/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1002 $
 * $Date: 2020-04-10 23:31:24 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.general.dtypes.editor2.TDataModel;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TNumeric;

/**
 * Representation of a TNumericModel - type
 * 
 * @author Johannes Kaisinger
 */
public class TNumericModel extends TDataModel {

	/** Data variable */
	private TNumeric data;

	/**
	 * Create object with size 1,1
	 * 
	 */
	public TNumericModel() {
		try {
			this.data = new TNumeric("0.0");
		} catch (TDataException e) {
		}
	}

	/**
	 * Create object with given data (TNumeric)
	 * 
	 * @param data
	 *            Reference of data (TNumeric)
	 */
	public TNumericModel(TNumeric data) {
		this.data = data;
	}

	/**
	 * Get all data (TNumeric)
	 * 
	 * @return data Reference of data (TNumeric)
	 */
	public TNumeric getTNumeric() {
		return this.data;
	}

	/**
	 * Get all data (TData)
	 * 
	 * @return Data Reference of data (TData)
	 */
	@Override
	public TData getData() {
		return this.data;
	}

	/**
	 * Get a single value as String
	 * 
	 * @param row
	 *            Index of row of array
	 * @param column
	 *            Index of column of array
	 * 
	 * @return Child Values of child as String
	 */
	public String getValue(int row, int column) {
		String child = null;
		if ((row > -1) && (column > -1)) {
			child = this.data.getStringVal(row, column);
		}
		return child;
	}

	/**
	 * Set a child value
	 * 
	 * @param row
	 *            Index of row of array
	 * @param column
	 *            Index of column of array
	 * 
	 * @param value
	 *            New value of child (String)
	 * 
	 */
	public void setValue(int row, int column, String value) {
		if ((row > -1) && (column > -1)) {
			try {
				this.data.setStringVal(row, column, value);
			} catch (TDataException e) {
			}
		}
	}

	/**
	 * Insert a empty row
	 * 
	 * @param row
	 *            Index of row of array (place of insert)
	 * 
	 */
	public void insertRow(final int row) {
		if (row > -1) {

			final TNumeric temp = this.data;

			final double[][] real = temp.getRealMatrix().getData();
			final double[][] imag = temp.getImagMatrix().getData();

			final int rowSize = real.length;
			final int colSize = real[0].length;

			final double[][] newReal = new double[rowSize + 1][colSize];
			final double[][] newImag = new double[rowSize + 1][colSize];

			final int newRowSize = newReal.length;
			final int newColSize = newReal[0].length;

			for (int m = 0; m < newRowSize; m++) {
				for (int n = 0; n < newColSize; n++) {
					if (m == row) {

						newReal[m][n] = 0.0;
						newImag[m][n] = 0.0;

					} else if (m < row) {
						newReal[m][n] = real[m][n];
						newImag[m][n] = imag[m][n];
					} else if (m > row) {
						newReal[m][n] = real[m - 1][n];
						newImag[m][n] = imag[m - 1][n];
					}
				}
			}

			try {
				this.data = new TNumeric(newReal, newImag);
			} catch (TDataException e) {
			}
		}
	}

	/**
	 * Insert a empty column
	 * 
	 * @param column
	 *            Index of column of array (place of insert)
	 * 
	 */
	public void insertColumn(final int column) {
		if (column > -1) {

			final TNumeric temp = this.data;

			final double[][] real = temp.getRealMatrix().getData();
			final double[][] imag = temp.getImagMatrix().getData();

			final int rowSize = real.length;
			final int colSize = real[0].length;

			final double[][] newReal = new double[rowSize][colSize + 1];
			final double[][] newImag = new double[rowSize][colSize + 1];

			final int newRowSize = newReal.length;
			final int newColSize = newReal[0].length;

			for (int m = 0; m < newRowSize; m++) {
				for (int n = 0; n < newColSize; n++) {
					if (n == column) {

						newReal[m][n] = 0.0;
						newImag[m][n] = 0.0;

					} else if (n < column) {
						newReal[m][n] = real[m][n];
						newImag[m][n] = imag[m][n];
					} else if (n > column) {
						newReal[m][n] = real[m][n - 1];
						newImag[m][n] = imag[m][n - 1];
					}
				}
			}

			try {
				this.data = new TNumeric(newReal, newImag);
			} catch (TDataException e) {
			}
		}
	}

	/**
	 * Delete a row
	 * 
	 * @param row
	 *            Index of row of array (place of delete)
	 * 
	 */
	public void deleteRow(final int row) {
		if (row > -1) {

			final TNumeric temp = this.data;

			final double[][] real = temp.getRealMatrix().getData();
			final double[][] imag = temp.getImagMatrix().getData();

			final int rowSize = real.length;
			final int colSize = real[0].length;

			if (rowSize > 1) {

				final double[][] newReal = new double[rowSize - 1][colSize];
				final double[][] newImag = new double[rowSize - 1][colSize];

				final int newRowSize = newReal.length;
				final int newColSize = newReal[0].length;

				int index = 0;
				for (int m = 0; m < newRowSize; m++) {
					for (int n = 0; n < newColSize; n++) {
						if (m == row) {
							index = 1;
							if ((m + index) <= newRowSize) {

								newReal[m][n] = real[m + index][n];
								newImag[m][n] = imag[m + index][n];
							}
						} else if ((m + index) <= newRowSize) {
							newReal[m][n] = real[m + index][n];
							newImag[m][n] = imag[m + index][n];
						}
					}
				}

				try {
					this.data = new TNumeric(newReal, newImag);

				} catch (final TDataException e) {
				}

			}
		}
	}

	/**
	 * Delete a column
	 * 
	 * @param column
	 *            Index of column of array (place of delete)
	 * 
	 */
	public void deleteColumn(int column) {

		if (column > -1) {

			final TNumeric temp = this.data;

			final double[][] real = temp.getRealMatrix().getData();
			final double[][] imag = temp.getImagMatrix().getData();

			final int rowSize = real.length;
			final int colSize = real[0].length;

			final double[][] newReal = new double[rowSize][colSize - 1];
			final double[][] newImag = new double[rowSize][colSize - 1];

			final int newRowSize = newReal.length;
			final int newColSize = newReal[0].length;

			for (int m = 0; m < newRowSize; m++) {
				int index = 0;
				for (int n = 0; n < newColSize; n++) {
					if (n == column) {
						index = 1;
						if ((n + index) <= newColSize) {

							newReal[m][n] = real[m][n + index];
							newImag[m][n] = imag[m][n + index];

						}
					} else if ((n + index) <= newColSize) {
						newReal[m][n] = real[m][n + index];
						newImag[m][n] = imag[m][n + index];
					}
				}
			}

			try {
				this.data = new TNumeric(newReal, newImag);
			} catch (final TDataException e) {
			}
		}

	}
}
