/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1002 $
 * $Date: 2020-04-10 23:31:24 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.general.dtypes.editor2;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.Vector;

import javax.swing.DefaultCellEditor;
import javax.swing.JButton;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.ListSelectionModel;
import javax.swing.SwingConstants;
import javax.swing.table.JTableHeader;
import javax.swing.text.JTextComponent;

import at.lcm.bu21.general.dtypes.editor.icons.IconEditor;
import at.lcm.bu21.general.dtypes.editor2.TDataModel.TStructModel;

/**
 * This panel allows to analyze and manipulate TStructs. It based on
 * TStructModel and TStructTableModel for JTables.
 * 
 * @author Johannes Kaisinger
 */
public class TStructPanel extends JPanel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/**
	 * Local reference of the tStructModel. It allows to manipulate the TStruct.
	 */
	private TStructModel tStructModel;

	/** Contains a table model for the JTable. */
	private TStructTableModel tableModel;

	/** Editor as table. */
	private JTable table;

	/** Contains a Listener. */
	private EditChildListener editChildListener;

	/**
	 * Creates an object
	 * 
	 * @param tStructModel
	 *            The tStructModel
	 */
	public TStructPanel(TStructModel tStructModel) {

		this.tStructModel = tStructModel;

		this.setLayout(new BorderLayout());

		// set up toolbar
		JToolBar toolbar = this.createToolBar();
		this.add(toolbar, BorderLayout.NORTH);

		// set up table
		this.tableModel = new TStructTableModel(tStructModel);
		this.table = new JTable(this.tableModel) {

			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			// Implement table cell tool tips.
			@Override
			public String getToolTipText(MouseEvent e) {
				String tip = null;
				java.awt.Point p = e.getPoint();
				int rowIndex = this.rowAtPoint(p);
				int colIndex = this.columnAtPoint(p);

				try {
					tip = this.getValueAt(rowIndex, colIndex).toString();
				} catch (RuntimeException e1) {
					// catch null pointer exception if mouse is over an empty
					// line
				}

				return tip;
			}
		};
		this.table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
		this.table.getColumnModel().getColumn(0).setPreferredWidth(100);
		this.table.getColumnModel().getColumn(1).setPreferredWidth(300);

		this.table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

		this.table.setDefaultEditor(String.class, new DefaultCellEditor(new JTextField()) {

			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			@Override
			public Component getTableCellEditorComponent(final JTable table, final Object value,
					final boolean isSelected, final int row, final int column) {
				Component c;
				c = super.getTableCellEditorComponent(table, value, isSelected, row, column);
				((JTextField) c).selectAll();
				return c;
			}
		});

		this.tableModel.setTStructKeyListener(new TStructKeyListener() {

			@Override
			public void keyChangedOccurred(TStructKeyChangedEvent evt) {
				String newKey = evt.getNewKey();

				for (int row = 0; row < TStructPanel.this.table.getRowCount(); row++) {
					String key = (String) TStructPanel.this.table.getValueAt(row, 0);
					if (newKey == key) {
						TStructPanel.this.table.addRowSelectionInterval(row, row);
						TStructPanel.this.table.addColumnSelectionInterval(0,
								TStructPanel.this.table.getColumnCount() - 1);
						break;
					}
				}
			}
		});

		this.createMouseListener();

		// configure table Header
		JTableHeader tableHeader = this.table.getTableHeader();
		tableHeader.setReorderingAllowed(false);

		JScrollPane scroll = new JScrollPane(this.table);

		this.addPopupMenu();

		this.add(scroll, BorderLayout.CENTER);
		this.setVisible(true);
	}

	/**
	 * Returns the panel.
	 * 
	 * @return panel Panel
	 */
	public JPanel getPanel() {
		return this;
	}

	/**
	 * Set a edit child listener.
	 * 
	 * @param editChildListener
	 *            The EditChildListener
	 */
	public void setEditChildListener(EditChildListener editChildListener) {
		this.editChildListener = editChildListener;
	}

	/**
	 * Adds a mouse listener to the table.
	 */
	private void createMouseListener() {
		this.table.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				if (e.getClickCount() == 2) {
					int column = TStructPanel.this.table.getSelectedColumn();
					if (column == 1) {
						TStructPanel.this.editChild();
					}
				}
			}

			@Override
			public void mousePressed(MouseEvent e) {

				if (e.getClickCount() == 2) {
					int column = TStructPanel.this.table.getSelectedColumn();
					if (column == 0) {
						Component editor = TStructPanel.this.table.getEditorComponent();
						editor.requestFocusInWindow();
						((JTextComponent) editor).selectAll();
					}

				}
			}

		});
	}

	/**
	 * Throws an edit child event.
	 */
	private void editChild() {
		int row = this.table.getSelectedRow();

		String key = (String) this.table.getValueAt(row, 0);
		EditChildEvent ev = new EditChildEvent(this, 0, 0, key);

		if (this.editChildListener != null) {
			this.editChildListener.editChildOccurred(ev);
		}
	}

	/**
	 * Adds a row; Selects the new row;
	 */
	private void addRow() {
		String newKey = this.tStructModel.addChild();
		this.tableModel.fireTableDataChanged();

		for (int row = 0; row < this.table.getRowCount(); row++) {
			String key = (String) this.table.getValueAt(row, 0);
			if (newKey == key) {
				this.table.addRowSelectionInterval(row, row);
				this.table.addColumnSelectionInterval(0, this.table.getColumnCount() - 1);
				break;
			}
		}

	}

	/**
	 * Removes a row; Selects the next row;
	 */
	private void removeRow() {
		int row = this.table.getSelectedRow();

		if (row > -1) {

			String key = (String) this.table.getValueAt(row, 0);

			this.tStructModel.removeChild(key);
			this.tableModel.fireTableDataChanged();

			if (this.table.getRowCount() > row) {
				this.table.addRowSelectionInterval(row, row);
				this.table.addColumnSelectionInterval(0, this.table.getColumnCount() - 1);
			} else if (this.table.getRowCount() > 0) {
				this.table.addRowSelectionInterval(this.table.getRowCount() - 1, this.table.getRowCount() - 1);
				this.table.addColumnSelectionInterval(0, this.table.getColumnCount() - 1);
			}

		}
	}

	/**
	 * Create the popupMenu and mousListener for the TStructPanel;
	 * 
	 * @return PopupMenu
	 */
	private JPopupMenu addPopupMenu() {

		final JPopupMenu popupMenu = new JPopupMenu();

		JMenuItem addChild = new JMenuItem("Add child");
		JMenuItem removeChild = new JMenuItem("Remove child");
		JMenuItem editChild = new JMenuItem("Edit child");

		popupMenu.add(addChild);
		popupMenu.add(removeChild);
		popupMenu.addSeparator();
		popupMenu.add(editChild);

		addChild.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TStructPanel.this.addRow();
			}
		});

		removeChild.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TStructPanel.this.removeRow();
			}
		});

		editChild.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				TStructPanel.this.editChild();
			}
		});

		this.table.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseReleased(MouseEvent e) {
				if (e.isPopupTrigger()) {
					JTable source = (JTable) e.getSource();
					int row = source.rowAtPoint(e.getPoint());
					int column = source.columnAtPoint(e.getPoint());

					if (!source.isRowSelected(row)) {
						source.changeSelection(row, column, false, false);
					}

					popupMenu.show(e.getComponent(), e.getX(), e.getY());
				}
			}

		});

		return popupMenu;
	}

	/**
	 * Create the toolBar for the TStructPanel;
	 * 
	 * @return toolBar
	 */
	private JToolBar createToolBar() {
		JToolBar tb = new JToolBar();
		tb.setPreferredSize(new Dimension(500, 50));

		Vector<JButton> buttons = new Vector<JButton>();

		final JButton addRow = new JButton("Add", IconEditor.ICON_ADD);
		addRow.setToolTipText("Add a row");
		addRow.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				TStructPanel.this.addRow();
			}

		});
		buttons.add(addRow);

		final JButton removeRow = new JButton("Remove", IconEditor.ICON_REMOVE);
		removeRow.setToolTipText("Remove selected row");
		removeRow.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				TStructPanel.this.removeRow();
			}

		});
		buttons.add(removeRow);

		for (final JButton i : buttons) {

			Dimension d = new Dimension(70, 50);
			i.setVerticalTextPosition(SwingConstants.BOTTOM);
			i.setHorizontalTextPosition(SwingConstants.CENTER);
			i.setMaximumSize(d);
			i.setMinimumSize(d);
			i.setPreferredSize(d);
			tb.add(i);
		}

		return tb;
	}
}
