from lxml import etree
from lcm.utils.common_fnc.array_fnc import is_empty


def listToString(lt):
    return ' '.join(map(str, lt))


class Fmi2ModelStructure:
    """
    Class for the <ModelStructure> xml-tag

    XML Schema Description
    ----------------------
    Ordered lists of outputs, exposed state derivatives and the initial unknowns.
    Optionally, the functional dependency of these variables can be defined.
    """

    def __init__(self, root, model_variables):
        from lcm.fmi.fmi2_xml_elements import Fmi2ModelVariables
        assert isinstance(model_variables, Fmi2ModelVariables)
        self.state_references = model_variables.state_references
        self.derivative_references = model_variables.derivative_references
        self.output_references = model_variables.output_references
        self.initial_unknown_references = model_variables.initial_unknown_references
        self.dependencies = model_variables.dependencies
        self.element = etree.SubElement(root, 'ModelStructure')
        self.__addOutputs()
        # ToDo: identify problem at Dymola import if self.__addDerivatives() is active
        self.__addDerivatives()
        self.__addInitialUnknowns()

    def __addOutputs(self):
        """
        Add the <Outputs> xml-tag to <ModelStructure>

        XML Schema Description
        ----------------------
        Ordered list of all outputs.
        Exactly all variables with causality="output" must be in this list.
        The dependency definition holds for Continuous-Time and for Event Mode (ModelExchange)
        and for Communication Points (CoSimulation).


        Returns
        -------

        """
        if not is_empty(self.output_references):
            sub_element = etree.SubElement(self.element, 'Outputs')
            for outputs in self.output_references:
                index = str(outputs)
                d = {'index': index}
                if index in self.dependencies:
                    d['dependencies'] = listToString(self.dependencies[index])
                etree.SubElement(sub_element, 'Unknown', **d)

    def __addDerivatives(self):
        """
        Add the <Derivatives> xml-tag to <ModelStructure>

        XML Schema Description
        ----------------------
        Ordered list of all exposed state derivatives (and therefore implicitely associated continuous-time states).
        Exactly all state derivatives of a ModelExchange FMU must be in this list.
        A CoSimulation FMU need not expose its state derivatives.
        If a model has dynamic state selection, introduce dummy state variables.
        The dependency definition holds for Continuous-Time and for Event Mode (ModelExchange)
        and for Communication Points (CoSimulation).


        Returns
        -------

        """
        if not is_empty(self.derivative_references):
            sub_element = etree.SubElement(self.element, 'Derivatives')
            for dervatives in self.derivative_references:
                index = str(dervatives)
                d = {'index': index}
                if index in self.dependencies:
                    d['dependencies'] = listToString(self.dependencies[index])

                etree.SubElement(sub_element, 'Unknown', **d)

    def __addInitialUnknowns(self):
        """
        Add the <InitialUnknowns> xml-tag to <ModelStructure>
        Initial Unknowns consist of the states, derivatives and outputs


        XML Schema Description
        ----------------------
        Ordered list of all outputs.
        Exactly all variables with causality="output" must be in this list.
        The dependency definition holds for Continuous-Time and for Event Mode (ModelExchange)
        and for Communication Points (CoSimulation).


        Returns
        -------

        """
        if not is_empty(self.initial_unknown_references):
            sub_element = etree.SubElement(self.element, 'InitialUnknowns')

            for unknowns in self.initial_unknown_references:
                index = str(unknowns)
                d = {'index': index}
                if index in self.dependencies:
                    d['dependencies'] = listToString(self.dependencies[index])
                etree.SubElement(sub_element, 'Unknown', **d)
