import numpy as np
from lxml import etree
from lcm.utils.common_fnc import is_integer
from lcm.utils.common_fnc import is_array_like


class Fmi2ModelVariables:
    """
    Class for the ModelVariables xml-tag

    XML Schema Description
    ----------------------
    Ordered list of all variables (first definition has index = 1).
    """

    def __init__(self, root):
        self._element = etree.SubElement(root, 'ModelVariables')
        self.variable_index = 0
        self.value_references = {'Real': -1, 'Integer': -1, 'Boolean': -1, 'String': -1, 'Enumeration': -1}
        self.state_references = []
        self.derivative_references = []
        self.output_references = []
        self.initial_unknown_references = []
        self.causalities = ["parameter", "calculatedParameter", "input", "output", "local", "independent"]
        self.variabilities = ["constant", "fixed", "tunable", "discrete", "continuous"]
        self.initials = ['exact', 'approx', 'calculated', None]

        self.dependencies = {}

        self.attributes = {}
        self.model_variables = []

    @property
    def is_derivative(self):
        derivative = self.attributes['derivative']
        if derivative is not None:
            if len(derivative) > 0:
                return True
            else:
                return False
        else:
            return False

    def fillElements(self):
        for value in self.model_variables:
            for index in range(len(value['variable_indices'])):
                scalar_variable = {'name': value['variable_names'][index],
                                   'valueReference': str(value['value_references'][index]),
                                   'description': value['variable_descriptions'][index],
                                   'causality': value['causality'],
                                   'variability': value['variability']}

                if value['initial'] is not None:
                    scalar_variable['initial'] = value['initial']

                variable_start = value['variable_start'][index]
                variable_derivative = value['variable_derivatives'][index]
                attributes = {}
                if value['type'] == 'Real':
                    attributes['unit'] = value['unit']
                if variable_start is not None:
                    attributes['start'] = str(variable_start)
                if variable_derivative is not None:
                    attributes['derivative'] = str(variable_derivative)

                self._element.append(
                    etree.Comment("Index of variable = " + "'" + str(value['variable_indices'][index]) + "'"))
                sub_element = etree.SubElement(self._element, 'ScalarVariable', **scalar_variable)
                etree.SubElement(sub_element, value['type'], **attributes)

    def addDependencies(self, name, dependency_references):
        flat_list = np.array(dependency_references).flatten()
        for value in self.model_variables:
            if value['name'] == name:
                for index in range(len(value['variable_indices'])):
                    key = str(value['variable_indices'][index])
                    if key in self.dependencies:
                        raise Exception
                    self.dependencies[key] = flat_list

    def __addValueReference(self):
        """
        Increase the value_reference,
        add an index comment tag
        """
        type = self.attributes['type']
        causality = self.attributes['causality']
        initial = self.attributes['initial']
        is_state = self.attributes['is_state']

        self.value_references[type] += 1

        self.variable_index += 1
        self.attributes['variable_indices'].append(self.variable_index)
        self.attributes['value_references'].append(self.value_references[type])

        if causality is "output" and initial in ['approx', 'calculated']:
            self.initial_unknown_references.append(self.variable_index)
        elif causality is "calculatedParameter":
            self.initial_unknown_references.append(self.variable_index)
        elif is_state and initial in ['approx', 'calculated']:
            self.initial_unknown_references.append(self.variable_index)
        elif self.is_derivative and initial in ['approx', 'calculated']:
            self.initial_unknown_references.append(self.variable_index)

        if causality is "output":
            self.output_references.append(self.variable_index)
        if is_state:
            self.state_references.append(self.variable_index)
        if self.is_derivative:
            self.derivative_references.append(self.variable_index)

    def add(self, name, description, unit,
            size, causality='local', variability='continuous', is_state=False,
            derivative=None,
            start=None, initial=None, variable_type='Real'):

        if derivative is None:
            derivative = []

        self.attributes = {'name': name,
                           'description': description,
                           'causality': causality,
                           'variability': variability,
                           'initial': initial,
                           'unit': unit,
                           'is_state': is_state,
                           'derivative': derivative,
                           'start': start,
                           'type': variable_type,
                           'size': size,
                           'value_references': [],
                           'variable_indices': [],
                           'variable_names': [],
                           'variable_descriptions': [],
                           'variable_derivatives': [],
                           'variable_start': []}

        self.__checkAttributes()
        self.__getDefaultInitial()
        self.__checkInitial()
        self.__checkAttributeCombination()
        self.__checkSize()
        self.__variableArray()
        self.model_variables.append(self.attributes)
        return self.attributes['variable_indices']

    def __variableArray(self):
        name = self.attributes['name']
        size = self.attributes['size']
        unit = self.attributes['unit']
        description = self.attributes['description']

        n_dim1 = size[0]
        n_dim2 = size[1]
        for dim1 in range(n_dim1):
            for dim2 in range(n_dim2):
                self.attributes['variable_descriptions'].append(description + '[' + unit + ']')
                if n_dim1 == 1 and n_dim2 == 1:
                    """scalar variable"""
                    self.attributes['variable_names'].append(name)
                elif (n_dim1 == 1 and n_dim2 > 1) or (n_dim1 > 1 and n_dim2 == 1):
                    """variable vector"""
                    index = dim1 if n_dim1 > n_dim2 else dim2
                    self.attributes['variable_names'].append(name + "[" + str(index) + "]")
                elif n_dim1 > 1 and n_dim2 > 1:
                    """variable matrix"""
                    self.attributes['variable_names'].append(name + "[" + str(dim1) + "," + str(dim2) + "]")
                else:
                    raise Exception("Wrong size dimensions")

                self.__scalarVariable(dim1, dim2)

    def __scalarVariable(self, dim1, dim2):
        type = self.attributes['type']
        size = self.attributes['size']
        start = self.attributes['start']
        derivative = self.attributes['derivative']
        n_dim1 = size[0]

        self.__addValueReference()

        if type in ['Real', 'Integer']:

            variable_derivative = derivative[n_dim1 * dim1 + dim2] if self.is_derivative else None
            variable_start = start[dim1, dim2] if start is not None else None
            self.attributes['variable_derivatives'].append(variable_derivative)
            self.attributes['variable_start'].append(variable_start)

        else:
            raise NotImplementedError('Variable Type: ' + str(type) + ' is not implemented!')

    def __checkSize(self):
        name = self.attributes['name']
        start = self.attributes['start']
        size = self.attributes['size']

        if is_integer(size):
            size = np.array(size)
            size = np.expand_dims(size, axis=0)
        if size.shape[0] == 1:
            size = np.append([1], size)

        if start is not None:
            if not np.array_equal(start.shape, size):
                raise Exception(
                    name + ' Sizes do not match, start.shape=' + str(start.shape) + ' size=' + str(size))

        if size.size != 2:
            raise Exception("Size dimension higher than 2 is not supported yet!")

        self.attributes['size'] = size

    def __checkAttributes(self):
        causality = self.attributes['causality']
        variability = self.attributes['variability']
        initial = self.attributes['initial']
        if causality not in self.causalities:
            raise ValueError('Causality:' + causality + ' is not supported')
        if variability not in self.variabilities:
            raise ValueError('Variability:' + variability + ' is not supported')
        if initial not in self.initials:
            raise ValueError('Initial:' + initial + ' is not supported')

    def __checkAttributeCombination(self):
        name = self.attributes['name']
        causality = self.attributes['causality']
        variability = self.attributes['variability']
        initial = self.attributes['initial']
        start = self.attributes['start']
        is_state = self.attributes['is_state']

        error_msg = name + ': not allowed combination causality: ' + causality + \
                    ' and variability: ' + variability + ' with initial: ' + str(initial)

        if causality not in ['output', 'local'] and variability not in ["discrete", "continuous"] and (
                    is_state or self.is_derivative):
            raise ValueError(name + ': not allowed combination causality: ' + causality + \
                             ' and variability: ' + variability + ' with is_state: ' + str(
                is_state) + ', is_derivative: ' + str(self.is_derivative))

        if causality == 'parameter' and variability in ["fixed", "tunable"] and initial != 'exact':
            raise ValueError(error_msg)
        elif causality == 'calculatedParameter' and variability in ["fixed", "tunable"] \
                and initial not in ['approx', 'calculated']:
            raise ValueError(error_msg)
        elif causality == 'input' and variability in ["discrete", "continuous"]:
            if initial is not None or start is None:
                raise ValueError(error_msg)
        elif causality == 'output':
            if variability == "constant" and (initial != 'exact'):
                raise ValueError(error_msg)
            elif variability in ["discrete", "continuous"] and initial not in ['exact', 'approx',
                                                                               'calculated']:
                raise ValueError(error_msg)
        elif causality == 'local':
            if variability == "constant" and (initial != 'exact'):
                raise ValueError(error_msg)
            elif variability in ["fixed", "tunable"] and initial not in ['approx', 'calculated']:
                raise ValueError(error_msg)
            elif variability in ["discrete", "continuous"] and initial not in ['exact', 'approx',
                                                                               'calculated']:
                raise ValueError(error_msg)
        elif causality == 'independent' and initial is not None:
            raise ValueError(error_msg)

    def __getDefaultInitial(self):
        causality = self.attributes['causality']
        variability = self.attributes['variability']
        initial = self.attributes['initial']
        start = self.attributes['start']
        if initial is None:
            initial = None
            if causality == 'parameter' and variability in ["fixed", "tunable"]:
                initial = 'exact'
            elif causality == 'calculatedParameter' and variability in ["fixed", "tunable"] and start is None:
                initial = 'calculated'
            elif causality == 'output':
                if variability == "constant":
                    initial = 'exact'
                elif variability in ["discrete", "continuous"] and start is None:
                    initial = 'calculated'
            elif causality == 'local':
                if variability == "constant":
                    initial = 'exact'
                elif variability in ["fixed", "tunable"] and start is None:
                    initial = 'calculated '
                elif variability in ["discrete", "continuous"] and start is None:
                    initial = 'calculated'
            self.attributes['initial'] = initial

    def __checkInitial(self):
        initial = self.attributes['initial']
        start = self.attributes['start']
        if initial is 'calculated' and start is not None:
            raise ValueError('Start attribute is not allowed for variables with initial=calculated')
