import os
import stat
import shutil
import subprocess
import msgpack
import matplotlib.pyplot as plt
from lxml import etree
import numpy as np
from abc import ABCMeta, abstractmethod
from lcm.fmi.fmi2_xml_elements import Fmi2ModelDescription
from lcm.fmi.fmi2_xml_elements import Fmi2ModelVariables
from lcm.fmi.fmi2_xml_elements.fmi2_model_type import Fmi2ModelExchange, Fmi2CoSimulation
from lcm.fmi.fmi2_xml_elements import Fmi2UnitDefinitions
from lcm.fmi.fmi2_xml_elements import Fmi2ModelStructure
from lcm.utils.common_fnc.os_fnc import handleRemoveReadonly
from lcm.fmi.fmu_model_data import AbstractFMUModelData
from symspace.osutils.pathutils import symspace_home


class AbstractFMUCreator(metaclass=ABCMeta):
    """
    Abstract class for FMU Creation
    """

    def __init__(self, model_name, model_data, fmu_image_path=None, *args, **kwargs):
        """

        Parameters
        ----------
        model_name : basestring
            Name of the fmu model
        model_data : AbstractFMUModelData
            * Model sizes
            * Resource data, message pack serializable
        fmu_image_path : basestring
            Path to a model image which will replace the model.png from the fmu_skeleton_dir dir (Optional),
             None (Default)
        args
        kwargs
        """
        # assert isinstance(model_data, AbstractFMUModelData)
        self.__model_name = model_name
        self.__model_data = model_data
        self.__logo_path = os.path.join(symspace_home, 'System', 'General', 'Python', 'resources', 'lcm_logo.png')
        self.__guid = "{8c4e810f-3df3-4a00-8276-176fa3c9f000}"
        self.__fmi_version = "2.0"
        self.__generation_tool = "SyMSpace"
        self.__fmu_out_dir = None
        self.__fmu_dir = None
        self.__fmu_image_path = fmu_image_path

    @property
    @abstractmethod
    def fmu_skeleton_dir(self):
        """
        Path to the directory of the FMU skeleton directory
        * includes binaries, documentation, resources, sources, model.png

        Returns
        -------
        basestring
        """
        pass

    @property
    def fmi_version(self):
        """
        FMI version

        Returns
        -------
        basestring
        """
        if self.__fmi_version != "2.0":
            raise NotImplementedError("FMI Versions other than 2.0 are not supported, yet!")
        return self.__fmi_version

    @property
    def model_data(self):
        """
        Model data includes e.g. the model sizes, can be message pack serializable resource data

        Returns
        -------
        AbstractFMUModelData or
        """
        return self.__model_data

    @property
    def model_name(self):
        """
        Name of the fmu model

        Returns
        -------
        basestring
        """
        return self.__model_name

    @property
    def model_identifier(self):
        """
        FMU model identifier (FMU model name without whitespaces)
        Returns
        -------
        basestring
        """
        return self.model_name.replace(" ", "_")

    @property
    def guid(self):
        """
        Globally Unique Identifier of the FMU

        Argument fmuGUID is used to check that the modelDescription.xml file is compatible with the C code of the FMU.
        It is a vendor specific globally unique identifier of the XML file
        (for example it is a “fingerprint” of the relevant information stored in the XML file).
        It is stored in the XML file as attribute “guid”.
        It must be identical to the one stored inside the fmi2Instantiate function.
        Otherwise the C code and the XML file of the FMU are not consistent to each other. This argument cannot be null.

        Returns
        -------
        basestring
        """

        return self.__guid

    @property
    def description(self):
        """
        FMU model description

        Optional string with a brief description of the model.

        Returns
        -------
        basestring
        """
        return self.__model_name

    @property
    def generation_tool(self):
        """
        Optional name of the tool that generated the XML file.

        Returns
        -------
        basestring
        """
        return self.__generation_tool

    @property
    def fmu_out_dir(self):
        """
        Path to the directory of the (zipped) FMU

        Returns
        -------
        basestring
        """

        if self.__fmu_out_dir is None:
            raise Exception("Output directory is not set!")
        return self.__fmu_out_dir

    @fmu_out_dir.setter
    def fmu_out_dir(self, var):
        self.__fmu_out_dir = var

    @property
    def fmu_dir(self):
        """
        Path to the directory of the (unzipped) FMU

        Returns
        -------
        basestring
        """
        return os.path.join(self.fmu_out_dir, "fmu_" + self.model_identifier)

    @property
    def bin_dir(self):
        """
        Path to the directory of the (unzipped) FMU binary directory

        Returns
        -------
        basestring
        """
        return os.path.join(self.fmu_dir, "binaries")

    @property
    def fmu_path(self):
        """
        Path to the (zipped) FMU file (*.fmu)

        Returns
        -------
        basestring
        """
        return os.path.join(self.fmu_out_dir, self.model_identifier + '.fmu')

    @property
    def msgpack_data_path(self):
        """
        Path to the  message pack data file of the (unzipped) FMU

        Returns
        -------

        """
        return os.path.join(self.fmu_dir, 'resources', 'data.mp')

    @property
    def model_description_path(self):
        """
        Path to the model description file of the (unzipped) FMU

        Returns
        -------

        """
        return os.path.join(self.fmu_dir, 'modelDescription.xml')

    @property
    @abstractmethod
    def physical_quantities(self):
        """
        List of physical quantities of the fmu variables

        e.g. ["torque", "angle", "angular_velocity", "current", "voltage", "flux", "power", "resistance"]

        Returns
        -------
        array_like
        """
        pass

    def __copy_fmu_skeleton(self):
        """
        Copy the fmu skeleton directory to the create a default template (with binaries, sources, model.png, etc.)
        for the (unzipped) FMU directory

        """
        if os.path.isdir(self.fmu_dir):
            '''If fmu_dir exist, delete it'''
            # ToDo Force remove
            shutil.rmtree(self.fmu_dir, ignore_errors=False, onerror=handleRemoveReadonly)

        '''Copy fmu_skeleton directory to the fmu directory'''
        shutil.copytree(self.fmu_skeleton_dir, self.fmu_dir)

    def __rename_compiled_files(self):
        """
        Rename files in the "/binaries" directory of type
        '*.dll' (Win),
        '*.so' (Linux) or
        '*.dylib'  (Darwin/MacOS)
        to match the "model_identifier"

        """
        self.__check_dir_exists(self.bin_dir)

        for (dirpath, dirnames, filenames) in os.walk(self.bin_dir):
            '''Walk through all directories in bin_dir and search for binary files, and rename them'''
            for file in os.listdir(dirpath):

                if 'fmi' in file:
                    '''rename only binaries which also include 'fmi' in the filename'''
                    original_file = os.path.join(dirpath, file)
                    renamed_file = os.path.join(dirpath, self.model_identifier)

                    if file.endswith('.dll'):
                        if original_file != renamed_file + '.dll':
                            shutil.move(original_file, renamed_file + '.dll')
                    if file.endswith('.so'):
                        if original_file != renamed_file + '.so':
                            shutil.move(original_file, renamed_file + '.so')
                    if file.endswith('.dylib'):
                        if original_file != renamed_file + '.dylib':
                            shutil.move(original_file, renamed_file + '.dylib')

    def __zip_fmu(self):
        """
        Create a zip-archive from the folder in the fmu directory
        and rename it to '*.fmu'

        Returns
        -------

        """
        ''' Create a zip-archive'''
        shutil.make_archive(self.model_identifier, 'zip', self.fmu_dir)
        ''' Move and rename the zip-archive'''
        shutil.move(self.model_identifier + '.zip', self.fmu_path)

    def check_fmu(self, fmu_checker_dir, options='me', log_level=3):
        """
        Checks the fmu using the fmuChecker

        Parameters
        ----------
        fmu_checker_dir : string
        options : string
            fmuChecker options:
            -k xml           Check XML only.
            -k me            Check XML and ME simulation.
            -k cs            Check XML and CS simulation.
                             Multiple -k options add up.
                             No -k option: test XML, simulate ME and CS respectively if supported.
        log_level : int
            -l <log level>
            Log level:
            0 - no logging
            1 - fatal errors only
            2 - errors
            3 - warnings
            4 - info
            5 - verbose,
            6 - debug

        """
        self.__check_dir_exists(self.bin_dir)
        self.__check_dir_exists(fmu_checker_dir)

        if os.name is 'nt':
            """execute this if run on a windows machine"""
            check_options = ('-k ', options)
            log_level_options = ('-l ', str(log_level))

            """Test 32bit FMU/dll"""
            self.__runFmuComplianceChecker(fmu_checker_dir, 'Win32', check_options, log_level_options)
            """Test 64bit FMU/dll"""
            self.__runFmuComplianceChecker(fmu_checker_dir, 'Win64', check_options, log_level_options)

    def __runFmuComplianceChecker(self, fmu_checker_dir, platform, check_options, log_level_options):
        """

        Parameters
        ----------
        fmu_checker_dir : string
        platform : string
            Machine Platform: 'Win32' or 'Win64'
        check_options : tuple
            fmuChecker options:
            ('-k', 'xml')           Check XML only.
            ('-k', 'me')            Check XML and ME simulation.
            ('-k', 'cs')            Check XML and CS simulation.
        log_level_options : tuple
            ('-l', '0') - no logging
            ('-l', '1') - fatal errors only
            ('-l', '2') - errors
            ('-l', '3') - warnings
            ('-l', '4') - info
            ('-l', '5') - verbose,
            ('-l', '6') - debug

        """

        """path to the binary file"""
        binary_file = os.path.join(self.bin_dir, platform, self.model_identifier + '.dll')

        if os.path.isfile(binary_file) or check_options[1] == 'xml':
            """run this, if a binary file exists or only the model_description.xml should be checked"""
            program = os.path.join(fmu_checker_dir, "fmuCheck." + platform.lower() + ".exe")
            process = subprocess.run([program, check_options, log_level_options, self.fmu_path], stderr=subprocess.PIPE,
                                     universal_newlines=True)

            if os.path.isfile(program):
                """run this, an executable fmuChecker exists"""
                fmu_check_sumary = process.stderr.split("FMU reported:\n")[-1]
                fmu_report, checker_report = fmu_check_sumary.split("Checker reported:\n")
                fmu_errors = int(fmu_report.split("warning(s)")[0])

                if fmu_errors > 0:
                    """FMU Errors or Warnings found!"""
                    print(process.stderr)
                    raise Exception(platform + ":\nFMU reported:\n" + fmu_report)

                else:
                    """No FMU Errors or Warnings found!"""
                    print((platform + ":\nFMU reported:\n" + fmu_report))

                xml_errors = int((checker_report.split("Warning(s)\n\t")[-1]).split('Error(s)\n')[0])
                xml_warnings = int(checker_report.split("Warning(s)\n\t")[0])

                if xml_errors > 0:
                    """XML Errors found!"""
                    print(process.stderr)
                    raise Exception(platform + ":\nChecker reported:\n" + checker_report)
                elif xml_warnings > 0:
                    """XML Warnings found!"""
                    print(process.stderr)
                    raise Warning(platform + ":\nChecker reported:\n" + checker_report)
                else:
                    """No XML Errors or Warnings found!"""
                    print(platform + ":\nChecker reported:\n" + checker_report)
            else:
                """executable fmuChecker does not exists"""
                raise FileNotFoundError("Executable fmuChecker" + str(program) + "does not exists")
        else:
            """The FMU does not contain a binaries, for this platform"""
            raise FileNotFoundError("The FMU does not contain a binaries, for this platform: " + str(binary_file))

    def __write_to_xml(self, xml_string):
        """
        Writes the 'self.__xml_string' into the 'modelDescription.xml'

        Returns
        -------
        xml_string : string
        """
        with open(self.model_description_path, "w") as text_file:
            print(xml_string, file=text_file)

    def __write_msgpack(self):
        """
        Create msgpack file representing an instance of this class
        ref: http://msgpack.org/
        """

        # # ToDo: Remove Obsolete Check if to_msgpack is callable for AbstractFMUModelData
        # """check if self.model_data.to_msgpack is callable """
        # to_msgpack = getattr(self.model_data, "to_msgpack", None)
        # if callable(to_msgpack):
        try:
            data = self.model_data.to_msgpack()
            data["guid"] = self.__guid

            with open(self.msgpack_data_path, "wb") as fh:
                msgpack.dump(data, fh)
        except Exception as e:
            print("Error during AbstractFMUCreator.__write_msgpack:\n" + str(e))

    def create_fmu(self, fmu_out_dir, fmu_checker_dir=None, model_type='me'):
        """
        Creates FMU file

        Parameters
        ----------
        fmu_out_dir : string
            Path to the directory of the (zipped) FMU
        fmu_checker_dir : string
            Path to the directory of the FMU compliance checker (Optional
        model_type : string
            FMU model type:
            * 'me' - model exchange (Default)
            * 'cs' - co simulation

        """

        """finally set fmu output directory"""
        self.fmu_out_dir = fmu_out_dir
        self.__check_dir_exists(self.fmu_skeleton_dir)

        if os.path.isdir(self.fmu_dir):
            """delete it, if the fmu directory already exists (from previous fmu creations)"""
            shutil.rmtree(self.fmu_dir, ignore_errors=False, onerror=handleRemoveReadonly)

        xml_string = self.__get_xml(model_type)
        self.__copy_fmu_skeleton()
        self.__write_to_xml(xml_string)
        self.__write_msgpack()
        self.__rename_compiled_files()
        self.__copy_model_icon()
        self.__zip_fmu()
        if fmu_checker_dir is not None:
            self.check_fmu(fmu_checker_dir)
        '''Remove unpacked fmu directory'''
        # ToDo Force remove
        # shutil.rmtree(self.fmu_dir)
        # shutil.rmtree(self.fmu_dir, ignore_errors=False, onerror=handleRemoveReadonly)
        print('Path to the created FMU: ' + self.fmu_path)

    def gen_documentation(self):
        """
        Generate Documentation "index.html"

        Returns
        -------

        """
        pass

    def __gen_model_icon(self):
        """
        Generate Model Icon "model.png"
        Returns
        -------

        """
        from PIL import Image

        def changeBackgroundColor(img):
            img.thumbnail(img.size, Image.ANTIALIAS)
            new = Image.new('RGBA', img.size, (255, 255, 255, 0))  # with alpha
            new.paste(img, ((img.size[0] - img.size[0]) // 2, (img.size[1] - img.size[1]) // 2))
            return new

        def makeSquare(img, pad=1.0):
            longer_side = int(max(img.size) * pad)
            horizontal_padding = (longer_side - img.size[0]) // 2
            vertical_padding = (longer_side - img.size[1]) // 2
            img = img.crop(
                (
                    -horizontal_padding,
                    -vertical_padding,
                    img.size[0] + horizontal_padding,
                    img.size[1] + vertical_padding
                )
            )
            return img

        def getModelImage(path, size=(800, 800)):
            img = Image.open(path)
            img.thumbnail(size, Image.ANTIALIAS)
            return img

        def getLogoImage(path):
            img = Image.open(path)
            return img

        def pasteImage(img1, img2):
            x, y = img1.size
            w, h = img2.size

            img1.paste(img2, (x - w, y - h, x, y))
            return img1

        if self.__fmu_image_path is None:
            self.__fmu_image_path = os.path.join(self.fmu_dir, 'model.png')
            model_img = getModelImage(self.__fmu_image_path)
        else:
            model_img = getModelImage(self.__fmu_image_path, size=(800, 800))
            logo_img = getLogoImage(self.__logo_path)
            model_img = changeBackgroundColor(model_img)
            model_img = pasteImage(model_img, logo_img)
            model_img = makeSquare(model_img, pad=1.1)

        return model_img

    # def copy_documentation(self, fmu_dir):
    #     pass

    def __copy_model_icon(self):

        try:
            image_path = os.path.join(self.fmu_dir, 'model.png')
            # os.chmod(self.__fmu_image_path, stat.S_IWRITE)
            if os.path.isfile(image_path):
                os.chmod(image_path, stat.S_IWRITE)
                # os.remove(image_path)
            model_img = self.__gen_model_icon()
            model_img.save(image_path)
        except Exception as e:
            print("Error during AbstractFMUCreator.__copy_model_icon:\n" + str(e))

    def __get_xml(self, model_type):
        fmi_model_description = Fmi2ModelDescription(self.fmi_version,
                                                     self.model_name,
                                                     self.guid,
                                                     self.description,
                                                     self.generation_tool).element

        if model_type == 'me':
            Fmi2ModelExchange(fmi_model_description, self.model_identifier, self.fmu_dir)
        else:
            Fmi2CoSimulation(fmi_model_description, self.model_identifier, self.fmu_dir)

        Fmi2UnitDefinitions(fmi_model_description, self.physical_quantities)
        model_variables = self.get_model_variables(Fmi2ModelVariables(fmi_model_description))
        model_variables.fillElements()
        Fmi2ModelStructure(fmi_model_description, model_variables)
        root_tree = etree.ElementTree(fmi_model_description)
        return etree.tostring(root_tree, pretty_print=True, xml_declaration=True, encoding='UTF-8').decode('ascii')

    @abstractmethod
    def get_model_variables(self, model_variables):
        """
        Return model variables for the model description file

        Parameters
        ----------
        model_variables : Fmi2ModelVariables

        Returns
        -------
        model_variables : Fmi2ModelVariables

        """
        pass

    @abstractmethod
    def simulate(self, tend=1.0):
        """
        Default Experiment (e.g. with PyFMI)

        Parameters
        ----------
        tend : float
            Simulation stop time in seconds 1.0 (Default)

        Returns
        -------

        """
        pass

    @staticmethod
    def __check_dir_exists(directory_path):
        """Perform check if a directory exists, otherwise raise Exception

        Parameters
        ----------
        directory_path : basestring
            Path to the directory

        Returns
        -------

        """
        if not os.path.isdir(directory_path):
            raise FileNotFoundError("Directory " + directory_path + " does not exist!")

    def createPlot(self, results, x_var, y_var, n_y=1, ax=None, refdata=None):
        if ax is None:
            plt.figure()
            ax = plt.gca()
        for index in range(n_y):
            variable = y_var if n_y == 1 else y_var + '[' + str(index) + ']'
            ax.plot(results[x_var], results[variable], label=variable)
            try:
                yinterp = np.interp(results[x_var], refdata[x_var], refdata[variable])
                ax.plot(results[x_var], yinterp, label='ref: ' + variable)
                # ax.plot(results[x_var], refdata[variable], label='ref: ' + variable)
            except Exception as e:
                print("Error during AbstractFMUCreator.createPlot:\n" + str(e))

        if x_var == 'time':
            xlabel = 'Time [s]'
        else:
            xlabel = results.result_data.vars[x_var].description
        ylabel = results.result_data.vars[variable].description
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.grid(True)
        ax.legend(loc=1)

    def createSubplot(self, ax, results, x_var, y_var, n_y, n_ax=0, refdata=None):
        self.createPlot(results, x_var, y_var, n_y, ax[n_ax], refdata)
        n_ax += 1
        return n_ax
