/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9715 $
 * $Date: 2020-03-19 10:01:15 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.symspace.center.views.tree;

/*
 * Copyright 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer. 
 *   
 * - Redistribution in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials
 *   provided with the distribution. 
 *   
 * Neither the name of Sun Microsystems, Inc. or the names of
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.  
 * 
 * This software is provided "AS IS," without a warranty of any
 * kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY
 * EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY
 * DAMAGES OR LIABILITIES SUFFERED BY LICENSEE AS A RESULT OF OR
 * RELATING TO USE, MODIFICATION OR DISTRIBUTION OF THIS SOFTWARE OR
 * ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE 
 * FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT,   
 * SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER  
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF 
 * THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF SUN HAS 
 * BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * 
 * You acknowledge that this software is not designed, licensed or
 * intended for use in the design, construction, operation or
 * maintenance of any nuclear facility.
 */

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragSource;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.DragSourceListener;
import java.awt.dnd.DropTarget;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.geom.Rectangle2D;
import java.util.EventObject;

import javax.swing.AbstractAction;
import javax.swing.DefaultCellEditor;
import javax.swing.DropMode;
import javax.swing.Icon;
import javax.swing.JComponent;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.JTree;
import javax.swing.KeyStroke;
import javax.swing.ListSelectionModel;
import javax.swing.Timer;
import javax.swing.UIManager;
import javax.swing.border.Border;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.TableCellRenderer;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeSelectionModel;
import javax.swing.tree.TreeCellRenderer;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.magopt.structure.Parameter;
import at.lcm.bu21.magopt.structure.Project;

/**
 * This example shows how to create a simple JTreeTable component, by using a
 * JTree as a renderer (and editor) for the cells in a particular column in the
 * JTable.
 * 
 * @version 1.2 10/27/98
 * 
 * @author Philip Milne
 * @author Scott Violet
 */
public class ProTreeTable extends JTable implements DragGestureListener, DragSourceListener {
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	/** A subclass of JTree. */

	/** time (in ms) to expand/collapse a node after mouse was moved over it */
	private static final int HOVER_TIME = 2000;

	protected TreeTableCellRenderer tree = null;

	private int _selectedIndex = 0;
	private ProParameterModel _proParameterModel = null;
	ProTreeTableIndexViewManager _idxMgr = null;
	public ProTreeTableModelAdapter _modelAdapter = null;

	// TableColumn _selectedIndexColumn = null;
	// ---------------------------
	/** last path to node */
	private TreePath _pathLast = null;
	/** rectangle for cue geometry */
	private Rectangle2D _raCueGeometry = new Rectangle2D.Float();
	/** rectangle for ghost geometry */
	private Rectangle2D _raGhost = new Rectangle2D.Float();
	/** Timer for automatically expanding/collapsing tree paths */
	private Timer _timerHover;

	/** true if dragged node should put dropped into closed folder */
	public boolean _dropIntoFolder = false;

	private DropTarget _dropTarget = null;

	// ---------------------------

	public ProTreeTable(ProParameterModel treeTableModel) {
		super();

		// Creates the tree. It will be used as a renderer and editor.
		this.tree = new TreeTableCellRenderer(treeTableModel);

		// Installs a tableModel representing the visible rows in the tree.
		// super.setModel(new ProTreeTableModelAdapter(treeTableModel, tree));

		this._modelAdapter = new ProTreeTableModelAdapter(treeTableModel, this.tree);
		super.setModel(this._modelAdapter);

		this.setProParameterModel(treeTableModel);

		// Forces the JTable and JTree to share their row selection models.
		ListToTreeSelectionModelWrapper selectionWrapper = new ListToTreeSelectionModelWrapper();
		this.tree.setSelectionModel(selectionWrapper);
		this.setSelectionModel(selectionWrapper.getListSelectionModel());

		// Installs the tree editor renderer and editor.
		this.setDefaultRenderer(TreeTableModel.class, this.tree);

		this.setDefaultEditor(TreeTableModel.class, new TreeTableCellEditor());
		this.setDefaultEditor(TData.class, new ProTableCellEditor());

		ListSelectionModel listSelectionModel = this.getSelectionModel();
		listSelectionModel.addListSelectionListener(new ProTableSelectionListener(this));

		// No grid.
		this.setShowGrid(false);

		// No intercell spacing
		this.setIntercellSpacing(new Dimension(0, 0));

		// And update the height of the trees row to match that of
		// the table.
		if (this.tree.getRowHeight() < 1) {
			// Metal looks better like this.
			this.setRowHeight(20);
		}

		this.tree.setTreeTable(this);

		this._idxMgr = new ProTreeTableIndexViewManager(this);

		treeTableModel.setIndexViewManager(this._idxMgr);
		treeTableModel.setColumnStructure(this._idxMgr.getVisibleColumns());
		this._idxMgr.enableActiveIndexView(true);

		Project proj = (Project) this._proParameterModel.getRoot();
		if (proj != null) {
			this._idxMgr.setSelectedIndex(proj.getIndex());
		}

		DragSource dragSource = DragSource.getDefaultDragSource();
		dragSource.createDefaultDragGestureRecognizer(this, DnDConstants.ACTION_COPY_OR_MOVE, this);

		dragSource.addDragSourceListener(this);

		// Also, make this JTree a drag target
		// ProjectDropTargetListener dtl = tree.new
		// ProjectDropTargetListener(tree);
		this._dropTarget = new DropTarget(this, this.tree.new ProjectDropTargetListener(this.tree));
		// DropTarget dropTarget = new DropTarget(this, dtl);
		this._dropTarget.setDefaultActions(DnDConstants.ACTION_COPY_OR_MOVE);

		this.setDefaultRenderer(TData.class, new ProTableCellRenderer());

		this.setFillsViewportHeight(true);
		// Set up a hover timer, so that a node will be automatically
		// expanded or collapsed
		// if the user lingers on it for more than a short time

		this._timerHover = new Timer(ProTreeTable.HOVER_TIME, new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {

				if (ProTreeTable.this.tree.isRootPath(ProTreeTable.this._pathLast)) {
					return; // Do nothing if we are hovering over the
							// root
				}
				// node
				if (ProTreeTable.this.tree.isExpanded(ProTreeTable.this._pathLast)) {
					ProTreeTable.this.tree.collapsePath(ProTreeTable.this._pathLast);
				} else {
					ProTreeTable.this.tree.expandPath(ProTreeTable.this._pathLast);
				}
			}
		});
		this._timerHover.setRepeats(false); // Set timer to one-shot mode

		// this.setSurrendersFocusOnKeystroke(true);
		final String solve = "Solve";
		KeyStroke enter = KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0);
		this.getInputMap(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT).put(enter, solve);
		this.getActionMap().put(solve, new EnterAction());

		this.putClientProperty("terminateEditOnFocusLost", true);
	}

	private class EnterAction extends AbstractAction {
		private static final long serialVersionUID = 1L;

		@Override
		public void actionPerformed(ActionEvent e) {
			ProTreeTable table = (ProTreeTable) e.getSource();
			int selRow = table.getSelectedRow();
			@SuppressWarnings("unused")
			int selCol = table.getSelectedColumn();
			table.changeSelection(selRow + 1, 1, false, false);
			table.editCellAt(selRow + 1, 1);
			DnDTreeForTable tree = table.getTree();
			tree.updateUI();
		}
	}

	/**
	 * Overridden to message super and forward the method to the tree. Since the
	 * tree is not actually in the component hierarchy it will never receive this
	 * unless we forward it in this manner.
	 */
	@Override
	public void updateUI() {

		super.updateUI();
		if (this.tree != null) {
			this.tree.updateUI();
			// Do this so that the editor is referencing the current renderer
			// from the tree. The renderer can potentially change each time
			// laf changes.
			// setDefaultEditor(ProTreeTableModel.class, new
			// TreeTableCellEditor());
		}
		// Use the tree's default foreground and background colors in the
		// table.
		// LookAndFeel.installColorsAndFont(this, "Tree.background",
		// "Tree.foreground", "Tree.font");
	}

	/**
	 * Get selected Parameter
	 * 
	 * @return Selected Parameter
	 */
	public Parameter getSelectedParameter() {

		Parameter param = null;

		if (this.tree.getSelectionCount() > 0) {
			param = (Parameter) this.tree.getSelectionPath().getLastPathComponent();
		}
		return param;
	}

	/**
	 * Workaround for BasicTableUI anomaly. Make sure the UI never tries to resize
	 * the editor. The UI currently uses different techniques to paint the renderers
	 * and editors; overriding setBounds() below is not the right thing to do for an
	 * editor. Returning -1 for the editing row in this case, ensures the editor is
	 * never painted.
	 */
	@Override
	public int getEditingRow() {
		return (this.getColumnClass(this.editingColumn) == TreeTableModel.class) ? -1 : this.editingRow;
	}

	/**
	 * Returns the actual row that is editing as <code>getEditingRow</code> will
	 * always return -1.
	 */
	private int realEditingRow() {
		return this.editingRow;
	}

	/**
	 * This is overridden to invoke super's implementation, and then, if the
	 * receiver is editing a Tree column, the editor's bounds is reset. The reason
	 * we have to do this is because JTable doesn't think the table is being edited,
	 * as <code>getEditingRow</code> returns -1, and therefore doesn't automatically
	 * resize the editor for us.
	 */
	@Override
	public void sizeColumnsToFit(int resizingColumn) {
		super.sizeColumnsToFit(resizingColumn);
		if ((this.getEditingColumn() != -1) && (this.getColumnClass(this.editingColumn) == TreeTableModel.class)) {
			Rectangle cellRect = this.getCellRect(this.realEditingRow(), this.getEditingColumn(), false);
			Component component = this.getEditorComponent();
			component.setBounds(cellRect);
			component.validate();
		}
	}

	/**
	 * Overridden to pass the new rowHeight to the tree.
	 */
	@Override
	public void setRowHeight(int rowHeight) {
		super.setRowHeight(rowHeight);
		if ((this.tree != null) && (this.tree.getRowHeight() != rowHeight)) {
			this.tree.setRowHeight(this.getRowHeight());
		}
	}

	/**
	 * Returns the tree that is being shared between the model.
	 */
	public DnDTreeForTable getTree() {

		return this.tree;
	}

	/**
	 * Overridden to invoke repaint for the particular location if the column
	 * contains the tree. This is done as the tree editor does not fill the bounds
	 * of the cell, we need the renderer to paint the tree in the background, and
	 * then draw the editor over it.
	 */
	@Override
	public boolean editCellAt(int row, int column, EventObject e) {

		boolean retValue = super.editCellAt(row, column, e);
		if (retValue && (this.getColumnClass(column) == TreeTableModel.class)) {
			this.repaint(this.getCellRect(row, column, false));
			System.out.println("katla");
		}
		return retValue;
	}

	/**
	 * A TreeCellRenderer that displays a JTree.
	 */
	public class TreeTableCellRenderer extends DnDTreeForTable implements TableCellRenderer/* , DragGestureListener */ {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
		/** Last table/tree row asked to renderer. */
		protected int visibleRow;
		/**
		 * Border to draw around the tree, if this is non-null, it will be painted.
		 */
		protected Border highlightBorder;

		public TreeTableCellRenderer(TreeModel model) {
			// ORI super(model);
			super();
			this.setModel(model);

			// MP
			// this.setBorder(new EmptyBorder(5, 5, 5, 5));
			this.setCellRenderer(new ProjectTreeCellRenderer());

			this.setDragEnabled(true);
			this.setAutoscrolls(true);
			this.getSelectionModel().setSelectionMode(TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
			this.setDropMode(DropMode.ON_OR_INSERT);
			// this.setTransferHandler(new ProjectTreeTransferHandler());

			DragSource dragSource = DragSource.getDefaultDragSource();
			dragSource.createDefaultDragGestureRecognizer(this, DnDConstants.ACTION_COPY_OR_MOVE, this);

			this.addMouseListener(new java.awt.event.MouseAdapter() {
				@Override
				public void mouseClicked(java.awt.event.MouseEvent e) {
					System.out.println("tree clicked event received!");
				}
			});
		}

		// public void dragGestureRecognized(DragGestureEvent dge) {
		// System.out.println("DnDTree DragGestureEvent event received!");
		// }

		/**
		 * updateUI is overridden to set the colors of the Tree's renderer to match that
		 * of the table.
		 */
		@Override
		public void updateUI() {
			super.updateUI();
			// Make the tree's cell renderer use the table's cell selection
			// colors.
			TreeCellRenderer tcr = this.getCellRenderer();
			if (tcr instanceof DefaultTreeCellRenderer) {
				DefaultTreeCellRenderer dtcr = ((DefaultTreeCellRenderer) tcr);
				// For 1.1 uncomment this, 1.2 has a bug that will cause an
				// exception to be thrown if the border selection color is
				// null.
				// dtcr.setBorderSelectionColor(null);
				dtcr.setTextSelectionColor(UIManager.getColor("Table.selectionForeground"));
				dtcr.setBackgroundSelectionColor(UIManager.getColor("Table.selectionBackground"));
			}
		}

		/**
		 * Sets the row height of the tree, and forwards the row height to the table.
		 */
		@Override
		public void setRowHeight(int rowHeight) {
			if (rowHeight > 0) {
				super.setRowHeight(rowHeight);
				if ((ProTreeTable.this != null) && (ProTreeTable.this.getRowHeight() != rowHeight)) {
					ProTreeTable.this.setRowHeight(this.getRowHeight());
				}
			}
		}

		/**
		 * This is overridden to set the height to match that of the JTable.
		 */
		@Override
		public void setBounds(int x, int y, int w, int h) {
			super.setBounds(x, 0, w, ProTreeTable.this.getHeight());
		}

		/**
		 * Sublcassed to translate the graphics such that the last visible row will be
		 * drawn at 0,0.
		 */
		@Override
		public void paint(Graphics g) {
			g.translate(0, -this.visibleRow * this.getRowHeight());
			super.paint(g);
			// Draw the Table border if we have focus.
			if (this.highlightBorder != null) {
				this.highlightBorder.paintBorder(this, g, 0, this.visibleRow * this.getRowHeight(), this.getWidth(),
						this.getRowHeight());
			}
		}

		// /**
		// * TreeCellRenderer method. Overridden to update the visible row.
		// */
		// public Component getTableCellRendererComponent_ORI(JTable table,
		// Object value, boolean isSelected, boolean hasFocus, int row,
		// int column) {
		// Color background;
		// Color foreground;
		//
		// if (isSelected) {
		// background = table.getSelectionBackground();
		// foreground = table.getSelectionForeground();
		// } else {
		// background = table.getBackground();
		// foreground = table.getForeground();
		// }
		// highlightBorder = null;
		// if (realEditingRow() == row && getEditingColumn() == column) {
		// background = UIManager.getColor("Table.focusCellBackground");
		// foreground = UIManager.getColor("Table.focusCellForeground");
		// } else if (hasFocus) {
		// highlightBorder = UIManager
		// .getBorder("Table.focusCellHighlightBorder");
		// if (isCellEditable(row, column)) {
		// background = UIManager
		// .getColor("Table.focusCellBackground");
		// foreground = UIManager
		// .getColor("Table.focusCellForeground");
		// }
		// }
		//
		// visibleRow = row;
		// setBackground(background);
		//
		// TreeCellRenderer tcr = getCellRenderer();
		// if (tcr instanceof DefaultTreeCellRenderer) {
		// DefaultTreeCellRenderer dtcr = ((DefaultTreeCellRenderer) tcr);
		// if (isSelected) {
		// dtcr.setTextSelectionColor(foreground);
		// dtcr.setBackgroundSelectionColor(background);
		// } else {
		// dtcr.setTextNonSelectionColor(foreground);
		// dtcr.setBackgroundNonSelectionColor(background);
		// }
		// }
		// return this;
		// }
		// }
		//

		/**
		 * TreeCellRenderer method. Overridden to update the visible row.
		 */
		@Override
		public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus,
				int row, int column) {
			Color background;
			Color foreground;

			if (isSelected) {
				background = table.getSelectionBackground();
				foreground = table.getSelectionForeground();
			} else {
				background = table.getBackground();
				foreground = table.getForeground();
			}
			this.highlightBorder = null;
			// TODO colors

			if ((ProTreeTable.this.realEditingRow() == row) && (ProTreeTable.this.getEditingColumn() == column)) {
				background = UIManager.getColor("Table.focusCellBackground");
				foreground = UIManager.getColor("Table.focusCellForeground");
			} else if (hasFocus) {
				this.highlightBorder = UIManager.getBorder("Table.focusCellHighlightBorder");
				// if (ProTreeTable.this.isCellEditable(row, column)) {
				// background = UIManager
				// .getColor("Table.focusCellBackground");
				// foreground = UIManager
				// .getColor("Table.focusCellForeground");
				// }
			}

			this.visibleRow = row;
			this.setBackground(background);

			TreeCellRenderer tcr = this.getCellRenderer();
			if (tcr instanceof DefaultTreeCellRenderer) {
				DefaultTreeCellRenderer dtcr = ((DefaultTreeCellRenderer) tcr);
				if (isSelected) {
					dtcr.setTextSelectionColor(foreground);
					dtcr.setBackgroundSelectionColor(background);
				} else {
					dtcr.setTextNonSelectionColor(foreground);
					dtcr.setBackgroundNonSelectionColor(background);
				}
			}

			return this;
		}
	}

	/**
	 * An editor that can be used to edit the tree column. This extends
	 * DefaultCellEditor and uses a JTextField (actually, TreeTableTextField) to
	 * perform the actual editing.
	 * <p>
	 * To support editing of the tree column we can not make the tree editable. The
	 * reason this doesn't work is that you can not use the same component for
	 * editing and renderering. The table may have the need to paint cells, while a
	 * cell is being edited. If the same component were used for the rendering and
	 * editing the component would be moved around, and the contents would change.
	 * When editing, this is undesirable, the contents of the text field must stay
	 * the same, including the caret blinking, and selections persisting. For this
	 * reason the editing is done via a TableCellEditor.
	 * <p>
	 * Another interesting thing to be aware of is how tree positions its render and
	 * editor. The render/editor is responsible for drawing the icon indicating the
	 * type of node (leaf, branch...). The tree is responsible for drawing any other
	 * indicators, perhaps an additional +/- sign, or lines connecting the various
	 * nodes. So, the renderer is positioned based on depth. On the other hand,
	 * table always makes its editor fill the contents of the cell. To get the
	 * allusion that the table cell editor is part of the tree, we don't want the
	 * table cell editor to fill the cell bounds. We want it to be placed in the
	 * same manner as tree places it editor, and have table message the tree to
	 * paint any decorations the tree wants. Then, we would only have to worry about
	 * the editing part. The approach taken here is to determine where tree would
	 * place the editor, and to override the <code>reshape</code> method in the
	 * JTextField component to nudge the textfield to the location tree would place
	 * it. Since JTreeTable will paint the tree behind the editor everything should
	 * just work. So, that is what we are doing here. Determining of the icon
	 * position will only work if the TreeCellRenderer is an instance of
	 * DefaultTreeCellRenderer. If you need custom TreeCellRenderers, that don't
	 * descend from DefaultTreeCellRenderer, and you want to support editing in
	 * JTreeTable, you will have to do something similiar.
	 */
	public class TreeTableCellEditor extends DefaultCellEditor {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;

		public TreeTableCellEditor() {
			super(new TreeTableTextField());
		}

		/**
		 * Overridden to determine an offset that tree would place the editor at. The
		 * offset is determined from the <code>getRowBounds</code> JTree method, and
		 * additionally from the icon DefaultTreeCellRenderer will use.
		 * <p>
		 * The offset is then set on the TreeTableTextField component created in the
		 * constructor, and returned.
		 */
		@Override
		public Component getTableCellEditorComponent(JTable table, Object value, boolean isSelected, int r, int c) {

			Component component = super.getTableCellEditorComponent(table, value, isSelected, r, c);

			JTree t = ProTreeTable.this.getTree();
			boolean rv = t.isRootVisible();
			int offsetRow = rv ? r : r - 1;
			Rectangle bounds = t.getRowBounds(offsetRow);
			int offset = bounds.x;
			TreeCellRenderer tcr = t.getCellRenderer();
			if (tcr instanceof DefaultTreeCellRenderer) {
				Object node = t.getPathForRow(offsetRow).getLastPathComponent();
				Icon icon;
				if (t.getModel().isLeaf(node)) {
					icon = ((DefaultTreeCellRenderer) tcr).getLeafIcon();
				} else if (ProTreeTable.this.tree.isExpanded(offsetRow)) {
					icon = ((DefaultTreeCellRenderer) tcr).getOpenIcon();
				} else {
					icon = ((DefaultTreeCellRenderer) tcr).getClosedIcon();
				}
				if (icon != null) {
					offset += ((DefaultTreeCellRenderer) tcr).getIconTextGap() + icon.getIconWidth();
				}
			}
			((TreeTableTextField) this.getComponent()).offset = offset;

			System.out.println("mp edit (r/c) " + r + " " + c);

			return component;
		}

		/*
		 * public boolean shouldSelectCell(EventObject anEvent) {
		 * 
		 * return true; }
		 */
		/**
		 * This is overridden to forward the event to the tree. This will return true if
		 * the click count &gt;= 3, or the event is null.
		 */
		@Override
		public boolean isCellEditable(EventObject e) {
			if (e instanceof MouseEvent) {
				MouseEvent me = (MouseEvent) e;
				// If the modifiers are not 0 (or the left mouse button),
				// tree may try and toggle the selection, and table
				// will then try and toggle, resulting in the
				// selection remaining the same. To avoid this, we
				// only dispatch when the modifiers are 0 (or the left mouse
				// button).

				// System.out.println("Mouse: " + e);
				// ProTreeTable.this.tree.dispatchEvent((MouseEvent) e);

				if ((me.getModifiers() == 0) || (me.getModifiers() == InputEvent.BUTTON1_MASK)) {
					for (int counter = ProTreeTable.this.getColumnCount() - 1; counter >= 0; counter--) {
						if (ProTreeTable.this.getColumnClass(counter) == TreeTableModel.class) {
							MouseEvent newME = new MouseEvent(ProTreeTable.this.tree, me.getID(), me.getWhen(),
									me.getModifiers(), me.getX() - ProTreeTable.this.getCellRect(0, counter, true).x,
									me.getY(), me.getClickCount(), me.isPopupTrigger());
							ProTreeTable.this.tree.dispatchEvent(newME);
							break;
						}
					}
				}
				// if (me.getClickCount() >= 3) {
				// return true;
				// }
				return false;
			}
			if (e == null) {
				return true;
			}
			return false;
		}
	}

	/**
	 * Component used by TreeTableCellEditor. The only thing this does is to
	 * override the <code>reshape</code> method, and to ALWAYS make the x location
	 * be <code>offset</code>.
	 */
	static class TreeTableTextField extends JTextField {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
		public int offset;

		@Override
		@SuppressWarnings("deprecation")
		public void reshape(int x, int y, int w, int h) {
			int newX = Math.max(x, this.offset);
			super.reshape(newX, y, w - (newX - x), h);
		}
	}

	/**
	 * ListToTreeSelectionModelWrapper extends DefaultTreeSelectionModel to listen
	 * for changes in the ListSelectionModel it maintains. Once a change in the
	 * ListSelectionModel happens, the paths are updated in the
	 * DefaultTreeSelectionModel.
	 */
	class ListToTreeSelectionModelWrapper extends DefaultTreeSelectionModel {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
		/** Set to true when we are updating the ListSelectionModel. */
		protected boolean updatingListSelectionModel;

		public ListToTreeSelectionModelWrapper() {
			super();
			this.getListSelectionModel().addListSelectionListener(this.createListSelectionListener());
		}

		/**
		 * Returns the list selection model. ListToTreeSelectionModelWrapper listens for
		 * changes to this model and updates the selected paths accordingly.
		 */
		ListSelectionModel getListSelectionModel() {
			return this.listSelectionModel;
		}

		/**
		 * This is overridden to set <code>updatingListSelectionModel</code> and message
		 * super. This is the only place DefaultTreeSelectionModel alters the
		 * ListSelectionModel.
		 */
		@Override
		public void resetRowSelection() {
			if (!this.updatingListSelectionModel) {
				this.updatingListSelectionModel = true;
				try {
					super.resetRowSelection();
				} finally {
					this.updatingListSelectionModel = false;
				}
			}
			// Notice how we don't message super if
			// updatingListSelectionModel is true. If
			// updatingListSelectionModel is true, it implies the
			// ListSelectionModel has already been updated and the
			// paths are the only thing that needs to be updated.
		}

		/**
		 * Creates and returns an instance of ListSelectionHandler.
		 */
		protected ListSelectionListener createListSelectionListener() {
			return new ListSelectionHandler();
		}

		/**
		 * If <code>updatingListSelectionModel</code> is false, this will reset the
		 * selected paths from the selected rows in the list selection model.
		 */
		protected void updateSelectedPathsFromSelectedRows() {
			if (!this.updatingListSelectionModel) {
				this.updatingListSelectionModel = true;
				try {
					// This is way expensive, ListSelectionModel needs an
					// enumerator for iterating.
					int min = this.listSelectionModel.getMinSelectionIndex();
					int max = this.listSelectionModel.getMaxSelectionIndex();

					this.clearSelection();
					if ((min != -1) && (max != -1)) {
						for (int counter = min; counter <= max; counter++) {
							if (this.listSelectionModel.isSelectedIndex(counter)) {
								TreePath selPath = ProTreeTable.this.tree.getPathForRow(counter);

								if (selPath != null) {
									this.addSelectionPath(selPath);
								}
							}
						}
					}
				} finally {
					this.updatingListSelectionModel = false;
				}
			}
		}

		/**
		 * Class responsible for calling updateSelectedPathsFromSelectedRows when the
		 * selection of the list changse.
		 */
		class ListSelectionHandler implements ListSelectionListener {
			@Override
			public void valueChanged(ListSelectionEvent e) {
				ListToTreeSelectionModelWrapper.this.updateSelectedPathsFromSelectedRows();
			}
		}

	}

	// Interface DragGestureListener

	@Override
	public void dragGestureRecognized(DragGestureEvent dge) {
		this.tree.dragGestureRecognized(dge);
	}

	// Interface DragSourceListener

	@Override
	public void dragDropEnd(DragSourceDropEvent dsde) {
		this.tree.dragDropEnd(dsde);
	}

	@Override
	public void dragEnter(DragSourceDragEvent dsde) {
		this.tree.dragEnter(dsde);
	}

	@Override
	public void dragExit(DragSourceEvent dse) {
		this.tree.dragExit(dse);
	}

	@Override
	public void dragOver(DragSourceDragEvent dsde) {
	}

	@Override
	public void dropActionChanged(DragSourceDragEvent dsde) {
		this.tree.dropActionChanged(dsde);
	}

	public void setSelectedIndex(int index) {

		this._selectedIndex = index;
		this._idxMgr.setSelectedIndex(index);
		// _gm.fireProjectChanged();
	}

	public int getSelectedIndex() {

		return this._selectedIndex;
	}

	public void rubCueLine() {

		this.paintImmediately(this._raCueGeometry.getBounds());
	}

	public void setCueGeometry(Rectangle2D cueGeometry, Color geometryColor) {

		Graphics2D g2 = (Graphics2D) this.getGraphics();
		this._raCueGeometry = cueGeometry;
		g2.setColor(geometryColor);
		g2.fill(this._raCueGeometry);
		this._raGhost = this._raGhost.createUnion(this._raCueGeometry);
	}

	public void repaintCueGeometry() {

		this.repaint(this._raGhost.getBounds()); // clear cue image
	}

	@Override
	public void editingStopped(ChangeEvent e) {

		super.editingStopped(e);

		// int _selectedColumn = this.getSelectedColumn();
		// int _selectedRow = this.getSelectedRow();

		// System.out.println("in editingStopped (c/r): " + _selectedColumn +
		// " "
		// + _selectedRow);
		// this._gm.showTitleMessage("editingStopped");

		// SwingUtilities.invokeLater(new Runnable() {
		// public void run() {
		// // getModelAdapter().fireTableDataChanged();
		// if (tree != null) {
		// tree.updateUI();
		// }
		// }
		// });
		// super.editingStopped(e);
	}

	@Override
	public void editingCanceled(ChangeEvent e) {

		super.editingCanceled(e);
		// System.out.println("editingCanceled");
	}

	public void setProParameterModel(ProParameterModel proParameterModel) {

		this._proParameterModel = proParameterModel;
	}

	public ProParameterModel getProParameterModel() {

		return this._proParameterModel;
	}

	public ProTreeTableModelAdapter getModelAdapter() {

		return this._modelAdapter;
	}

	@Override
	public DropTarget getDropTarget() {

		return this._dropTarget;
	}

	public ProTreeTableIndexViewManager getIndexViewManager() {

		return this._idxMgr;
	}
}
