/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9893 $
 * $Date: 2020-06-25 13:21:01 +0200 (Thu, 25 Jun 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.rmi.RemoteException;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.magopt.structure.CollectorStructure.CollectorStructureException;
import at.lcm.bu21.magopt.structure.client.CParamCollector;
import at.lcm.bu21.magopt.structure.rmi.RParamCollector;

/**
 * Parameter for data-representation
 * 
 * @author Werner Koppelsttter
 */
public class ParamCollector extends ParamContainer implements IParamCollector, Function {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Dependency of Parameter */
	private Depend depend = new Depend(this);

	@Override
	public List<String> listTabNames() {
		List<String> list = super.listTabNames();
		return list;
	}

	/**
	 * Constructor for loading a xml-file
	 * 
	 * @throws Exception
	 */
	ParamCollector() throws Exception {
		this("XML_Load", "XML load constructor");
	}

	/**
	 * Constructor with name
	 * 
	 * @param name
	 *            Name of Parameter
	 * @throws Exception
	 */
	public ParamCollector(String name) throws Exception {
		this(name, name);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name
	 *            Name of Parameter
	 * @param description
	 *            Description of Parameter
	 * @throws Exception
	 */
	public ParamCollector(String name, String description) throws Exception {
		super(name, description);
		this.setAccess(Access.PROTECTED);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ParamCollector other = (ParamCollector) obj;
		if (this.depend == null) {
			if (other.depend != null) {
				return false;
			}
		} else if (!this.depend.equals(other.depend)) {
			return false;
		}
		return true;
	}

	@Override
	public ParamCollector deepClone() throws CloneException {
		return (ParamCollector) Clone.deepClone(this);
	}

	@Override
	public ParamContainer convertToParamContainer() {
		Parameter[] list;
		ParamContainer obj = super.convertToParamContainer();

		// clear locks
		list = obj.getParameters(ParamField.class);
		for (Parameter param : list) {
			((ParamField) param).setLock(false);
		}

		this.refresh();
		obj.refresh();

		return obj;
	}

	@Override
	void cloneParam(Parameter clone) throws CloneException, InstantiationException, IllegalAccessException {

		super.cloneParam(clone);

		if (clone instanceof ParamCollector) {
			ParamCollector other = (ParamCollector) clone;

			this.depend.cloneDepend(other.depend);
		}
	}

	@Override
	public Depend getDepend() {
		return this.depend;
	}

	@Override
	public void refreshDependency() {
		DependItem[] depend;
		PortDesc[] desc;
		String[][] assign;

		depend = this.listImports();
		desc = new PortDesc[depend.length];
		assign = new String[depend.length][2];
		for (int i = 0; i < depend.length; i++) {
			desc[i] = depend[i].getVarDesc();
			assign[i][0] = depend[i].getVarName();
			assign[i][1] = depend[i].getParam();
		}
		this.getDepend().setImportVars(desc);
		this.getDepend().setImportAssignments(assign);
	}

	/**
	 * List all sub-parameter as imports
	 * 
	 * @return List of imports
	 */
	protected DependItem[] listImports() {
		LinkedList<DependItem> depend;

		depend = new LinkedList<DependItem>();
		for (Parameter p : this.getParam()) {
			if ((p instanceof ParamCollector) || !(p instanceof ParamContainer)) {
				depend.add(new DependItem(new PortDesc(p.getName(), "Sub parameter", "", false), p.getIdentifier()));
			}
		}

		return depend.toArray(new DependItem[0]);
	}

	@Override
	public void refreshDepend(boolean dirty, boolean load) {
		this.depend.refreshDepend(dirty, load);
	}

	@Override
	public boolean getUsed() {
		return (this.getUsedBy().length != 0);
	}

	@Override
	public TStruct getData() {
		TStruct struct;
		CollectorStructure collect, sub;

		struct = new TStruct();
		collect = new CollectorStructure(struct);

		for (DependItem d : this.getDepend().getImports()) {
			if (d.getHandle() != null) {
				try {
					if (d.getHandle() instanceof ParamCollector) {
						collect.addData(null, d.getHandle().getName(),
								new CollectorStructure((TStruct) d.getHandle().getData()));
					} else {
						sub = collect.addData(null, d.getHandle().getName());
						sub.setDescription(null, d.getHandle().getDescription());
						if (d.getHandle() instanceof IParamField) {
							try {
								sub.setUnit(null, ((IParamField) d.getHandle()).getUnit());
							} catch (RemoteException e) {
								this.warning(e.getMessage(), e.getStackTrace());
							}
						}
						sub.setValue(null, d.getHandle().getData());
					}
				} catch (CollectorStructureException e) {
					this.warning(e.getMessage(), e.getStackTrace());
				}
			}
		}

		return struct;
	}

	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();
		str.append("ParamCollector\n");
		str.append("\n\tName           : ");
		str.append(this.getName());
		str.append("\n\tDescription    : ");
		str.append(this.getDescription());
		str.append("\n\tData           : ");
		str.append(this.getData());
		return str.toString();
	}

	@Override
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("ParamContainer")) {
				super.xml2obj(xmlReader);
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();

		this.getDepend().replaceImports(this.listImports());
	}

	@Override
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("ParamCollector");
		xmlWriter.writeAttribute("object", this.getClass().getSimpleName());
		xmlWriter.writeAttribute("name", this.getName());

		// super class
		super.obj2xml(xmlWriter);

		xmlWriter.writeEndElement();
	}

	@Override
	public boolean getUpdateRequired() {
		return this.getModified();
	}

	@Override
	public boolean getUpdateRequired(int index) {
		return this.getModified(index);
	}

	@Override
	public void setUpdateRequired(boolean update) {
		this.setUpdateRequired(this.getIndex(), update);

	}

	@Override
	public void setUpdateRequired(int index, boolean update) {
		if (update) {
			this.setModified(index, true);
		}
	}

	@Override
	public void reset() {
	}

	@Override
	public boolean checkDocFiles(int index) {
		return false;
	}

	@Override
	public int update(Map<String, Object> args) throws Exception {
		return 0;
	}

	@Override
	public IParamCollector getRMIImpl() throws RemoteException {
		return new RParamCollector(this);
	}

	@Override
	public IParamCollector getClientImpl() {
		return new CParamCollector(this);
	}
}
