/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9817 $
 * $Date: 2020-05-05 21:02:41 +0200 (Tue, 05 May 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.io.File;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.general.utils.MathUtils;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.structure.client.CParamMaterial;
import at.lcm.bu21.magopt.structure.rmi.RParamMaterial;
import at.lcm.bu21.magopt.utils.MagOptConfig;
import magopt.structure.database.DBItem;
import magopt.structure.database.DatabaseConnection;
import magopt.structure.database.FileConnection;
import magopt.structure.database.MaterialConnection;
import magopt.structure.database.MaterialDiagram;
import magopt.structure.database.MaterialProperty;

/**
 * Parameter for data-representation
 * 
 * @author Werner Koppelsttter
 */
public class ParamMaterial extends ParamContainer implements IParamMaterial, Function {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** database user */
	private static String dbUser = "magopt_basic";
	/** database password */
	private static String dbPwd = "magopt";
	/** database file */
	private static String dbPath = null;

	static {
		ParamMaterial.resetDbUser();
		String path = MagOptConfig.getConfig("DatabaseFile");
		if (path != null) {
			File file = new File(path);
			if (file.exists() && file.isFile()) {
				ParamMaterial.setDbPath(file.getPath());
			}
		}
	}

	/** Selected dataset */
	private long datasetID = -1;

	/** Name of material-class */
	private String materialClass = "---";

	/** Name of material */
	private String material = "---";

	/** Name of dataset */
	private String dataset = "---";

	/** Dependency of Parameter */
	private Depend depend = new Depend(this);

	/**
	 * Is offline database is selected
	 * 
	 * @return Is offline
	 */
	public static boolean isOffline() {
		return (ParamMaterial.dbPath != null) && (ParamMaterial.dbPath.length() != 0)
				&& new File(ParamMaterial.dbPath).exists();
	}

	/**
	 * Get database user
	 * 
	 * @return User name
	 */
	public static String getDbUser() {
		return ParamMaterial.dbUser;
	}

	/**
	 * Set database user
	 * 
	 * @param user
	 *            User name
	 * @param pwd
	 *            Password
	 */
	public static void setDbUser(String user, String pwd) {
		ParamMaterial.dbUser = user;
		ParamMaterial.dbPwd = pwd;
	}

	/**
	 * Reset database-user to default
	 */
	public static void resetDbUser() {
		ParamMaterial.setDbUser("magopt_basic", "magopt");
	}

	/**
	 * Set path to database-file (offline)
	 * 
	 * @param path
	 *            Database file
	 */
	public static void setDbPath(String path) {
		ParamMaterial.dbPath = path;
	}

	/**
	 * Get database connection
	 * 
	 * @return Connection to database
	 * @throws Exception
	 */
	public static MaterialConnection getConnection() throws Exception {
		File file;

		if ((ParamMaterial.dbPath != null) && (ParamMaterial.dbPath.length() != 0)) {
			file = new File(ParamMaterial.dbPath);
			if (!file.exists()) {
				throw new Exception("Database file doesn't exist");
			} else {
				return new FileConnection(file);
			}
		} else {
			return new DatabaseConnection(ParamMaterial.dbUser, ParamMaterial.dbPwd);
		}
	}

	@Override
	public List<String> listTabNames() {
		List<String> list = super.listTabNames();
		list.set(0, "Material Info");
		return list;
	}

	/**
	 * Constructor for loading a xml-file
	 * 
	 * @throws Exception
	 */
	ParamMaterial() throws Exception {
		this("XML_Load", "XML load constructor");
	}

	/**
	 * Constructor with name
	 * 
	 * @param name
	 *            Name of Parameter
	 * @throws Exception
	 */
	public ParamMaterial(String name) throws Exception {
		this(name, name);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name
	 *            Name of Parameter
	 * @param description
	 *            Description of Parameter
	 * @throws Exception
	 */
	public ParamMaterial(String name, String description) throws Exception {
		super(name, description);
		this.setAccess(Access.PROTECTED);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ParamMaterial other = (ParamMaterial) obj;
		if (this.datasetID != other.datasetID) {
			return false;
		}
		if (this.dataset == null) {
			if (other.dataset != null) {
				return false;
			}
		} else if (!this.dataset.equals(other.dataset)) {
			return false;
		}
		if (this.material == null) {
			if (other.material != null) {
				return false;
			}
		} else if (!this.material.equals(other.material)) {
			return false;
		}
		if (this.materialClass == null) {
			if (other.materialClass != null) {
				return false;
			}
		} else if (!this.materialClass.equals(other.materialClass)) {
			return false;
		}
		if (this.depend == null) {
			if (other.depend != null) {
				return false;
			}
		} else if (!this.depend.equals(other.depend)) {
			return false;
		}
		return true;
	}

	@Override
	public ParamMaterial deepClone() throws CloneException {
		return (ParamMaterial) Clone.deepClone(this);
	}

	@Override
	public ParamContainer convertToParamContainer() {
		Parameter[] list;
		ParamContainer obj = super.convertToParamContainer();

		// clear locks
		list = obj.getParameters(ParamField.class);
		for (Parameter param : list) {
			((ParamField) param).setLock(false);
		}

		this.refresh();
		obj.refresh();

		return obj;
	}

	@Override
	void cloneParam(Parameter clone) throws CloneException, InstantiationException, IllegalAccessException {

		super.cloneParam(clone);

		if (clone instanceof ParamMaterial) {
			ParamMaterial other = (ParamMaterial) clone;
			other.datasetID = this.datasetID;
			other.materialClass = this.materialClass;
			other.material = this.material;
			other.dataset = this.dataset;

			this.depend.cloneDepend(other.depend);
		}
	}

	@Override
	public Depend getDepend() {
		return this.depend;
	}

	@Override
	public void refreshDependency() {
		DependItem[] depend;
		PortDesc[] desc;
		String[][] assign;

		depend = this.listImports();
		desc = new PortDesc[depend.length];
		assign = new String[depend.length][2];
		for (int i = 0; i < depend.length; i++) {
			desc[i] = depend[i].getVarDesc();
			assign[i][0] = depend[i].getVarName();
			assign[i][1] = depend[i].getParam();
		}
		this.getDepend().setImportVars(desc);
		this.getDepend().setImportAssignments(assign);
	}

	/**
	 * List all sub-parameter as imports
	 * 
	 * @return List of imports
	 */
	protected DependItem[] listImports() {
		DependItem[] depend;
		Parameter[] param;

		param = this.getParameters(Parameter.class);
		depend = new DependItem[param.length - 1];
		for (int i = 0; i < (param.length - 1); i++) {
			depend[i] = new DependItem(new PortDesc("p" + i, "Sub parameter", "", false), param[i + 1].getIdentifier());
		}

		return depend;
	}

	@Override
	public void refreshDepend(boolean dirty, boolean load) {
		this.depend.refreshDepend(dirty, load);
	}

	@Override
	public boolean getUsed() {
		return (this.getUsedBy().length != 0);
	}

	@Override
	public String getMaterialClassName() {
		return this.materialClass;
	}

	@Override
	public String getMaterialName() {
		return this.material;
	}

	@Override
	public String getDatasetName() {
		return this.dataset;
	}

	@Override
	public long getDatasetID() {
		return this.datasetID;
	}

	@Override
	public void setDatasetID(DBItem dataset, MaterialConnection conn) {
		this.setDatasetID(dataset.getID(), conn);
	}

	@Override
	public void setDatasetID(long datasetID, MaterialConnection conn) {
		Parameter[] list;

		if (conn != null) {
			for (int i = 0; i < this.getLength(); i++) {
				this.setModified(i, true);
			}

			try {
				// remove children
				this.changeProperties(conn, false);
				this.changeDiagrams(conn, false);

				// clear locks
				list = this.getParameters(ParamField.class);
				for (Parameter param : list) {
					((ParamField) param).setLock(false);
				}

				// change Material
				this.datasetID = datasetID;
				this.changeValue(conn);

				// remove children
				this.changeProperties(conn, true);
				this.changeDiagrams(conn, true);
			} catch (Exception ex) {
				this.error(ex.getMessage(), ex.getStackTrace());
			}
		}
	}

	/**
	 * Change values
	 * 
	 * @param conn
	 *            Connection to database
	 * @throws Exception
	 */
	protected void changeValue(MaterialConnection conn) throws Exception {
		ParameterPath param;
		ParamField field;
		DBItem[] items;

		items = conn.getMaterialItems(this.getDatasetID());

		// Type
		this.materialClass = items[0].getName();
		param = this.getParamByUniqueIdentifier(this.getName() + ".Type", true);
		if ((param != null) && (param.getParameter() != null)) {
			param.getParameter().removeParam();
		}
		field = new ParamField("Type", "Type of material");
		this.addParam(field);
		field.setConstant(true);
		field.setType("attribute");
		field.setTexString("Type");
		field.setData(new TString(this.materialClass));
		field.setLock(true);

		// Name
		this.material = items[1].getName();
		param = this.getParamByUniqueIdentifier(this.getName() + ".Name", true);
		if ((param != null) && (param.getParameter() != null)) {
			param.getParameter().removeParam();
		}
		field = new ParamField("Name", "Name of material");
		this.addParam(field);
		field.setConstant(true);
		field.setType("attribute");
		field.setTexString("Name");
		field.setData(new TString(this.material));
		field.setLock(true);

		// Dataset
		this.dataset = items[2].getName();
		param = this.getParamByUniqueIdentifier(this.getName() + ".Dataset", true);
		if ((param != null) && (param.getParameter() != null)) {
			param.getParameter().removeParam();
		}
		field = new ParamField("Dataset", "Dataset of material");
		this.addParam(field);
		field.setConstant(true);
		field.setType("attribute");
		field.setTexString("Dataset");
		field.setData(new TString(this.dataset));
		field.setLock(true);
	}

	/**
	 * Change material property
	 * 
	 * @param conn
	 *            Connection to database
	 * 
	 * @param add
	 *            Add or remove properties
	 */
	protected void changeProperties(MaterialConnection conn, boolean add) {
		ParamField field;
		ParameterPath param;
		LinkedList<MaterialProperty<Double>> list;

		list = conn.getProperties(this.getDatasetID(), ((Project) this.getRoot()).getEventLog());
		for (MaterialProperty<Double> prop : list) {
			try {
				param = this.getParamByUniqueIdentifier(this.getName() + "." + prop.getName(), true);
				if ((param != null) && (param.getParameter() != null)) {
					param.getParameter().removeParam();
				}
				if (add) {
					field = new ParamField(prop.getName());
					this.addParam(field);
					field.setConstant(true);
					field.setDescription(prop.getDescription());
					field.setType(prop.getGroup());
					field.setUnit(prop.getUnit());
					field.setData(new TNumeric(prop.getValue()));
					field.setTexString(prop.getTex());
					field.setLock(prop.getLocked());
				}
			} catch (Exception ex) {
				this.error(ex.getMessage(), ex.getStackTrace());
			}
		}
	}

	/**
	 * Change material diagrams
	 * 
	 * @param conn
	 *            Connection to database
	 * 
	 * @param add
	 *            Add or remove
	 */
	protected void changeDiagrams(MaterialConnection conn, boolean add) {
		ParamContainer container, curve;
		ParameterPath param;
		ParamField field;
		LinkedList<MaterialDiagram> list;
		TNumeric values;
		String curveName;

		list = conn.getDiagrams(this.getDatasetID(), ((Project) this.getRoot()).getEventLog());
		for (MaterialDiagram diag : list) {
			try {
				if (add) {
					// create diagram container if not exists
					container = null;
					param = this.getParamByUniqueIdentifier(this.getName() + "." + diag.getName(), true);
					if ((param != null) && (param.getParameter() != null)) {
						if (param.getParameter() instanceof ParamContainer) {
							container = (ParamContainer) param.getParameter();
						} else {
							param.getParameter().removeParam();
						}
					}
					if (container == null) {
						container = new ParamContainer(diag.getName());
						this.addParam(container);
					}

					// create curve container if needed and not exists
					curveName = "";
					for (MaterialProperty<Double[]> axes : diag.getAxis()) {
						if (axes.getFixed()) {
							curveName += "_" + axes.getName() + "_" + axes.getValue()[0].toString().replace('.', '_');
						}
					}
					if (curveName.length() != 0) {
						curveName = curveName.substring(1);
						curve = null;
						param = container.getParamByUniqueIdentifier(container.getName() + "." + curveName, true);
						if ((param != null) && (param.getParameter() != null)) {
							if (param.getParameter() instanceof ParamContainer) {
								curve = (ParamContainer) param.getParameter();
							} else {
								param.getParameter().removeParam();
							}
						}
						if (curve == null) {
							curve = new ParamContainer(curveName);
							container.addParam(curve);
						}
					} else {
						curve = container;
					}
					curve.setDescription(diag.getDescription());

					// add axes
					for (MaterialProperty<Double[]> axes : diag.getAxis()) {
						try {
							param = curve.getParamByUniqueIdentifier(curve.getName() + "." + axes.getName(), true);
							if ((param != null) && (param.getParameter() != null)) {
								param.getParameter().removeParam();
							}

							field = new ParamField(axes.getName());
							curve.addParam(field);
							field.setConstant(true);
							field.setDescription(axes.getDescription());
							field.setType(axes.getGroup());
							field.setUnit(axes.getUnit());
							field.setTexString(axes.getTex());
							if (axes.getFixed()) {
								values = new TNumeric(axes.getValue()[0]);
							} else {
								values = new TNumeric(axes.getValue());
							}
							field.setData(values);
							field.setLock(axes.getLocked());
						} catch (Exception ex) {
							this.error(ex.getMessage(), ex.getStackTrace());
						}
					}
				} else {
					param = this.getParamByUniqueIdentifier(this.getName() + "." + diag.getName(), true);
					if ((param != null) && (param.getParameter() != null)
							&& (param.getParameter() instanceof ParamContainer)) {
						container = (ParamContainer) param.getParameter();

						// get curve container to remove
						curveName = "";
						for (MaterialProperty<Double[]> axes : diag.getAxis()) {
							if (axes.getFixed()) {
								curveName += "_" + axes.getName() + "_"
										+ axes.getValue()[0].toString().replace('.', '_');
							}
						}
						if (curveName.length() != 0) {
							curveName = curveName.substring(1);
							curve = null;
							param = container.getParamByUniqueIdentifier(container.getName() + "." + curveName, true);
							if ((param != null) && (param.getParameter() != null)
									&& (param.getParameter() instanceof ParamContainer)) {
								curve = (ParamContainer) param.getParameter();
							}
						} else {
							curve = container;
						}

						// remove curve
						if (curve != null) {
							// remove axes
							for (MaterialProperty<Double[]> axes : diag.getAxis()) {
								param = curve.getParamByUniqueIdentifier(curve.getName() + "." + axes.getName(), true);
								if ((param != null) && (param.getParameter() != null)) {
									param.getParameter().removeParam();
								}
							}

							// remove empty curve
							if (curve.getParam().length == 0) {
								curve.removeParam();
							}
						}

						// remove empty container
						if (container.getParam().length == 0) {
							container.removeParam();
						}
					}
				}
			} catch (Exception ex) {
				this.error(ex.getMessage(), ex.getStackTrace());
			}
		}
	}

	@Override
	public TStruct getData() {
		return this.getDataStructRec();
	}

	@Override
	public TStruct getDataStructRec() {
		TStruct struct;

		struct = new TStruct();
		for (Parameter param : this.getParam()) {
			struct.put(param.getName(), param.getDataStructRec());
		}

		return struct;
	}

	// @Override
	// public TStruct getDataStruct(){
	// TStruct struct;
	//
	// struct = new TStruct();
	// for (Parameter param : this.getParam()) {
	// struct.put(param.getName(), param.getDataStruct());
	// }
	//
	// return struct;
	// }

//	protected TStruct getDataStruct(ParamContainer container) {
//		TStruct struct, item;
//
//		struct = new TStruct();
//		for (Parameter param : container.getParam()) {
//			if (param instanceof ParamField) {
//				ParamField field = (ParamField) param;
//				item = new TStruct();
//				item.put("value__", field.getData());
//				item.put("unit__", new TString(field.getUnit()));
//				struct.put(param.getName(), item);
//			} else if (param instanceof ParamContainer) {
//				struct.put(param.getName(), this.getDataStruct((ParamContainer) param));
//			} else {
//				struct.put(param.getName(), new TNull());
//			}
//		}
//		return struct;
//
//	}

	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();
		str.append("ParamMaterial\n");
		str.append("\n\tName           : ");
		str.append(this.getName());
		str.append("\n\tDescription    : ");
		str.append(this.getDescription());
		str.append("\n\tMaterial-class : ");
		str.append(this.materialClass);
		str.append("\n\tMaterial       : ");
		str.append(this.material);
		str.append("\n\tDataset        : ");
		str.append(this.dataset);
		str.append("\n\tData           : ");
		str.append(this.getData());
		return str.toString();
	}

	@Override
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("ParamContainer")) {
				super.xml2obj(xmlReader);
			} else if (element.equals("DSID")) {
				xmlReader.next();
				this.datasetID = Long.parseLong(xmlReader.getText());
				xmlReader.next();
				xmlReader.next();
			} else if (element.equals("MCName")) {
				xmlReader.next();
				this.materialClass = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.materialClass += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("MName")) {
				xmlReader.next();
				this.material = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.material += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("DSName")) {
				xmlReader.next();
				this.dataset = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.dataset += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();

		this.getDepend().replaceImports(this.listImports());
	}

	@Override
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("ParamMaterial");
		xmlWriter.writeAttribute("object", this.getClass().getSimpleName());
		xmlWriter.writeAttribute("name", this.getName());

		// super class
		super.obj2xml(xmlWriter);

		// Dataset-id
		xmlWriter.writeStartElement("DSID");
		xmlWriter.writeCharacters(Long.toString(this.datasetID));
		xmlWriter.writeEndElement();

		// Material-name
		xmlWriter.writeStartElement("MCName");
		if (this.materialClass != null) {
			xmlWriter.writeCharacters(this.materialClass);
		}
		xmlWriter.writeEndElement();

		// Material-name
		xmlWriter.writeStartElement("MName");
		if (this.material != null) {
			xmlWriter.writeCharacters(this.material);
		}
		xmlWriter.writeEndElement();

		// Material-name
		xmlWriter.writeStartElement("DSName");
		if (this.dataset != null) {
			xmlWriter.writeCharacters(this.dataset);
		}
		xmlWriter.writeEndElement();

		xmlWriter.writeEndElement();
	}

	@Override
	public boolean getUpdateRequired() {
		return this.getModified();
	}

	@Override
	public boolean getUpdateRequired(int index) {
		return this.getModified(index);
	}

	@Override
	public void setUpdateRequired(boolean update) {
		this.setUpdateRequired(this.getIndex(), update);

	}

	@Override
	public void setUpdateRequired(int index, boolean update) {
		if (update) {
			this.setModified(index, true);
		}
	}

	@Override
	public void reset() {
	}

	@Override
	public void generateDocFiles(File directory) throws Exception {
		File file;
		DocTexWriter content;

		try {
			// report datasheet
			file = this.getUniqueDocFileName(new File(directory, "Datasheet.tex"));
			content = new DocTexWriter(file);
			this.reportDatasheet(content);
			content.write();
			this.registerDocFile("Datasheet", file.getPath(), "Datasheet "
					+ ((TString) this.getParamByUniqueIdentifier(this.getName() + ".Name").getParameter().getData())
							.getString());
		} catch (Exception ex) {
			this.warning(this.getIdentifier(), ex.getMessage(), ex.getStackTrace());
		}
	}

	/**
	 * Report datasheet from material
	 * 
	 * @param content
	 *            Tex-writer
	 * @throws Exception
	 */
	protected void reportDatasheet(DocTexWriter content) throws Exception {

		// name, type and dataset
		content.add("\\begin{datasheet}{");
		content.addString(((TString) this.getParamByUniqueIdentifier(this.getName() + ".Name").getParameter().getData())
				.getString());
		content.add("}\n\n");

		content.add("\\vspace{3ex}\n");
		content.add(
				"\\fcolorbox{tablegray}{tablegray}{\\parbox{1\\textwidth}{\\color{white} \\textbf{Material Class}}}\\\\[1.5ex]\n");
		content.add("\\textsc{");
		content.addString(((TString) this.getParamByUniqueIdentifier(this.getName() + ".Type").getParameter().getData())
				.getString());
		content.add("}\n\n");

		content.add("\\vspace{3ex}\n");
		content.add(
				"\\fcolorbox{tablegray}{tablegray}{\\parbox{1\\textwidth}{\\color{white} \\textbf{Description}}}\\\\[1.5ex]\n");
		content.add("\\textsc{");
		content.addString(
				((TString) this.getParamByUniqueIdentifier(this.getName() + ".Dataset").getParameter().getData())
						.getString());
		content.add("}\n\n");

		// Property group
		this.reportProperties(content);

		// diagrams
		this.reportDiagrams(content);

		content.add("\\end{datasheet}");
	}

	/**
	 * Write properties of material to report
	 * 
	 * @param content
	 *            Tex-writer
	 */
	protected void reportProperties(DocTexWriter content) {
		Hashtable<String, ArrayList<ParamField>> properties;
		ArrayList<ParamField> list;
		ParamField field;
		String type;
		String[] groups;

		// group properties
		properties = new Hashtable<String, ArrayList<ParamField>>();
		for (Parameter p : this.getParam()) {
			if (p instanceof ParamField) {
				field = (ParamField) p;
				type = field.getType();
				list = properties.get(type);
				if (list == null) {
					list = new ArrayList<ParamField>();
					properties.put(type, list);
				}
				list.add(field);
			}
		}
		properties.remove("attribute");

		// write groups
		groups = properties.keySet().toArray(new String[0]);
		Arrays.sort(groups);
		for (String group : groups) {
			content.add("\\vspace{3ex}\n");

			content.add("\\ifdefined\\usehtml\n");
			content.add("\\begin{tabular}{m{5cm} >{\\raggedleft}m{2cm} m{2cm} m{5cm}}\n");
			content.add("\\else\n");
			content.add("\\begin{longtable}{m{5cm} >{\\raggedleft}m{2cm} m{2cm} m{5cm}}\n");
			content.add("\\fi\n\n");

			content.add("\\multicolumn{4}{>{\\columncolor{tablegray}\\color{white}}l}{\\textbf{");
			if (group.length() != 0) {
				content.addString(group + " ");
			}
			content.add("Properties");
			content.add("}}\\\\\n");
			content.add("\\multicolumn{1}{>{\\columncolor{tablegray}\\color{white}}l}{\\small \\emph{Property}} &\n");
			content.add("\\multicolumn{2}{>{\\columncolor{tablegray}\\color{white}}c}{\\small \\emph{Value/Unit}} &\n");
			content.add("\\multicolumn{1}{>{\\columncolor{tablegray}\\color{white}}l}{\\small \\emph{Comment}} \\\\\n");
			list = properties.get(group);
			for (ParamField f : list) {
				content.addString(f.getDescription());
				content.add(" &");
				content.addString(f.getData().toString());
				content.add(" &");
				content.addString(f.getUnit());
				content.add(" &");
				content.addString(f.getComment());
				content.add("\\\\\\hline\n");
			}

			content.add("\\ifdefined\\usehtml");
			content.add("\\end{tabular}");
			content.add("\\else");
			content.add("\\end{longtable}");
			content.add("\\fi\n");
		}
	}

	/**
	 * Write diagrams of material to report
	 * 
	 * @param content
	 *            Tex-writer
	 */
	protected void reportDiagrams(DocTexWriter content) {
		DocTexWriter localContent, curvesContent;
		ParamContainer diagram, curve;
		ParamField x = null, y = null, z;
		String name;
		String[] axes, params;
		double[] extrema = new double[4];
		TData value;
		String[] colors = { "blue", "green", "red", "cyan", "magenta", "orange" };
		int colorIndex;
		double[] limits;

		localContent = new DocTexWriter(content.getFile());

		for (Parameter p : this.getParam()) {
			name = p.getName();
			if ((p instanceof ParamContainer) && name.contains("vs")) {
				diagram = (ParamContainer) p;
				params = name.split("at");
				axes = params[0].split("vs");
				if (axes.length == 2) {
					curvesContent = new DocTexWriter(localContent.getFile());

					extrema[0] = Double.POSITIVE_INFINITY; // minX
					extrema[1] = Double.NEGATIVE_INFINITY; // maxX
					extrema[2] = Double.POSITIVE_INFINITY; // minY
					extrema[3] = Double.NEGATIVE_INFINITY; // maxY

					if (params.length > 1) {
						colorIndex = 0;
						for (Parameter c : diagram.getParam()) {
							if (c instanceof ParamContainer) {
								curve = (ParamContainer) c;
								x = (ParamField) curve.getParamByUniqueIdentifier(curve.getName() + "." + axes[0])
										.getParameter();
								y = (ParamField) curve.getParamByUniqueIdentifier(curve.getName() + "." + axes[1])
										.getParameter();
								this.reportDiagramCurve(curvesContent, x, y, extrema, colors[colorIndex++]);
								curvesContent.add("\t\t\\addlegendentry{");
								for (int k = 1; k < params.length; k++) {
									z = (ParamField) curve.getParamByUniqueIdentifier(curve.getName() + "." + params[k])
											.getParameter();
									if (k != 1) {
										curvesContent.add(", ");
									}
									curvesContent.addString(z.getName() + "=");
									value = z.getData();
									try {
										curvesContent.addString(((TNumeric) value).getReal(0, 0) + "");
									} catch (Exception ex) {
										curvesContent.addString(value.toString());
									}
									curvesContent.addString(" " + z.getUnit());
								}
								curvesContent.add("}\n\n");
								if (colorIndex >= colors.length) {
									colorIndex = 0;
								}
							}
						}
					} else {
						x = (ParamField) diagram.getParamByUniqueIdentifier(diagram.getName() + "." + axes[0])
								.getParameter();
						y = (ParamField) diagram.getParamByUniqueIdentifier(diagram.getName() + "." + axes[1])
								.getParameter();
						this.reportDiagramCurve(curvesContent, x, y, extrema, colors[0]);
					}

					// diagram
					localContent.add("\\begin{tikzpicture}\n");
					localContent.add("\t\\begin{axis}[\n");
					localContent.add("\t\ttitle={\\parbox{0.7\\textwidth}{\\raggedleft \\textbf{");

					// name
					for (int k = 0; k < axes.length; k++) {
						if (k > 0) {
							localContent.addString(" vs ");
						}
						localContent.addString(axes[k]);
					}
					for (int k = 1; k < params.length; k++) {
						localContent.addString(" @ ");
						localContent.addString(params[k]);
					}

					localContent.add("}}},\n");
					localContent.add("\t\txlabel={");
					localContent.addString(x.getName());
					localContent.add(" (");
					localContent.addString(x.getUnit());
					localContent.add(")},\n");
					localContent.add("\t\tylabel={");
					localContent.addString(y.getName());
					localContent.add(" (");
					localContent.addString(y.getUnit());
					localContent.add(")},\n");
					localContent.add("\t\tgrid = major,\n");
					limits = MathUtils.roundLimits(new double[] { extrema[0], extrema[1] });
					localContent.add("\t\txmin = " + limits[0] + ", xmax = " + limits[1] + ",\n");
					limits = MathUtils.roundLimits(new double[] { extrema[2], extrema[3] });
					localContent.add("\t\tymin = " + limits[0] + ", ymax = " + limits[1] + ",\n");
					localContent.add("\t\tlegend style={at={(0.03,0.97)},legend pos=outer north east}]\n\n");

					// curve
					localContent.addWriter(curvesContent);

					localContent.add("\t\\end{axis}\n");
					localContent.add("\\end{tikzpicture}\n\n");

					localContent.add("\\vspace{2ex}\n\n");
				}
			}
		}

		if (localContent.length() > 0) {
			content.add("\\vspace{3ex}\n");
			content.add(
					"\\fcolorbox{tablegray}{tablegray}{\\parbox{1\\textwidth}{\\color{white} \\textbf{Characteristic Curves}}}\n");
			content.add("\\vspace{2ex}\n");
			content.add("\\begin{center}\n");
			content.addWriter(localContent);
			content.add("\\end{center}\n");
		}

	}

	/**
	 * Draw curve of diagram
	 * 
	 * @param content
	 *            DocTexWriter
	 * @param x
	 *            Value for x-axis
	 * @param y
	 *            Value for y-axis
	 * @param extrema
	 *            Array with limits of x-/y-axis
	 * @param color
	 *            Color to draw
	 */
	protected void reportDiagramCurve(DocTexWriter content, ParamField x, ParamField y, double[] extrema,
			String color) {
		TNumeric xVal, yVal;
		int[] xSize, ySize;
		double val;
		int num;
		DocTexWriter localContent;

		xVal = ((TNumeric) x.getData());
		yVal = ((TNumeric) y.getData());

		xSize = xVal.size();
		ySize = yVal.size();

		if ((xSize[1] == ySize[1]) && (xSize[0] == 1) && (ySize[0] == 1)) {
			num = xSize[1];
			localContent = new DocTexWriter(content.getFile());

			try {
				for (int k = 0; k < num; k++) {
					// x
					val = xVal.getReal(0, k);
					if (val < extrema[0]) {
						extrema[0] = val;
					}
					if (val > extrema[1]) {
						extrema[1] = val;
					}
					localContent.add("\t\t\t(" + val + ",");

					// y
					val = yVal.getReal(0, k);
					if (val < extrema[2]) {
						extrema[2] = val;
					}
					if (val > extrema[3]) {
						extrema[3] = val;
					}
					localContent.add(val + ")\n");
				}

				content.add("\t\t\\addplot[smooth,mark=*,color=" + color + "] plot coordinates {\n");
				content.addWriter(localContent);
				content.add("\t\t};\n");
			} catch (Exception ex) {
			}
		}
	}

	@Override
	public boolean checkDocFiles(int index) {
		return false;
	}

	@Override
	public int update(Map<String, Object> args) throws Exception {
		return 0;
	}

	@Override
	public IParamMaterial getRMIImpl() throws RemoteException {
		return new RParamMaterial(this);
	}

	@Override
	public IParamMaterial getClientImpl() {
		return new CParamMaterial(this);
	}
}
