/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9734 $
 * $Date: 2020-04-06 09:51:22 +0200 (Mon, 06 Apr 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.awt.Toolkit;
import java.awt.datatransfer.Transferable;
import java.io.File;
import java.io.IOException;
import java.rmi.RemoteException;
import java.security.NoSuchAlgorithmException;
import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TNull;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.docu.ReportStructure;
import at.lcm.bu21.magopt.utils.ParameterTransferable;

/**
 * Basic class of all Parameters with all basic features
 * 
 * @author Werner Koppelsttter
 */
public abstract class Parameter implements IParameter {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** ID of parameter */
	private String pId = "";

	/** Parameter Name */
	private String name;

	/** Parameter Description */
	private String description;

	/** Parameter Comment */
	private String comment = "";

	/** Constant flag of parameter */
	private boolean constant = false;

	/** print flag for documentation */
	private boolean print = true;

	/** Reference to parent parameter */
	private ParamContainer parent;

	/** List of Parameter which use this Parameter */
	private ArrayList<Parameter> usedBy = new ArrayList<Parameter>(0);

	/** List of Parameter which export to this Parameter */
	private ArrayList<Parameter> setBy = new ArrayList<Parameter>(0);

	/** Temporary data */
	private transient Hashtable<String, Object> tmpData = null;

	/**
	 * Class that define the path to the parameter with all container-interfaces
	 * 
	 * @author Werner Koppelsttter
	 */
	public class ParameterPath {

		/** Parameter */
		private Parameter parameter;

		/** Path to parameter */
		private ParameterPath path = null;

		/**
		 * Create ParamterPath to parameter
		 * 
		 * @param parameter
		 *            Parameter
		 */
		ParameterPath(Parameter parameter) {
			this.parameter = parameter;
		}

		/**
		 * Get parameter
		 * 
		 * @return Parameter
		 */
		public Parameter getParameter() {
			return this.parameter;
		}

		/**
		 * Get path to parameter
		 * 
		 * @return Path to parameter
		 */
		public ParameterPath getPath() {
			return this.path;
		}

		/**
		 * Set path to parameter
		 * 
		 * @param path
		 *            Path
		 */
		void setPath(ParameterPath path) {
			this.path = path;
		}

	}

	@Override
	public DocFile[] getDocFiles() {
		Dataset ds;
		DocFile[] docs = null;

		ds = this.getDataset();
		if (ds != null) {
			docs = ds.getDocFiles(this.getPId());
		}

		return docs;
	}

	@Override
	public DocFile getDocFile(String name) {
		Dataset ds;
		DocFile doc = null;

		ds = this.getDataset();
		if (ds != null) {
			doc = ds.getDocFile(this.getPId(), name);
		}
		return doc;
	}

	@Override
	public File getFileFromDocFile(DocFile doc) {
		String path;
		File file = null;
		Parameter root;

		root = this.getRoot();
		if (root instanceof Project) {
			path = ((Project) root).getFilesDirectory().getPath();
			file = FileUtils.trimPath(new File(path, doc.getFile()));
		}

		return file;
	}

	@Override
	public void registerDocFile(String name, String strFile, String description)
			throws NoSuchAlgorithmException, IOException {
		Dataset ds;
		File file = new File(strFile);
		DocFile docFile;
		String tmp;

		ds = this.getDataset();
		if ((ds != null) && file.exists()) {
			docFile = new DocFile();
			docFile.setName(name);
			ds.setDocFile(this.getPId(), docFile);

			tmp = this.getIdentifier();
			if (tmp.indexOf(".") != -1) {
				tmp = tmp.substring(tmp.indexOf(".") + 1);
			}
			docFile.setParamName(tmp);

			docFile.setFile(
					FileUtils.getRelativePath(((Project) this.getRoot()).getFilesDirectory().getPath(), strFile));

			docFile.setDescription(description);
			docFile.setChecksum(FileUtils.checkSum(file));

		}
	}

	/**
	 * Unregister all doc files
	 */
	@Override
	public void unregisterDocFiles() {
		Dataset ds;

		ds = this.getDataset();
		if (ds != null) {
			ds.unregisterDocFiles(this.getPId());
		}
	}

	/**
	 * Unregister specified doc file
	 * 
	 * @param name
	 *            Name of docfile to remove
	 */
	@Override
	public void unregisterDocFile(String name) {
		Dataset ds;

		ds = this.getDataset();
		if (ds != null) {
			ds.unregisterDocFile(this.getPId(), name);
		}
	}

	/**
	 * Reset DocFiles (set files to invalid)
	 */
	@Override
	public void resetDocFiles() {
		this.resetDocFiles(this.getIndex());
	}

	/**
	 * Reset DocFiles (set files to invalid)
	 * 
	 * @param index
	 *            Index to reset
	 */
	@Override
	public void resetDocFiles(int index) {
		Dataset ds;

		ds = this.getDataset(index);
		if (ds != null) {
			ds.resetDocFiles(this.getPId());
		}
	}

	/**
	 * Delete registered doc-files
	 */
	@Override
	public void deleteDocFiles() {
		this.deleteDocFiles(this.getIndex());
	}

	/**
	 * Delete registered doc-files
	 * 
	 * @param index
	 *            Index of dataset
	 */
	@Override
	public void deleteDocFiles(int index) {
		DocFile[] docs;
		File file;
		File[] fileList;
		String[] fileParts, parts;

		docs = this.getDocFiles();
		if (docs != null) {
			for (DocFile d : docs) {
				file = new File(FileUtils.trimPath(((Project) this.getRoot()).getFilesDirectory()), d.getFile());
				fileParts = FileUtils.splitName(file.getName());
				fileList = file.getParentFile().listFiles();
				if (fileList != null) {
					for (File f : fileList) {
						parts = FileUtils.splitName(f.getName());
						if (fileParts[0].compareToIgnoreCase(parts[0]) == 0) {
							f.delete();
						}
					}
				}
				this.unregisterDocFile(d.getName());
			}
		}
	}

	/**
	 * Check if all files are valid (exist, checksum)
	 * 
	 * @return Files valid
	 */
	@Override
	public boolean checkDocFiles() {
		return this.checkDocFiles(this.getIndex());
	}

	/**
	 * Check if all files are valid (exist, checksum)
	 * 
	 * @param index
	 *            Index to reset
	 * @return Files valid
	 */
	@Override
	public boolean checkDocFiles(int index) {
		Dataset ds;
		boolean valid = true;
		DocFile[] list;
		String path = ((Project) this.getRoot()).getFilesDirectory(index).getPath();

		ds = this.getDataset(index);
		if (ds != null) {
			list = ds.getDocFiles(this.getPId());
			if (list.length == 0) {
				valid = false;
			} else {
				for (DocFile doc : list) {
					valid &= this.checkDocFile(doc, path);
				}
			}
		}

		return valid;
	}

	@Override
	public boolean checkDocFile(String docfile) {
		return this.checkDocFile(this.getIndex(), docfile);
	}

	/**
	 * Check if all files are valid (exist, checksum)
	 * 
	 * @param index
	 *            Index to reset
	 * @param docfile
	 *            Name of docfile
	 * @return Files valid
	 */
	@Override
	public boolean checkDocFile(int index, String docfile) {
		Dataset ds;
		boolean valid = false;
		DocFile[] list;
		String path = ((Project) this.getRoot()).getFilesDirectory().getPath();

		ds = this.getDataset(index);
		if (ds != null) {
			list = ds.getDocFiles(this.getPId());
			for (DocFile doc : list) {
				if (doc.getName().compareTo(docfile) == 0) {
					valid = this.checkDocFile(doc, path);
				}
			}
		}

		return valid;
	}

	/**
	 * Check doc-file if file exist and is checksum is equal
	 * 
	 * @param docfile
	 *            Docfile to check
	 * @param path
	 *            Path of docfiles directory
	 * @return Docfile valid
	 */
	protected boolean checkDocFile(DocFile docfile, String path) {
		boolean valid = false;
		File file;

		if (docfile.getChecksum().length() != 0) {
			file = FileUtils.trimPath(new File(path, docfile.getFile()));
			if (file.exists()) {
				try {
					valid = FileUtils.checkSum(file).compareTo(docfile.getChecksum()) == 0;
				} catch (Exception e) {
				}
			}
		}

		return valid;
	}

	@Override
	public List<String> listTabNames() {
		List<String> list = new LinkedList<String>();
		list.add("DocFiles");
		return list;
	}

	/**
	 * Constructor for loading a xml-file
	 * 
	 * @throws Exception
	 */
	Parameter() throws Exception {
		this("XML_Load", "XML load constructor");
	}

	/**
	 * Constructor with name
	 * 
	 * @param name
	 *            Name of Parameter
	 * @throws Exception
	 */
	public Parameter(String name) throws Exception {
		this(name, name);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name
	 *            Name of Parameter
	 * @param description
	 *            Description of Parameter
	 * @throws Exception
	 */
	public Parameter(String name, String description) throws Exception {
		if ((name == null) || name.isEmpty()) {
			throw new Exception("Name is necessary!");
		} else if (name.indexOf(".") != -1) {
			throw new Exception("'.' is not allowed in name!");
		}
		this.name = name;
		this.setDescription(description);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		Parameter other = (Parameter) obj;
		if (this.pId == null) {
			if (other.pId != null) {
				return false;
			}
		} else if (!this.pId.equals(other.pId)) {
			return false;
		}
		if (this.name == null) {
			if (other.name != null) {
				return false;
			}
		} else if (!this.name.equals(other.name)) {
			return false;
		}
		if (this.description == null) {
			if (other.description != null) {
				return false;
			}
		} else if (!this.description.equals(other.description)) {
			return false;
		}
		if (this.comment == null) {
			if (other.comment != null) {
				return false;
			}
		} else if (!this.comment.equals(other.comment)) {
			return false;
		}
		if (this.constant != other.constant) {
			return false;
		}
		if (this.print != other.print) {
			return false;
		}
		return true;
	}

	/**
	 * Get Classname
	 * 
	 * @return classname
	 */
	protected String getClassName() {
		return this.getClass().getSimpleName();
	}

	@Override
	public Object clone() throws CloneNotSupportedException {
		return super.clone();
	}

	/**
	 * Create a deep clone of given parameter
	 * 
	 * @return Deep clone of parameter
	 * @throws CloneException
	 */
	public Parameter deepClone() throws CloneException {
		return (Parameter) Clone.deepClone(this);
	}

	/**
	 * Clone parameter
	 * 
	 * @return Clone of parameter
	 * @throws Exception
	 */
	public Parameter cloneParam() throws Exception {
		Parameter clone = this.cloneParam(null, -1);
		return clone;

	}

	/**
	 * Clone parameter
	 * 
	 * @param clone
	 *            Existing object to clone to
	 * @param index
	 *            Index to clone (-1 for all datasets)
	 * @return Clone of parameter
	 * @throws Exception
	 */
	public Parameter cloneParam(Parameter clone, int index) throws Exception {
		return this.cloneParam(clone, index, true);
	}

	/**
	 * Clone parameter
	 * 
	 * @param clone
	 *            Existing object to clone to
	 * @param index
	 *            Index to clone (-1 for all datasets)
	 * @param docFiles
	 *            Clone doc-files
	 * @return Clone of parameter
	 * @throws Exception
	 */
	public Parameter cloneParam(Parameter clone, int index, boolean docFiles) throws Exception {
		Dataset dsThis, dsClone;
		ParamDataStruct pdThis, pdClone;
		File dir;

		if (clone == null) {
			clone = this.getClass().newInstance();
			if (!(clone instanceof Project)) {
				int k = 0;
				do {
					dir = FileUtils.getFile(System.getProperty("java.io.tmpdir"), "MagOpt", Long.toString(k++, 36));
				} while (dir.exists());
				dir.mkdirs();
				dir.deleteOnExit();
				Project p = new Project("Temp", "Temp", new File(dir, "Temp.mop"));
				p.addParam(clone);
			}

		}

		this.cloneParam(clone);

		pdThis = this.getParameterStruct();
		pdClone = clone.getParameterStruct();

		if ((pdThis != null) && (pdClone != null)) {
			pdThis.cloneParam(pdClone);
			if (index < 0) {
				for (int k = 0; k < this.getLength(); k++) {
					dsThis = this.getDataset(k);
					dsClone = new Dataset(dsThis.getDsId());
					dsClone.setComment(dsThis.getComment());
					dsClone.setEventLog(dsThis.getEventLog().deepClone());
					pdClone.addParamDataset(dsClone);
					this.cloneParam(clone, dsThis, dsClone, docFiles);
				}
				clone.setIndex(this.getIndex());
			} else {
				dsThis = this.getDataset(index);
				dsClone = new Dataset(dsThis.getDsId());
				dsClone.setComment(dsThis.getComment());
				dsClone.setEventLog(dsThis.getEventLog().deepClone());
				pdClone.addParamDataset(dsClone);
				this.cloneParam(clone, dsThis, dsClone, docFiles);
				clone.setIndex(0);
			}

		}
		return clone;
	}

	/**
	 * Clone parameter structure
	 * 
	 * @param clone
	 *            Existing object to clone to
	 * @throws CloneException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 */
	void cloneParam(Parameter clone) throws CloneException, InstantiationException, IllegalAccessException {
		clone.pId = this.pId;
		clone.name = this.name;
		clone.description = this.description;
		clone.comment = this.comment;

		clone.print = this.print;
		clone.constant = this.constant;
	}

	/**
	 * Clone parameter data
	 * 
	 * @param clone
	 *            Existing object to clone to
	 * @param dsThis
	 *            Dataset of this parameter
	 * @param dsClone
	 *            Dataset of thie clone parameter
	 * @param docFiles
	 *            Clone doc-files
	 * @throws CloneException
	 */
	void cloneParam(Parameter clone, Dataset dsThis, Dataset dsClone, boolean docFiles) throws CloneException {
		TData data;
		File thisDir, cloneDir, thisFile, cloneFile;

		data = dsThis.getValue(this.getPId(), ParamSetKey.MODIFIED);
		if (data != null) {
			dsClone.setValue(clone.getPId(), ParamSetKey.MODIFIED, data.deepClone());
		} else {
			dsClone.rmValue(clone.getPId(), ParamSetKey.MODIFIED);
		}

		data = dsThis.getValue(this.getPId(), ParamSetKey.UNCERTAIN);
		if (data != null) {
			dsClone.setValue(clone.getPId(), ParamSetKey.UNCERTAIN, data.deepClone());
		} else {
			dsClone.rmValue(clone.getPId(), ParamSetKey.UNCERTAIN);
		}

		data = dsThis.getValue(this.getPId(), ParamSetKey.ERROR);
		if (data != null) {
			dsClone.setValue(clone.getPId(), ParamSetKey.ERROR, data.deepClone());
		} else {
			dsClone.rmValue(clone.getPId(), ParamSetKey.ERROR);
		}

		thisDir = new File(((Project) this.getRoot()).getAllFilesDirectory(), dsThis.getDsId());
		cloneDir = new File(((Project) clone.getRoot()).getAllFilesDirectory(), dsClone.getDsId());
		// DocFiles
		if (docFiles) {
			for (DocFile doc : dsThis.getDocFiles(this.getPId())) {
				dsClone.setDocFile(clone.getPId(), (DocFile) Clone.deepClone(doc));
				// copy file
				thisFile = new File(thisDir, doc.getFile());
				cloneFile = new File(cloneDir, doc.getFile());
				if (thisFile.exists()) {
					try {
						FileUtils.copy(thisFile, cloneFile);
					} catch (IOException e) {
					}
				}
			}
		}

		// ParameterDirectories
		thisDir = new File(thisDir, "param");
		cloneDir = new File(cloneDir, "param");
		thisFile = new File(thisDir, this.getPId());
		cloneFile = new File(cloneDir, clone.getPId());
		if (thisFile.exists()) {
			try {
				FileUtils.copy(thisFile, cloneFile);
			} catch (IOException e) {
			}
		}
	}

	@Override
	public String getName() {
		return this.name;
	}

	@Override
	public void setName(String name) throws Exception {
		String ident, par;
		ParamContainer cont;
		Parameter[] params;

		if ((name == null) || name.isEmpty()) {
			this.warning("Name is necessary!");
		} else if (name.indexOf(".") != -1) {
			this.warning("'.' is not allowed in name!");
		} else if (name.equals(this.name)) {
		} else if ((this.parent != null) && (this.parent.existsName(name))) {
			this.warning("Object with identical name existing!");
		} else {
			ident = this.getIdentifier();

			// change dependency entries
			params = this.getParameters(Parameter.class);
			for (Parameter param : params) {
				//
				Parameter[] usedList = param.getUsedBy();
				for (Parameter used : usedList) {
					this.renameDepends(param, ((Dependency) used).getDepend().getImports(), ident, name);
				}

				Parameter[] setList = param.getSetBy();
				for (Parameter set : setList) {
					this.renameDepends(param, ((Dependency) set).getDepend().getExports(), ident, name);
				}
			}

			// change interfaces entries of parents
			cont = this.getParent();
			while (cont != null) {
				for (String iname : cont.listPublicInterfaceName()) {
					if (cont.getPublicInterfaceParameter(iname) != null) {
						par = this.changeParamString(cont.getPublicInterfaceParam(iname),
								ident.substring(cont.getIdentifier().length() + 1), name);
						if (par.compareTo(cont.getPublicInterfaceParam(iname)) != 0) {
							cont.setPublicInterfaceParam(iname, par);
						}
					}

				}
				cont = cont.getParent();
			}

			this.name = name;
			this.refresh();
		}
	}

	/**
	 * Rename depends if name is in parameter-string
	 * 
	 * @param param
	 *            Parameter to check dependencies
	 * @param items
	 *            Dependency-items to check of parameter
	 * @param ident
	 *            Identifier of parameter with old name
	 * @param name
	 *            New name of parameter
	 */
	private void renameDepends(Parameter param, DependItem[] items, String ident, String name) {
		ParameterPath path;
		String id, par;

		for (DependItem item : items) {
			if (item.getHandle() == param) {
				// get parameter of direct link or parameter of first interface
				path = this.getObjectPath(item.getParam());
				if (path != null) {
					while (path.getPath() != null) {
						path = path.getPath();
					}
					// chek if found parameter is a sub-parameter of identifier
					id = path.getParameter().getIdentifier();
					if (id.startsWith(ident + ".") || (id.compareTo(ident) == 0)) {
						// change parameter-string
						par = this.changeParamString(item.getParam(), ident, name);
						if (par.compareTo(item.getParam()) != 0) {
							item.setParam(par);
						}
					}
				}
			}
		}
	}

	/**
	 * Change parameter-string
	 * 
	 * @param param
	 *            Parameter-string used
	 * @param ident
	 *            Identifier of parameter with old name
	 * @param name
	 *            New name of parameter
	 * @return Changed Parameter-string
	 */
	private String changeParamString(String param, String ident, String name) {
		String[] params;
		String str;
		int k, index;

		// search for parts of param
		params = param.split("\\.");
		str = "." + ident + ".";
		k = 0;
		while ((k < params.length) && ((index = str.indexOf("." + params[k] + ".")) != -1)) {
			str = str.substring(index + params[k].length() + 1);
			k++;
		}
		// replace old name of parent dependency
		if (str.length() == 1) {
			params[k - 1] = name;
		}

		// concate all param-parts
		str = "";
		for (String tmp : params) {
			str += "." + tmp;
		}
		;

		return str.substring(1);
	}

	@Override
	public String getDescription() {
		return this.description;
	}

	@Override
	public void setDescription(String description) {
		if (description != null) {
			this.description = description;
		} else {
			this.description = "";
		}
	}

	@Override
	public String getComment() {
		return this.comment;
	}

	@Override
	public void setComment(String comment) {
		if (comment != null) {
			this.comment = comment;
		} else {
			this.comment = "";
		}
	}

	@Override
	public boolean getPrint() {
		return this.print;
	}

	@Override
	public void setPrint(boolean print) {
		this.print = print;
	}

	@Override
	public boolean getConstant() {
		return this.constant;
	}

	@Override
	public void setConstant(boolean constant) {
		if (this.getConstant() != constant) {
			this.constant = constant;
		}
	}

	@Override
	public boolean getUsed() {
		return (this.usedBy.size() != 0) || (this.setBy.size() != 0);
	}

	@Override
	public boolean getReadOnly() {
		boolean bool = this.setBy.size() != 0;
		for (int i = 0; i < this.setBy.size(); i++) {
			for (int j = 0; j < this.usedBy.size(); j++) {
				if (this.setBy.get(i) == this.usedBy.get(j)) {
					bool = false;
				}
			}
		}
		return bool;
	}

	@Override
	// FIXME RMI
	public IParamContainer getParentParameter() {
		return this.parent;
	}

	/**
	 * Get parent-parameter
	 * 
	 * @return Parent
	 */
	public ParamContainer getParent() {
		return this.parent;
	}

	/**
	 * Set parent of Parameter
	 * 
	 * @param parent
	 *            Parent
	 */
	void setParent(ParamContainer parent) {
		this.parent = parent;
	}

	@Override
	public Parameter getRoot() {
		Parameter param = null;

		if (this.parent == null) {
			param = this;
		} else {
			param = this.parent.getRoot();
		}
		return param;
	}

	@Override
	public TData getData() {
		return this.getData(this.getIndex());
	}

	@Override
	public TData getData(int index) {
		return new TNull();
	}

	/**
	 * Get structure of data
	 * 
	 * @return Structure of data
	 */
	public TStruct getDataStruct() {
		TStruct struct;

		struct = new TStruct();
		struct.put("value__", this.getData());
		struct.put("description__", new TString(this.getDescription()));

		return struct;
	}

	/**
	 * Get recursive structure of data
	 * 
	 * @return Structure of data
	 */
	public TStruct getDataStructRec() {
		return this.getDataStruct();
	}

	@Override
	public void setData(TData data) throws Exception {
		this.warning("Unable to set data to " + this.getClass().getSimpleName());
	}

	@Override
	public boolean getModified() {
		return this.getModified(this.getIndex());
	}

	/**
	 * Get modified-status
	 * 
	 * @param index
	 *            Dataset-index
	 * @return modified-flag
	 */
	public boolean getModified(int index) {
		Dataset ds;
		TData value;
		boolean bool = true;

		ds = this.getDataset(index);
		if (ds != null) {
			value = ds.getValue(this.getPId(), ParamSetKey.MODIFIED);
			if (value != null) {
				bool = ((TBoolean) value).getBoolean();
			}
		}

		return bool;
	}

	@Override
	public void setModified(boolean modified) {
		this.setModified(this.getIndex(), modified);
		// FIXME setUncertain (refresh)
		this.refreshFlags();
	}

	/**
	 * Set modified-status of given index
	 * 
	 * @param index
	 *            Index of dataset
	 * @param modified
	 *            modified-flag
	 */
	void setModified(int index, boolean modified) {
		Dataset ds;
		if (modified != this.getModified(index)) {
			ds = this.getDataset(index);
			if (ds != null) {
				ds.setValue(this.getPId(), ParamSetKey.MODIFIED, new TBoolean(modified));
			}
			this.resetDocFiles(index);
		}
	}

	@Override
	public Parameter[] getUsedBy() {
		return this.usedBy.toArray(new Parameter[0]);
	}

	/**
	 * Add parameter to used-by-list
	 * 
	 * @param param
	 *            Parameter to add
	 */
	protected void addUsedBy(Parameter param) {
		this.usedBy.add(param);
	}

	/**
	 * Remove Parameter from used-by-list
	 * 
	 * @param param
	 *            Parameter to remove
	 */
	protected void rmUsedBy(Parameter param) {
		this.usedBy.remove(param);
	}

	/**
	 * Clear used-by-list
	 */
	protected void clearUsedBy() {
		this.usedBy.clear();
	}

	@Override
	public Parameter[] getSetBy() {
		return this.setBy.toArray(new Parameter[0]);
	}

	/**
	 * Add parametert to set-by-list
	 * 
	 * @param param
	 *            Parameter to add
	 */
	protected void addSetBy(Parameter param) {
		this.setBy.add(param);
	}

	/**
	 * Remove parameter from set-by-list
	 * 
	 * @param param
	 *            Parameter to remove
	 */
	protected void rmSetBy(Parameter param) {
		this.setBy.remove(param);
	}

	/**
	 * Clear set-by-list
	 */
	protected void clearSetBy() {
		this.setBy.clear();
	}

	public boolean cut() {
		if (!(this instanceof Project)) {
			this.getParent().rmParameter(this);
			return this.copy();
		}
		return false;
	}

	public boolean copy() {
		Toolkit.getDefaultToolkit().getSystemClipboard().setContents(new ParameterTransferable(this), null);
		return true;
	}

	public boolean paste() {
		ParamContainer parent;
		Transferable content;
		Parameter param = null;

		if (this instanceof IParamContainer) {
			parent = (ParamContainer) this;
		} else {
			parent = this.getParent();
		}

		content = Toolkit.getDefaultToolkit().getSystemClipboard().getContents(this);
		try {
			// FIXME check files
			param = (Parameter) content.getTransferData(ParameterTransferable._paramFlavor);
		} catch (Exception ex1) {
			this.warning("Unable to paste clipboard-content as parameter (" + ex1.getMessage() + ")",
					ex1.getStackTrace());
		}

		if (param != null) {
			try {
				parent.addParam(param);
				return true;
			} catch (Exception e) {
				parent.warning("Unable to paste parameter (" + e.getMessage() + ")", e.getStackTrace());
			}
		}

		return false;
	}

	@Override
	public Parameter removeParam() {
		ParamContainer par;

		par = this.parent;
		if (par != null) {
			par.rmParameter(this);
			this.setParent(null);
			par.refresh();
			this.refresh();
		}

		return this;
	}

	@Override
	public void moveUp() {
		int index;

		if (this.parent != null) {
			index = this.getIndexOfParameter();
			if ((index != -1) && (index > 0)) {
				try {
					this.parent.addParam(this, index - 1);
				} catch (Exception e) {
				}
			}
		}
	}

	@Override
	public void moveDown() {
		int index;

		if (this.parent != null) {
			index = this.getIndexOfParameter();
			if ((index != -1) && (index < (this.parent.getParam().length - 1))) {
				try {
					this.parent.addParam(this, index + 2);
				} catch (Exception e) {
				}
			}
		}
	}

	@Override
	public int getIndexOfParameter() {
		int index = -1;

		if (this.parent != null) {
			index = this.parent.getIndexOfParam(this);
		}

		return index;
	}

	@Override
	public boolean getUncertain() {
		return this.getUncertain(this.getIndex());
	}

	/**
	 * Get uncertain-status
	 * 
	 * @param index
	 *            Index of dataset
	 * @return uncertain-flag
	 */
	public boolean getUncertain(int index) {
		Dataset ds;
		TData value;
		boolean bool = false;

		ds = this.getDataset(index);
		if (ds != null) {
			value = ds.getValue(this.getPId(), ParamSetKey.UNCERTAIN);
			if (value != null) {
				bool = ((TBoolean) value).getBoolean();
			}
		}

		return bool;
	}

	/**
	 * Set uncertain-status
	 * 
	 * @param uncertain
	 *            uncertain-flag
	 */
	void setUncertain(boolean uncertain) {
		this.setUncertain(this.getIndex(), uncertain);
	}

	/**
	 * Set uncertain-status
	 * 
	 * @param index
	 *            Index of dataset
	 * @param uncertain
	 *            uncertain-flag
	 */
	void setUncertain(int index, boolean uncertain) {
		Dataset ds;

		ds = this.getDataset(index);
		if (ds != null) {
			ds.setValue(this.getPId(), ParamSetKey.UNCERTAIN, new TBoolean(uncertain));
		}

	}

	@Override
	public boolean getError() {
		return this.getError(this.getIndex());
	}

	/**
	 * Get error-status of given dataset
	 * 
	 * @param index
	 *            Index of dataset
	 * @return error-flag
	 */
	public boolean getError(int index) {
		Dataset ds;
		TData value;
		boolean bool = false;

		ds = this.getDataset(index);
		if (ds != null) {
			value = ds.getValue(this.getPId(), ParamSetKey.ERROR);
			if (value != null) {
				bool = ((TBoolean) value).getBoolean();
			}
		}

		return bool;
	}

	@Override
	public boolean[] getAllError() {
		boolean[] list = new boolean[this.getLength()];

		for (int i = 0; i < this.getLength(); i++) {
			list[i] = this.getError(i);
		}

		return list;
	}

	@Override
	public void setError(boolean error) {
		this.setError(this.getIndex(), error);
	}

	/**
	 * Set error-status of current dataset
	 * 
	 * @param index
	 *            Index of dataset
	 * @param error
	 *            error-flag
	 */
	void setError(int index, boolean error) {
		this.setErrorStructure(index, error);
	}

	/**
	 * Set error-status of current dataset
	 * 
	 * @param index
	 *            Index of dataset
	 * @param error
	 *            error-flag
	 */
	protected void setErrorStructure(int index, boolean error) {
		Dataset ds;

		ds = this.getDataset(index);
		if (ds != null) {
			ds.setValue(this.getPId(), ParamSetKey.ERROR, new TBoolean(error));
		}
		if (error && (this.parent != null)) {
			this.parent.setErrorStructure(index, error);
		}
	}

	@Override
	public String getPId() {
		return this.pId;
	}

	/**
	 * Set ID of parameter
	 * 
	 * @param pId
	 *            Parameter-ID
	 */
	void setPId(String pId) {
		this.pId = pId;
	}

	@Override
	public String getIdentifier() {
		Parameter parent;
		if ((parent = this.getParent()) == null) {
			return this.getName();
		} else {
			return parent.getIdentifier() + "." + this.getName();
		}
	}

	@Override
	public String getShortIdentifier() {
		LinkedList<LinkedList<String>> list;
		LinkedList<String> param;
		StringBuffer uName;
		int i;
		boolean equal;

		// list parameters with equal names
		list = new LinkedList<LinkedList<String>>();
		for (Parameter p : this.getRoot().getParameters(Parameter.class)) {
			if ((p != this) && p.getName().equals(this.name)) {
				list.add(new LinkedList<String>(Arrays.asList(p.getIdentifier().split("\\."))));
				list.getLast().removeLast();
			}
		}

		uName = new StringBuffer();
		for (String str : this.getIdentifier().split("\\.")) {
			i = 0;
			equal = true;
			while (i < list.size()) {
				param = list.get(i);
				if (!param.isEmpty() && param.get(0).equals(str)) {
					param.remove(0);
					i++;
				} else {
					equal = false;
					list.remove(i);
				}
			}
			if (!equal) {
				uName.append(".");
				uName.append(str);
			}
		}
		uName.append(".");
		uName.append(this.getName());

		return uName.deleteCharAt(0).toString();
	}

	/**
	 * Execute methode execute of ExecStructure on all Parameter in structure.
	 * 
	 * @param wrapper
	 *            ExecStructure object
	 */
	void execStructure(ExecStructure wrapper) {
		wrapper.execute(this);
	}

	/**
	 * Get a list of Parameter from the structure. Only return Parameter when method
	 * accept of FileterStructure element retrun true.
	 * 
	 * @param filter
	 *            FilterStructure object
	 * @return List of Parameter
	 */
	AbstractList<Parameter> filterParameter(FilterStructure filter) {
		ArrayList<Parameter> list = new ArrayList<Parameter>(0);
		if (filter.accept(this)) {
			list.add(this);
		}
		return list;
	}

	@Override
	public void refresh() {
		this.refresh(false);
	}

	/**
	 * Refresh structure and flags
	 * 
	 * @param load
	 *            Flag to indicate loading
	 */
	public void refresh(boolean load) {
		this.refresh(load, false);
	}

	/**
	 * Short refresh structure and flags
	 */
	public void refreshPartial() {
		this.refresh(false, true);
	}

	/**
	 * Refresh structure and flags
	 * 
	 * @param load
	 *            Flag to indicate loading
	 * @param partial
	 *            Flag to indicate short refresh
	 */
	void refresh(boolean load, boolean partial) {
		if (this.parent != null) {
			this.parent.refresh(load, partial);
		} else {
			this.refresh(load, partial, this.getParameters(Dependency.class));
		}
	}

	/**
	 * Refresh structure and flags
	 * 
	 * @param load
	 *            Flag to indicate loading
	 * @param partial
	 *            Short refresh (without refresh Dependency)
	 * @param dependencies
	 *            List of dependencies
	 */
	void refresh(boolean load, boolean partial, Parameter[] dependencies) {
		Parameter[] list;
		DependItem[] ex;
		Parameter node;

		// if (this.parent != null) {
		// this.parent.refresh(load, fast, dependencies);
		// } else {

		// update dependencies (import/export)
		for (Parameter p : dependencies) {
			((Dependency) p).refreshDepend(false, load);
		}

		if (!partial) {
			// refresh components
			for (Parameter p : this.getParameters(ParamContainer.class)) {
				Component comp = ((ParamContainer) p).getComponent();
				if (comp != null) {
					comp.refresh();
				}
			}

			// Refresh dynamic Parameter
			for (Parameter p : dependencies) {
				((Dependency) p).refreshDependency();
			}

			// Refresh File
			for (Parameter p : this.getParameters(ParamFile.class)) {
				((ParamFile) p).refreshFile();
			}
		}

		// Check exports
		for (Parameter p : dependencies) {
			ex = ((Dependency) p).getDepend().getExports();
			for (int i = 0; i < ex.length; i++) {
				node = ex[i].getHandle();
				if (node != null) {
					list = node.getSetBy();
					if (list.length == 1) {
						if (node.getConstant()) {
							this.warning("Export-parameter is constant! (" + node.getIdentifier() + ")\n Set by "
									+ list[0].getIdentifier());
						}
					} else if (list.length > 1) {
						this.warning("Parameter is set by more than one parameter! (" + node.getIdentifier()
								+ ")\n Set by " + p.getIdentifier());
					}
				}
			}
		}

		// refresh Uncertain & incomplete
		this.refreshFlags();
		// }
	}

	/**
	 * Refresh flags
	 */
	public void refreshFlags() {
		AbstractList<Parameter> list;

		if (this.parent != null) {
			this.parent.refreshFlags();
		} else {
			// clear uncertain flags
			this.execStructure(new ExecStructure() {
				@Override
				public void execute(Parameter param) {
					param.setUncertain(false);
				}
			});

			// modified and updateRequired list
			list = this.filterParameter(new FilterStructure() {
				@Override
				public boolean accept(Parameter param) {
					return param.getModified()
							|| ((param instanceof Function) && ((Function) param).getUpdateRequired());
				}
			});

			// set uncertain
			for (Parameter p : list) {
				p.refreshUncertain();
			}

			// set incompleteFlag
			this.refreshIncomplete();
		}

	}

	/**
	 * Check if refresh uncertain is necessary
	 */
	private void refreshUncertainNeeds() {
		if (!this.getUncertain() && !this.getModified()
				&& !((this instanceof Function) && ((Function) this).getUpdateRequired())) {
			this.refreshUncertain();
		}
	}

	/**
	 * Refresh uncertain flags
	 */
	void refreshUncertain() {
		DependItem[] ex;
		Parameter node;

		// set uncertain flag
		if (!this.getModified()) {
			this.setUncertain(true);
		}

		// function used this parameter
		for (int i = 0; i < this.usedBy.size(); i++) {
			node = this.usedBy.get(i);
			if (node != this) {
				node.refreshUncertainNeeds();
			}
		}

		// export of this parameter
		if (this instanceof Dependency) {
			ex = ((Dependency) this).getDepend().getExports();
			for (int i = 0; i < ex.length; i++) {
				node = ex[i].getHandle();
				if ((node != null) && (node != this)) {
					node.refreshUncertainNeeds();
				}
			}
		}
	}

	/**
	 * Refresh incomplete-status
	 * 
	 * @return incomplete-status
	 */
	protected boolean refreshIncomplete() {
		return this.getModified() || this.getUncertain() || this.getError();
	}

	@Override
	public Parameter[] getParameters(String c) {
		Parameter[] params = new Parameter[0];
		try {
			params = this.getParameters(Class.forName(c));
		} catch (ClassNotFoundException e) {
		}
		return params;
	}

	/**
	 * Get list of Parameter that are instances of class
	 * 
	 * @param c
	 * @return List of parameter
	 */
	public Parameter[] getParameters(final Class<?> c) {

		return this.filterParameter(new FilterStructure() {
			@Override
			public boolean accept(Parameter param) {
				return c.isAssignableFrom(param.getClass());
			};
		}).toArray(new Parameter[0]);
	}

	/**
	 * Get list of used input Parameter.
	 * 
	 * @return List of parameter
	 */
	public Parameter[] getUsedInputParameter() {
		return this.filterParameter(new FilterStructure() {
			@Override
			public boolean accept(Parameter param) {
				return (param instanceof ParamField) && param.getUsed() && !param.getReadOnly();
			};
		}).toArray(new Parameter[0]);
	}

	@Override
	public Parameter[] getDynamicParams() {
		if (this.parent != null) {
			return this.parent.getDynamicParams();
		} else {
			// TODO Workaround for Matlab->Java
			return this.filterParameter(new FilterStructure() {
				@Override
				public boolean accept(Parameter param) {
					boolean bool = false;
					TData data;

					if (param instanceof ParamField) {
						if (param.getUsed() && !param.getConstant() && !param.getReadOnly()) {
							data = ((ParamField) param).getData();
							if (data instanceof TNumeric) {
								if (((TNumeric) data).isScalar()) {
									bool = true;
								}
							}
						}
					}
					return bool;
				};
			}).toArray(new Parameter[0]);
		}
	}

	/**
	 * Get list of dynamic output parameters. Dynamic output parameter are parameter
	 * which can diversify. ParamField which are used, not constant, not writable
	 * and has one numeric value
	 * 
	 * @return List of parameter
	 */
	public Parameter[] getDynamicOutputs() {
		if (this.parent != null) {
			return this.parent.getDynamicOutputs();
		} else {
			// TODO Workaround for Matlab->Java
			return this.filterParameter(new FilterStructure() {
				@Override
				public boolean accept(Parameter param) {
					boolean bool = false;
					TData data;

					if (param instanceof ParamField) {
						if (param.getUsed() && !param.getConstant() && param.getReadOnly()) {
							data = ((ParamField) param).getData();
							if (data instanceof TNumeric) {
								if (((TNumeric) data).isScalar()) {
									bool = true;
								}
							}
						}
					}
					return bool;
				};
			}).toArray(new Parameter[0]);
		}
	}

	@Override
	public Parameter[] getAnalysisParams() {
		if (this.parent != null) {
			return this.parent.getAnalysisParams();
		} else {
			// TODO Workaround for Matlab->Java
			return this.filterParameter(new FilterStructure() {
				@Override
				public boolean accept(Parameter param) {
					return ((param instanceof ParamField) || (param instanceof ParamFormula)) && param.getUsed();
				};
			}).toArray(new Parameter[0]);
		}
	}

	@Override
	public Parameter getUniqueObject(String uniqueIdentifier) {
		Parameter root;
		ParameterPath p;
		Parameter obj = null;

		root = this.getRoot();
		p = root.getParamByUniqueIdentifier(uniqueIdentifier, true);
		if (p != null) {
			obj = p.getParameter();
		}

		return obj;
	}

	@Override
	public Parameter getObject(String identifier) {
		Parameter obj = null;
		ParameterPath p;

		p = this.getObjectPath(identifier);
		if (p != null) {
			obj = p.getParameter();
		}

		return obj;
	}

	/**
	 * Get first object which identifier correlate with given identifier
	 * 
	 * @param identifier
	 *            Identifier
	 * @return Found parameter
	 */
	public ParameterPath getObjectPath(String identifier) {
		AbstractList<ParameterPath> list;
		ParameterPath path = null;

		list = this.getObjectPaths(identifier);
		if (!list.isEmpty()) {
			path = list.get(0);
		}

		return path;
	}

	/**
	 * Get list of objects which identifier correlate with given identifier
	 * 
	 * @param identifier
	 *            Identifier
	 * @return Found parameter
	 */
	public AbstractList<ParameterPath> getObjectPaths(String identifier) {
		Parameter root;
		Parameter param;
		ParameterPath p;
		ArrayList<ParameterPath> list;

		list = new ArrayList<ParameterPath>();
		// unique object
		root = this.getRoot();
		p = root.getParamByUniqueIdentifier(identifier, false);
		if (p != null) {
			list.add(p);
		}

		// recursive objects
		param = this;
		while (param != null) {
			if ((param.getParent() == null)
					|| ((param instanceof ParamContainer) && (((ParamContainer) param).getAccess() != Access.PUBLIC))) {
				list.addAll(param.getParamByIdentifier(identifier, true));
			}
			param = param.getParent();
		}

		return list;
	}

	/**
	 * Return Parameter that identifier match with given identifier
	 * 
	 * @param uIdName
	 *            Unique identifier
	 * @return Found parameter
	 */
	public ParameterPath getParamByUniqueIdentifier(String uIdName) {
		return this.getParamByUniqueIdentifier(uIdName, false);
	}

	/**
	 * Return Parameter that identifier match with given identifier
	 * 
	 * @param uIdName
	 *            Unique identifier
	 * @param allAccess
	 *            Get parameter without access restriction
	 * @return Found parameter
	 */
	public ParameterPath getParamByUniqueIdentifier(String uIdName, boolean allAccess) {
		if (uIdName.compareTo(this.getName()) == 0) {
			return new ParameterPath(this);
		}
		return null;
	}

	/**
	 * Return a list of Parameter which identifier correlate with given identifier
	 * 
	 * @param idName
	 *            Identifier
	 * @return List of Parameter
	 */
	public AbstractList<ParameterPath> getParamByIdentifier(String idName) {
		return this.getParamByIdentifier(idName, false);
	}

	/**
	 * Return a list of Parameter which identifier correlate with given identifier
	 * 
	 * @param idName
	 *            Identifier
	 * @param thisPublic
	 *            Get parameter with public access on this object
	 * @return List of Parameter
	 */
	protected AbstractList<ParameterPath> getParamByIdentifier(String idName, boolean thisPublic) {
		ArrayList<ParameterPath> list = new ArrayList<ParameterPath>(0);
		if (idName.compareTo(this.getName()) == 0) {
			list.add(new ParameterPath(this));
		}
		return list;
	}

	@Override
	public Parameter getParamByPid(String pid) {
		Parameter p = null;

		if (this.getPId().compareTo(pid) == 0) {
			p = this;
		}

		return p;
	}

	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();
		str.append("Parameter\n");
		str.append("\n\tName        : ");
		str.append(this.getName());
		str.append("\n\tDescription : ");
		str.append(this.getDescription());
		str.append("\n\tComment : ");
		str.append(this.getComment());
		return str.toString();
	}

	/**
	 * Get a stack-trace of caller
	 * 
	 * @return stack-trace
	 */
	protected StackTraceElement[] getStackTrace() {
		StackTraceElement[] stack = new Throwable().getStackTrace();
		StackTraceElement[] newStack = new StackTraceElement[stack.length - 2];

		System.arraycopy(stack, 2, newStack, 0, newStack.length);
		return newStack;

	}

	@Override
	public void info(String message) {
		this.info(message, this.getStackTrace());
	}

	@Override
	public void info(String message, StackTraceElement[] stack) {
		this.info(this.getIdentifier(), message, stack);
	}

	@Override
	public void info(String identifier, String message, StackTraceElement[] stack) {
		Parameter parent;
		if ((parent = this.getParent()) != null) {
			parent.info(identifier, message, stack);
		} else {
			System.out.println("INFO: " + message);
		}
	}

	@Override
	public void warning(String message) {
		this.warning(message, this.getStackTrace());
	}

	@Override
	public void warning(String message, StackTraceElement[] stack) {
		this.warning(this.getIdentifier(), message, stack);
	}

	@Override
	public void warning(String identifier, String message, StackTraceElement[] stack) {
		Parameter parent;
		if ((parent = this.getParent()) != null) {
			parent.warning(identifier, message, stack);
		} else {
			System.out.println("WARNING: " + message);
		}
	}

	@Override
	public void error(String message) {
		this.error(message, this.getStackTrace());
	}

	@Override
	public void error(String message, StackTraceElement[] stack) {
		this.error(this.getIdentifier(), message, stack);
	}

	@Override
	public void error(String identifier, String message, StackTraceElement[] stack) {
		this.setError(true);

		this.errorStructure(identifier, message, stack);
	}

	/**
	 * Add error message to event-log up to the root (without ser error-flag)
	 * 
	 * @param identifier
	 *            Identifier
	 * @param message
	 *            Message
	 * @param stack
	 *            Stack-trace
	 */
	protected void errorStructure(String identifier, String message, StackTraceElement[] stack) {
		Parameter parent;

		if ((parent = this.getParent()) != null) {
			parent.errorStructure(identifier, message, stack);
		} else {
			System.out.println("ERROR: " + message);
		}
	}

	/**
	 * Read object from xml-file
	 * 
	 * @param xmlReader
	 *            XMLReader to read xml-file
	 * @throws Exception
	 */
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("PID")) {
				xmlReader.next();
				this.pId = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.pId += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("Name")) {
				xmlReader.next();
				this.name = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.name += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("Description")) {
				xmlReader.next();
				this.description = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.description += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("Comment")) {
				xmlReader.next();
				this.comment = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.comment += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("Constant")) {
				xmlReader.next();
				this.constant = Boolean.parseBoolean(xmlReader.getText());
				xmlReader.next();
				xmlReader.next();
			} else if (element.equals("Print")) {
				xmlReader.next();
				this.print = Boolean.parseBoolean(xmlReader.getText());
				xmlReader.next();
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();
	}

	/**
	 * Write object to xml-file
	 * 
	 * @param xmlWriter
	 *            XMLWriter to write xml-file
	 * @throws Exception
	 */
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {
		xmlWriter.writeStartElement("Parameter");
		xmlWriter.writeAttribute("object", this.getClass().getSimpleName());
		xmlWriter.writeAttribute("name", this.name);

		// ID
		xmlWriter.writeStartElement("PID");
		if (this.pId != null) {
			xmlWriter.writeCharacters(this.pId);
		}
		xmlWriter.writeEndElement();

		// Name
		xmlWriter.writeStartElement("Name");
		if (this.name != null) {
			xmlWriter.writeCharacters(this.name);
		}
		xmlWriter.writeEndElement();

		// Description
		xmlWriter.writeStartElement("Description");
		if (this.description != null) {
			xmlWriter.writeCharacters(this.description);
		}
		xmlWriter.writeEndElement();

		// Comment
		xmlWriter.writeStartElement("Comment");
		if (this.comment != null) {
			xmlWriter.writeCharacters(this.comment);
		}
		xmlWriter.writeEndElement();

		// Constant
		xmlWriter.writeStartElement("Constant");
		xmlWriter.writeCharacters(Boolean.toString(this.constant));
		xmlWriter.writeEndElement();

		// Print
		xmlWriter.writeStartElement("Print");
		xmlWriter.writeCharacters(Boolean.toString(this.print));
		xmlWriter.writeEndElement();

		xmlWriter.writeEndElement();
	}

	/**
	 * Get temporary data
	 * 
	 * @param key
	 *            Key of data
	 * @return Data
	 */
	public Object getTmpData(String key) {
		Object obj = null;

		if (this.tmpData != null) {
			obj = this.tmpData.get(key);
		}

		return obj;
	}

	/**
	 * Set temporary data
	 * 
	 * @param key
	 *            Key of data
	 * @param obj
	 *            Data
	 */
	public void setTmpData(String key, Object obj) {
		if (this.tmpData == null) {
			this.tmpData = new Hashtable<String, Object>();
		}
		this.tmpData.put(key, obj);
	}

	/**
	 * Remove temporary data
	 * 
	 * @param key
	 *            Key of data
	 */
	public void rmTmpData(String key) {
		if (this.tmpData != null) {
			this.tmpData.remove(key);
			if (this.tmpData.isEmpty()) {
				this.tmpData = null;
			}
		}
	}

	/**
	 * Clear temporary data
	 */
	public void clearTmpData() {
		if (this.tmpData != null) {
			this.tmpData = null;
		}
	}

	/**
	 * Get temporary dataset data
	 * 
	 * @param key
	 *            Key of data
	 * @return Data
	 */
	public Object getTmpDSData(String key) {
		Dataset ds;
		Object obj = null;

		ds = this.getDataset();
		if (ds != null) {
			obj = ds.getTmpData(this.getPId(), key);
		}

		return obj;
	}

	/**
	 * Set temporary dataset data
	 * 
	 * @param key
	 *            Key of data
	 * @param obj
	 *            Data
	 */
	public void setTmpDSData(String key, Object obj) {
		Dataset ds;

		ds = this.getDataset();
		if (ds != null) {
			ds.setTmpData(this.getPId(), key, obj);
		}
	}

	/**
	 * Remove temporary dataset data
	 * 
	 * @param key
	 *            Key of data
	 */
	public void rmTmpDSData(String key) {
		Dataset ds;

		ds = this.getDataset();
		if (ds != null) {
			ds.rmTmpData(this.getPId(), key);
		}
	}

	/**
	 * Clear temporary dataset data
	 */
	public void clearTmpDSData() {
		Dataset ds;

		ds = this.getDataset();
		if (ds != null) {
			ds.clearTmpData(this.getPId());
		}
	}

	/**
	 * Report table overview
	 * 
	 * @param content
	 *            tex-writer
	 * @param indent
	 *            indent for name
	 */
	void reportTable(DocTexWriter content, int indent) {

		content.add("\\hspace{" + (3 * indent) + "ex}");
		content.add("\\hyperlink{");
		content.addName("D_" + this.getIdentifier());
		content.add("}{\\hypertarget{");
		content.addName("T_" + this.getIdentifier());
		content.add("}{");
		content.add("\\moparam{");
		// FIXME tex Color
		if (!this.getUsed()) {
			content.add("\\textcolor{gray}{");
		}
		content.addString(this.getName());
		if (!this.getUsed()) {
			content.add("}");
		}
		content.add("}");
		content.add("}} & ");

		if (!this.getUsed()) {
			content.add("\\textcolor{gray}{");
		}
		content.addString(this.getDescription());
		if (!this.getUsed()) {
			content.add("}");
		}
		content.add(" & ");
	}

	/**
	 * Report detail-description
	 * 
	 * @param content
	 *            tex-writer
	 * @param struct
	 *            Report-structure
	 * @throws IOException
	 */
	void reportDetail(DocTexWriter content, ReportStructure struct) throws IOException {
		DocTexWriter localContent;
		DocFile[] files;
		DocFile file;

		localContent = new DocTexWriter(content.getFile());

		// template and code
		if (struct.getTemplate()) {
			file = this.getDocFile("Template");
			if (file != null) {
				localContent.addTex(file);
			}

			if (struct.getDatasheet()) {
				file = this.getDocFile("Datasheet");
				if (file != null) {
					localContent.add("\\vspace{2ex}\\hyperlink{");
					localContent.addName("DS_" + this.getIdentifier());
					localContent.add("}{Data-sheet: ");
					localContent.addString(this.getName());
					localContent.add("}\\vspace{2ex}\n");
				}
			}
		}

		// parameter info
		if (struct.getInfo()) {
			localContent.add("\n\\begin{tabular}{m{3cm} m{10cm}}\n");
			this.reportInfo(localContent);
			localContent.add("\\end{tabular}\n\n");
		}

		// imports/exports
		if (struct.getImExport()) {
			this.reportDepend(localContent);
		}
		// usedBy/setby
		if (struct.getUsedSet()) {
			if (this.usedBy.size() != 0) {
				localContent.add("\\subsection*{Used}\n");
				this.reportUsedSet(localContent, this.usedBy);
				localContent.add("\n");
			}
			if (this.setBy.size() != 0) {
				localContent.add("\\subsection*{Set by}\n");
				this.reportUsedSet(localContent, this.setBy);
				localContent.add("\n");
			}
		}

		// DocFiles
		if (struct.getFiles()) {
			files = this.getDocFiles();
			if (files.length > 0) {
				localContent.add("\\subsection*{Files}\n");
				for (DocFile f : files) {
					localContent.add("\t");
					localContent.add("\\begin{figure}[H]");
					localContent.add("\\centering");
					localContent.addGraphic(f, 15);
					localContent.add("\\caption{");
					localContent.addString(f.getDescription());
					localContent.add(".}");
					localContent.add("\\end{figure} ");
					localContent.add("\n");
				}
			}
		}

		// add local content to content
		if (localContent.length() > 0) {
			// FIXME link to table
			// content.add("\\hyperlink{");
			// content.addName("T_" + this.getIdentifier());
			// content.add("}{");
			content.add("\\hypertarget{");
			content.addName("D_" + this.getIdentifier());
			content.add("}{");
			content.add("\\section*{");
			content.addString(this.getDescription());
			content.add(" - \\moparam{");
			content.addString(this.getName());
			content.add("}}}");
			// content.add("}");
			content.add("\n");
			content.add("\\ifdefined\\usehtml\\else\n");
			content.add("\t\\addcontentsline{toc}{section}{\\moparam{");
			content.addString(this.getName());
			content.add("} - ");
			content.addString(this.getDescription());
			content.add("}\n");
			content.add("\\fi\n\n");

			content.add("\\index{");
			content.addString(this.getName().toUpperCase());
			content.add("@\\moparam{");
			content.addString(this.getName());
			content.add("}}\n");

			content.addWriter(localContent);

			if (struct.getFunction()) {
				if (this instanceof ParamFunction) {
					ParamFunction func = (ParamFunction) this;
					content.add("\\vspace{2ex}\\hyperlink{");
					content.addName("F_" + func.getClassName());
					content.add("}{Function: ");
					content.addString(func.getClassName());
					content.add("}\n");
				}
			}

			if (struct.getSummary()) {
				content.add("\n");
				content.add("\\vspace{2ex}\\hyperlink{");
				content.addName("T_" + this.getIdentifier());
				content.add("}{back}\n\n");
			}

		}
	}

	/**
	 * Report overview-information for detail-description
	 * 
	 * @param content
	 *            tex-writer
	 */
	void reportInfo(DocTexWriter content) {

		content.add("\tName & ");
		content.addString(this.getName());
		content.add(" \\\\\n");

		content.add("\tIdentifier & ");
		content.addString(this.getIdentifier());
		content.add(" \\\\\n");

		content.add("\tDescription & ");
		content.addString(this.getDescription());
		content.add(" \\\\\n");
	}

	/**
	 * Report dependencies infos
	 * 
	 * @param content
	 *            Tex-writer
	 */
	void reportDepend(DocTexWriter content) {

	}

	/**
	 * Report used-by information
	 * 
	 * @param content
	 *            tex-writer
	 * @param items
	 */
	protected void reportUsedSet(DocTexWriter content, AbstractList<Parameter> items) {
		Parameter item;

		content.add("\\begin{tabular}{m{13cm}}\n");
		for (int k = 0; k < items.size(); k++) {
			item = items.get(k);

			content.add("\\hyperlink{");
			content.addName("D_" + item.getIdentifier());
			content.add("}{");
			content.addString(item.getIdentifier());
			content.add("} \\\\\n");
		}
		content.add("\\end{tabular}\n");
	}

	/**
	 * Function for generating documentation-files
	 * 
	 * @param directory
	 *            Directory for export-files
	 * @throws Exception
	 */
	protected void generateDocFiles(String directory) throws Exception {
		this.generateDocFiles(new File(directory));
	}

	/**
	 * Function for generating documentation-files
	 * 
	 * @param directory
	 *            Directory for export-files
	 * @throws Exception
	 */
	protected void generateDocFiles(File directory) throws Exception {
		this.deleteDocFiles();
	}

	/**
	 * Get Parameter-data-structure
	 * 
	 * @return Parameter-data-structure
	 */
	protected ParamDataStruct getParameterStruct() {
		ParamDataStruct struct = null;

		if (this.getParent() != null) {
			struct = this.getParent().getParameterStruct();
		}

		return struct;
	}

	/**
	 * Get index of current dataset
	 * 
	 * @return Current index
	 */
	protected Dataset getDataset() {
		ParamDataStruct pds;
		Dataset ds = null;

		pds = this.getParameterStruct();
		if (pds != null) {
			ds = pds.getDataset();
		}

		return ds;
	}

	/**
	 * Get index of current dataset
	 * 
	 * @param index
	 *            Index of dataset
	 * @return Current index
	 */
	Dataset getDataset(int index) {
		ParamDataStruct pds;
		Dataset ds = null;

		pds = this.getParameterStruct();
		if (pds != null) {
			ds = pds.getDataset(index);
		}

		return ds;
	}

	@Override
	public int getLength() {
		ParamDataStruct pds;
		int len = 0;

		pds = this.getParameterStruct();
		if (pds != null) {
			len = pds.getLength();
		}

		return len;
	}

	@Override
	public String getDsId() {
		Dataset ds;
		String dsid = null;

		ds = this.getDataset();
		if (ds != null) {
			dsid = ds.getDsId();
		}
		return dsid;
	}

	/**
	 * Get identifier of datasets
	 * 
	 * @param index
	 *            Index of dataset
	 * @return Dataset-identifier
	 */
	public String getDsId(int index) {
		Dataset ds;
		String dsid = null;

		ds = this.getDataset(index);
		if (ds != null) {
			dsid = ds.getDsId();
		}
		return dsid;
	}

	@Override
	public int getIndex() {
		ParamDataStruct pds;
		int idx = -1;

		pds = this.getParameterStruct();
		if (pds != null) {
			idx = pds.getIndex();
		}

		return idx;
	}

	@Override
	public void setIndex(int index) {
		if (this.parent != null) {
			this.parent.setIndex(index);
		}
	}

	@Override
	public File getParamDirectory() {
		return this.getParamDirectory(this.getIndex());
	}

	/**
	 * Get dataset-directory of parameter
	 * 
	 * @param index
	 *            Index of dataset
	 * @return Parameter-directory
	 */
	public File getParamDirectory(int index) {
		File dir = null;
		Parameter root;

		root = this.getRoot();
		if (root instanceof Project) {
			dir = FileUtils.getFile(((Project) root).getFilesDirectory(index).getPath(), "param", this.getPId());
			if (!dir.exists()) {
				dir.mkdirs();
			}
		}

		return dir;
	}

	/**
	 * List key of parameter
	 * 
	 * @return List parameter key
	 */
	List<ParamSetKey> listParamSetKey() {
		List<ParamSetKey> list;

		list = new LinkedList<ParamSetKey>();
		list.add(ParamSetKey.MODIFIED);
		list.add(ParamSetKey.UNCERTAIN);
		list.add(ParamSetKey.ERROR);

		return list;
	}

	@Override
	public String getUniqueDocFileName(String file) {
		return this.getUniqueDocFileName(new File(file)).getPath();
	}

	@Override
	public File getUniqueDocFileName(File file) {
		File path, uniqueFile;
		String name;
		String[] fileparts, parts;
		int i = 0;
		boolean exists = true;

		if (this.getParent() != null) {
			uniqueFile = this.getParent().getUniqueDocFileName(file);
		} else {
			path = file.getParentFile();
			fileparts = FileUtils.splitName(file.getName());
			name = fileparts[0];
			if (path.exists()) {
				while (exists) {
					exists = false;
					for (File f : path.listFiles()) {
						parts = FileUtils.splitName(f.getName());
						exists |= (name.compareToIgnoreCase(parts[0]) == 0);
					}
					if (exists) {
						i++;
						name = fileparts[0] + "(" + i + ")";
					}
				}
			}
			uniqueFile = new File(path, name + fileparts[1]);
		}

		return uniqueFile;
	}

	/**
	 * Get RMI-adapter of parameter
	 * 
	 * @return RMI-adapter
	 * @throws RemoteException
	 */
	public abstract IParameter getRMIImpl() throws RemoteException;

	@Override
	public String toStringI() throws RemoteException {
		return this.toString();
	}

}
