/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.converter;

import at.lcm.bu21.general.utils.XMLNode;

/**
 * Convert to V_0_27.
 * 
 * - Partition of structure and data
 * 
 * @author Werner Koppelsttter
 */
public class Converter_27 {

	/**
	 * Convert the given node
	 * 
	 * @param node
	 *            Data to convert
	 * @return Changed data node
	 */
	static XMLNode convert(XMLNode node) {

		// Convert Project
		Converter_27.convertProject(node);

		return node;
	}

	/**
	 * Convert Project
	 * 
	 * @param node
	 *            Node to update
	 */
	private static void convertProject(XMLNode node) {
		XMLNode paramDataStruct, datasets, dsid, comment, log, ds, item;

		// create ParamDataStruct
		paramDataStruct = new XMLNode("ParamDataStruct");
		node.addNode(paramDataStruct);

		// remove length
		node.rmNode(0);
		// move CurIndex
		paramDataStruct.addNode(node.rmNode(0));
		// move NextPID
		paramDataStruct.addNode(node.rmNode(0));
		// move NextDSID
		paramDataStruct.addNode(node.rmNode(0));

		// move DSID (names of datasets)
		dsid = node.rmNode(2);
		paramDataStruct.addNode(dsid);
		// remove dataset comments
		comment = node.rmNode(2);
		// remove eventlogs
		log = node.rmNode(2);

		// create dataset-list
		datasets = new XMLNode("DS");
		paramDataStruct.addNode(datasets);

		for (int k = 0; k < dsid.getNodeSize(); k++) {
			// create dataset
			ds = new XMLNode("item");
			datasets.addNode(ds);

			// dataset-id
			item = new XMLNode("DSID");
			item.setText(dsid.getNode(k).getText());
			ds.addNode(item);

			// move comment
			item = comment.getNode(k);
			item.setName("Comment");
			ds.addNode(item);

			// ParamData
			item = new XMLNode("ParamData");
			ds.addNode(item);

			// move EventLog
			item = log.getNode(k);
			item.setName("EventLog");
			ds.addNode(item);

		}

		// convert ParamContainer
		Converter_27.convertContainer(node.getNode(0), datasets);
	}

	/**
	 * Convert ParamContainer
	 * 
	 * @param node
	 *            Node of container
	 * @param datasets
	 *            List of datasets
	 * @return Pid of container
	 */
	private static String convertContainer(XMLNode node, XMLNode datasets) {
		String pid;
		XMLNode list, item;

		// convert Parameter
		pid = Converter_27.convertParameter(node.getNode(0), datasets);

		// Parameter-list
		for (XMLNode n : node.getNode(1).getNodes()) {
			if (n.getName().compareTo("ParamContainer") == 0) {
				Converter_27.convertContainer(n, datasets);
			} else if (n.getName().compareTo("ParamMatlabFunction") == 0) {
				Converter_27.convertMatlabFunction(n, datasets);
			} else if (n.getName().compareTo("ParamJavaFunction") == 0) {
				Converter_27.convertJavaFunction(n, datasets);
			} else if (n.getName().compareTo("ParamField") == 0) {
				Converter_27.convertField(n, datasets);
			} else if (n.getName().compareTo("ParamComboBox") == 0) {
				Converter_27.convertComboBox(n, datasets);
			} else if (n.getName().compareTo("ParamFormula") == 0) {
				Converter_27.convertFormula(n, datasets);
			} else if (n.getName().compareTo("ParamMaterial") == 0) {
				Converter_27.convertMaterial(n, datasets);
			} else if (n.getName().compareTo("ParamDirectory") == 0) {
				Converter_27.convertDirectory(n, datasets, false);
			} else if (n.getName().compareTo("ParamFile") == 0) {
				Converter_27.convertFile(n, datasets);
			}
		}

		// Modified
		list = node.rmNode(2);
		for (int k = 0; k < list.getNodeSize(); k++) {
			item = list.getNode(k);
			item.setName("Boolean");
			Converter_27.addData(datasets, k, pid, "INCOMPLETE", item);
		}

		return pid;
	}

	/**
	 * Convert Parameter
	 * 
	 * @param node
	 *            Node of parameter
	 * @param datasets
	 *            List of datasets
	 * @return Pid of parameter
	 */
	private static String convertParameter(XMLNode node, XMLNode datasets) {
		String pid;
		XMLNode list, item;

		pid = node.getNode(0).getText();

		// Modified
		list = node.rmNode(6);
		for (int k = 0; k < list.getNodeSize(); k++) {
			item = list.getNode(k);
			item.setName("Boolean");
			Converter_27.addData(datasets, k, pid, "MODIFIED", item);
		}

		// Uncertain
		list = node.rmNode(6);
		for (int k = 0; k < list.getNodeSize(); k++) {
			item = list.getNode(k);
			item.setName("Boolean");
			Converter_27.addData(datasets, k, pid, "UNCERTAIN", item);
		}

		// Error
		list = node.rmNode(6);
		for (int k = 0; k < list.getNodeSize(); k++) {
			item = list.getNode(k);
			item.setName("Boolean");
			Converter_27.addData(datasets, k, pid, "ERROR", item);
		}

		// DocFiles
		list = node.rmNode(6);
		for (int k = 0; k < list.getNodeSize(); k++) {
			item = list.getNode(k);
			Converter_27.addDocFiles(datasets, k, pid, item);
		}

		return pid;
	}

	/**
	 * Convert ParamMatlabFunction
	 * 
	 * @param node
	 *            Node of matlab-function
	 * @param datasets
	 *            List of datasets
	 * @return Pid of matalab-function
	 */
	private static String convertMatlabFunction(XMLNode node, XMLNode datasets) {
		String pid;

		// convert ParamFunction
		pid = Converter_27.convertFunction(node.getNode(0), datasets);

		return pid;
	}

	/**
	 * Convert ParamJavaFunction
	 * 
	 * @param node
	 *            Node of java-function
	 * @param datasets
	 *            List of datasets
	 * @return Pid of java-function
	 */
	private static String convertJavaFunction(XMLNode node, XMLNode datasets) {
		String pid;

		// convert ParamFunction
		pid = Converter_27.convertFunction(node.getNode(0), datasets);

		return pid;
	}

	/**
	 * Convert ParamFunction
	 * 
	 * @param node
	 *            Node of function
	 * @param datasets
	 *            List of datasets
	 * @return Pid of function
	 */
	private static String convertFunction(XMLNode node, XMLNode datasets) {
		String pid;
		XMLNode list, item;

		// convert ParamContainer
		pid = Converter_27.convertContainer(node.getNode(0), datasets);

		// Data
		list = node.rmNode(2);
		for (int k = 0; k < list.getNodeSize(); k++) {
			item = list.getNode(k);
			Converter_27.addData(datasets, k, pid, "DATA", item);
		}

		// UpdateRequired
		list = node.rmNode(2);
		for (int k = 0; k < list.getNodeSize(); k++) {
			item = list.getNode(k);
			item.setName("Boolean");
			Converter_27.addData(datasets, k, pid, "UPDATEREQUIRED", item);
		}

		return pid;
	}

	/**
	 * Convert ParamField
	 * 
	 * @param node
	 *            Node of field
	 * @param datasets
	 *            List of datasets
	 * @return Pid of field
	 */
	private static String convertField(XMLNode node, XMLNode datasets) {
		String pid;
		XMLNode list, item;
		boolean constant;

		// convert Parameter
		pid = Converter_27.convertParameter(node.getNode(0), datasets);

		constant = node.getNode(0).getNode(4).getText().compareTo("true") == 0;

		// Data
		list = node.rmNode(2);
		if (!constant) {
			for (int k = 0; k < list.getNodeSize(); k++) {
				item = list.getNode(k);
				Converter_27.addData(datasets, k, pid, "DATA", item);
			}
			list.clearNodes();
		}
		node.addNode(1, list);

		return pid;
	}

	/**
	 * Convert ParamComboBox
	 * 
	 * @param node
	 *            Node of combobox
	 * @param datasets
	 *            List of datasets
	 * @return Pid of combobox
	 */
	private static String convertComboBox(XMLNode node, XMLNode datasets) {
		String pid;

		// convert ParamField
		pid = Converter_27.convertField(node.getNode(0), datasets);

		return pid;
	}

	/**
	 * Convert ParamFormula
	 * 
	 * @param node
	 *            Node of formula
	 * @param datasets
	 *            List of datasets
	 * @return Pid of formula
	 */
	private static String convertFormula(XMLNode node, XMLNode datasets) {
		String pid;
		XMLNode list, item;

		// convert ParamField
		pid = Converter_27.convertField(node.getNode(0), datasets);

		// UpdateRequired
		list = node.rmNode(3);
		for (int k = 0; k < list.getNodeSize(); k++) {
			item = list.getNode(k);
			item.setName("Boolean");
			Converter_27.addData(datasets, k, pid, "UPDATEREQUIRED", item);
		}

		return pid;
	}

	/**
	 * Convert ParamMaterial
	 * 
	 * @param node
	 *            Node of material
	 * @param datasets
	 *            List of datasets
	 * @return Pid of material
	 */
	private static String convertMaterial(XMLNode node, XMLNode datasets) {
		String pid;

		// convert ParamContainer
		pid = Converter_27.convertContainer(node.getNode(0), datasets);

		return pid;
	}

	/**
	 * Convert ParamDirectory
	 * 
	 * @param node
	 *            Node of directory
	 * @param datasets
	 *            List of datasets
	 * @param constant
	 *            Is directory constant (e.g. ParamFile)
	 * @return Pid of directory
	 */
	private static String convertDirectory(XMLNode node, XMLNode datasets,
			boolean constant) {
		String pid;
		XMLNode list, item;

		// convert Parameter
		pid = Converter_27.convertParameter(node.getNode(0), datasets);

		constant |= node.getNode(0).getNode(4).getText().compareTo("true") == 0;

		// Directory
		list = node.getNode(2);
		if (!constant) {
			for (int k = 0; k < list.getNodeSize(); k++) {
				item = list.getNode(k);
				item.setName("String");
				Converter_27.addData(datasets, k, pid, "DIRECTORY", item);
			}
			list.clearNodes();
		} else {
			item = list.getNode(0);
			list.clearNodes();
			if (item.getText() != null) {
				list.setText(item.getText());
			}
		}

		// Extension
		list = node.rmNode(3);
		if (!constant) {
			for (int k = 0; k < list.getNodeSize(); k++) {
				item = list.getNode(k);
				item.setName("String");
				Converter_27.addData(datasets, k, pid, "EXTENSION", item);
			}
		}

		return pid;
	}

	/**
	 * Convert ParamFile
	 * 
	 * @param node
	 *            Node of file
	 * @param datasets
	 *            List of datasets
	 * @return Pid of file
	 */
	private static String convertFile(XMLNode node, XMLNode datasets) {
		String pid;
		XMLNode list, item;
		boolean constant;

		// convert ParamDirectory
		pid = Converter_27.convertDirectory(node.getNode(0), datasets, true);

		constant = node.getNode(0).getNode(0).getNode(4).getText()
				.compareTo("true") == 0;

		// File
		list = node.getNode(1);
		if (!constant) {
			for (int k = 0; k < list.getNodeSize(); k++) {
				item = list.getNode(k);
				item.setName("String");
				Converter_27.addData(datasets, k, pid, "FILE", item);
			}
			list.clearNodes();
		} else {
			item = list.getNode(0);
			list.clearNodes();
			if (item.getText() != null) {
				list.setText(item.getText());
			}
		}

		// CheckSum
		list = node.getNode(2);
		if (!constant) {
			for (int k = 0; k < list.getNodeSize(); k++) {
				item = list.getNode(k);
				item.setName("String");
				Converter_27.addData(datasets, k, pid, "CHECKSUM", item);
			}
		} else {
			item = list.getNode(0);
			list.clearNodes();
			if (item.getText() != null) {
				list.setText(item.getText());
			}
		}

		return pid;
	}

	/**
	 * Add data of parameter-identifier
	 * 
	 * @param datasets
	 *            List of datasets
	 * @param dsNum
	 *            Number of dataset
	 * @param pid
	 *            Parameter-identifier
	 * @param key
	 * @param data
	 */
	private static void addData(XMLNode datasets, int dsNum, String pid,
			String key, XMLNode data) {
		XMLNode paramset, item;

		paramset = Converter_27.getParamSet(datasets, dsNum, pid);
		item = new XMLNode(key);
		item.addNode(data);
		paramset.getNode(0).addNode(item);

	}

	/**
	 * Add DocFile of parameter-identifier
	 * 
	 * @param datasets
	 *            List of datasets
	 * @param dsNum
	 *            Number of dataset
	 * @param pid
	 *            Parameter-identifier
	 * @param docfiles
	 *            DocFile to add
	 */
	private static void addDocFiles(XMLNode datasets, int dsNum, String pid,
			XMLNode docfiles) {
		XMLNode paramset, node;

		paramset = Converter_27.getParamSet(datasets, dsNum, pid);
		node = paramset.getNode(1);
		for (XMLNode item : docfiles.getNodes()) {
			node.addNode(item);
		}

	}

	/**
	 * Get ParamSet for parameter-identifier
	 * 
	 * @param datasets
	 *            List of datasets
	 * @param dsNum
	 *            Number of dataset
	 * @param pid
	 *            Parameter-identifier
	 * @return ParamSet of parameter-identifier
	 */
	private static XMLNode getParamSet(XMLNode datasets, int dsNum, String pid) {
		XMLNode paramData, paramset = null;
		String val;

		// get list of paramsets
		paramData = datasets.getNode(dsNum).getNode(2);
		// search for existing paramset
		for (XMLNode p : paramData.getNodes()) {
			val = p.getAttribute("pid");
			if ((val != null) && (pid.compareTo(val) == 0)) {
				paramset = p;
			}
		}

		// if no paramset exists create a new and add
		if (paramset == null) {
			paramset = new XMLNode("ParamSet");
			paramset.setAttribute("pid", pid);
			paramset.addNode(new XMLNode("Data"));
			paramset.addNode(new XMLNode("DocFiles"));
			paramData.addNode(paramset);
		}

		return paramset;
	}
}
