/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.event;

import java.io.Serializable;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.LinkedList;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.utils.Clone;

/**
 * Event-base-object for information of current status
 * 
 * @author Werner Koppelsttter
 */
public class Event implements Serializable {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Date-formatter for reading and writing date */
	protected static SimpleDateFormat sdf = new SimpleDateFormat(
			"dd-MM-yyyy HH:mm");

	/** Type of event */
	private EventType type = EventType.INFO;

	/** Event message */
	private String message = new String();

	/** Date when event occur */
	private Date date = new Date();

	/** Identifier where the event occur */
	private String identifier = new String();

	/** Calling stack of event */
	private StackTraceElement[] stack = new StackTraceElement[0];

	/**
	 * Create empty object to loading
	 */
	Event() {
	}

	/**
	 * Create a event with given data and calling stack
	 * 
	 * @param type
	 *            Type of event
	 * @param identifier
	 *            Identifier where the event occurs
	 * @param message
	 *            Description of the Info
	 * @param stack
	 *            Calling stack
	 */
	public Event(EventType type, String identifier, String message,
			StackTraceElement[] stack) {
		this.type = type;
		if (identifier == null) {
			identifier = "";
		}
		if (message == null) {
			message = "";
		}
		if (stack == null) {
			stack = new StackTraceElement[0];
		}

		this.identifier = identifier;
		this.message = message;
		this.stack = stack;
	}

	/**
	 * Get type of event
	 * 
	 * @return Event-type
	 */
	public EventType getType() {
		return this.type;
	}

	/**
	 * Set type of event
	 * 
	 * @param type
	 *            Type of event
	 * 
	 */
	public void setType(EventType type) {
		this.type = type;
	}

	/**
	 * Create a deep clone of given data-structure
	 * 
	 * @return Deep clone of structure
	 * @throws Exception
	 */
	public Event deepClone() throws Exception {
		return (Event) Clone.deepClone(this);
	}

	/**
	 * Get Description message of event
	 * 
	 * @return Description of event
	 */
	public String getDescription() {
		return this.message;
	}

	/**
	 * Set Description message of event
	 * 
	 * @param message
	 *            Description of event
	 */
	public void setDescription(String message) {
		this.message = message;
	}

	/**
	 * Get date when event occur
	 * 
	 * @return Date of event
	 */
	public Date getDate() {
		return this.date;
	}

	/**
	 * Get date-string when event occur
	 * 
	 * @return date-string of event
	 */
	public String getDateString() {
		return Event.sdf.format(this.date);
	}

	/**
	 * Get identifier where the event occur
	 * 
	 * @return Identifier of event
	 */
	public String getIdentifier() {
		return this.identifier;
	}

	/**
	 * Get calling stack of the event
	 * 
	 * @return calling stack-array
	 */
	public StackTraceElement[] getStack() {
		return this.stack;
	}

	/**
	 * Create a string representation
	 * 
	 * @return String representation
	 */
	@Override
	public String toString() {
		return String.format("%1$-8s %2$-8s %3$s", this.type,
				this.getDateString(), this.getDescription());

	}

	/**
	 * Create a full string representation
	 * 
	 * @return String representation
	 */
	public String toFullString() {
		StringBuffer str = new StringBuffer();
		str.append(this);

		for (int i = 0; i < this.stack.length; i++) {
			str.append("\n    ");
			str.append(this.stack[i]);
		}

		return str.toString();
	}

	/**
	 * Read object from xml-file
	 * 
	 * @param xmlReader
	 *            XMLReader to read xml-file
	 * @throws Exception
	 */
	public void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element, tmp;
		String filename, classname, methodename;
		int line;
		LinkedList<StackTraceElement> stack = new LinkedList<StackTraceElement>();

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {

			element = xmlReader.getLocalName();
			if (element.equals("Type")) {
				xmlReader.next();
				tmp = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					tmp += xmlReader.getText();
					xmlReader.next();
				}
				this.type = EventType.valueOf(tmp);
				xmlReader.next();
			} else if (element.equals("Message")) {
				xmlReader.next();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.message += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("Date")) {
				xmlReader.next();
				tmp = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					tmp += xmlReader.getText();
					xmlReader.next();
				}
				this.date = Event.sdf.parse(tmp);
				xmlReader.next();
			} else if (element.equals("Identifier")) {
				xmlReader.next();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.identifier += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("Stack")) {
				xmlReader.next();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					xmlReader.next();

					// file
					xmlReader.next();
					filename = new String();
					while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
						filename += xmlReader.getText();
						xmlReader.next();
					}
					xmlReader.next();

					// class
					xmlReader.next();
					classname = new String();
					while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
						classname += xmlReader.getText();
						xmlReader.next();
					}
					xmlReader.next();

					// method
					xmlReader.next();
					methodename = new String();
					while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
						methodename += xmlReader.getText();
						xmlReader.next();
					}
					xmlReader.next();

					// line
					xmlReader.next();
					line = Integer.parseInt(xmlReader.getText());
					xmlReader.next();
					xmlReader.next();

					xmlReader.next();
					stack.add(new StackTraceElement(classname, methodename,
							filename, line));
				}
				this.stack = stack.toArray(new StackTraceElement[0]);
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: "
						+ xmlReader.getLocalName());
			}
		}
		xmlReader.next();
	}

	/**
	 * Write object to xml-file
	 * 
	 * @param xmlWriter
	 *            XMLWriter to write xml-file
	 * @throws Exception
	 */
	public void obj2xml(XMLStreamWriter xmlWriter) throws Exception {
		StackTraceElement element;

		xmlWriter.writeStartElement("Event");

		// Type
		xmlWriter.writeStartElement("Type");
		xmlWriter.writeCharacters(this.type.name());
		xmlWriter.writeEndElement();

		// DescrptionStr
		xmlWriter.writeStartElement("Message");
		xmlWriter.writeCharacters(this.message);
		xmlWriter.writeEndElement();

		// Date
		xmlWriter.writeStartElement("Date");
		xmlWriter.writeCharacters(Event.sdf.format(this.date));
		xmlWriter.writeEndElement();

		// Identifier
		xmlWriter.writeStartElement("Identifier");
		xmlWriter.writeCharacters(this.identifier);
		xmlWriter.writeEndElement();

		// Stack
		xmlWriter.writeStartElement("Stack");
		if (this.stack != null) {
			for (int i = 0; i < this.stack.length; i++) {
				element = this.stack[i];
				xmlWriter.writeStartElement("element");
				// file
				xmlWriter.writeStartElement("File");
				if (element.getFileName() != null) {
					xmlWriter.writeCharacters(element.getFileName());
				}
				xmlWriter.writeEndElement();

				// class
				xmlWriter.writeStartElement("Class");
				xmlWriter.writeCharacters(element.getClassName());
				xmlWriter.writeEndElement();

				// method
				xmlWriter.writeStartElement("Methode");
				xmlWriter.writeCharacters(element.getMethodName());
				xmlWriter.writeEndElement();

				// Line
				xmlWriter.writeStartElement("Line");
				xmlWriter.writeCharacters(Integer.toString(element
						.getLineNumber()));
				xmlWriter.writeEndElement();
				xmlWriter.writeEndElement();
			}
		}
		xmlWriter.writeEndElement();

		xmlWriter.writeEndElement();
	}

}
