/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.formula;

import java.io.Reader;
import java.io.StringReader;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TDataException;

/**
 * Parse syntax of formula (syntax analyze)
 *
 * @author Werner Koppelsttter
 */
public class Parser {

	/** List of parsed variables */
	protected List<String> vars;

	/** Lexical scanner */
	protected Scanner scan;

	/** Parsed syntax tree */
	protected SyntaxNode syntax;

	/** Scanned token */
	protected Token t;

	/** Look-ahead token */
	protected Token la;

	/** Kind of look-ahead token */
	protected TokenKind sym;

	/**
	 * Initialize Parser with formula
	 * 
	 * @param formula
	 *            Formula to parse
	 * 
	 * @throws ScannerException
	 * @throws ParserException
	 */
	public Parser(String formula) throws ParserException, ScannerException {
		this(new StringReader(formula));
	}

	/**
	 * Initialize Parser with reader of formula
	 * 
	 * @param reader
	 *            Reader with formula
	 * 
	 * @throws ScannerException
	 * @throws ParserException
	 */
	public Parser(Reader reader) throws ParserException, ScannerException {
		this.vars = new LinkedList<String>();

		// lexical analyze
		this.scan = new Scanner(reader);
		// syntax analyze
		this.parse();
	}

	/**
	 * Get variables of formula
	 * 
	 * @return List of variables
	 */
	public String[] getVariables() {
		return this.vars.toArray(new String[0]);
	}

	/**
	 * Parse formula to create syntax-tree and list variables
	 * 
	 * @throws ParserException
	 * @throws ScannerException
	 */
	private void parse() throws ParserException, ScannerException {
		// syntax analyzer
		this.scan();
		this.syntax = this.Statement();
		this.check(TokenKind.EOF);
	}

	/**
	 * Evaluate formula
	 * 
	 * @param vars
	 *            Map with variables
	 * 
	 * @return Evaluated data
	 * 
	 * @throws EvalException
	 * @throws TDataException
	 * @throws UserException
	 */
	public TData eval(Map<String, TData> vars) throws EvalException, UserException, TDataException {
		return this.syntax.eval(vars);
	}

	/**
	 * Scan next token
	 * 
	 * @throws ScannerException
	 */
	private void scan() throws ScannerException {
		this.t = this.la;
		this.la = this.scan.next();
		this.sym = this.la.kind;
	}

	/**
	 * Check if next token is the expected
	 * 
	 * @param expected
	 *            Expected token
	 * @throws ScannerException
	 * @throws ParserException
	 */
	private void check(TokenKind expected) throws ScannerException, ParserException {
		if (this.sym == expected) {
			this.scan();
		} else {
			throw new ParserException(expected.getDescription() + " was expected", this.la);
		}
	}

	/**
	 * Statement = Condition.
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ParserException
	 * @throws ScannerException
	 */
	private SyntaxNode Statement() throws ParserException, ScannerException {
		SyntaxNode sn;

		// switch (Parser.sym) {
		// default:
		sn = this.Condition();
		// break;
		// }

		return sn;
	}

	/**
	 * Condition = CondTerm {"||" CondTerm}.
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ScannerException
	 * @throws ParserException
	 */
	private SyntaxNode Condition() throws ParserException, ScannerException {
		SyntaxNode sn, tmp;

		sn = this.CondTerm();
		while (this.sym == TokenKind.OR) {
			tmp = new SyntaxNode(this.la);
			this.scan();
			tmp.addChild(sn);
			tmp.addChild(this.CondTerm());
			sn = tmp;
		}

		return sn;
	}

	/**
	 * CondTerm = CondFact {"&&" CondFact}.
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ScannerException
	 * @throws ParserException
	 */
	private SyntaxNode CondTerm() throws ParserException, ScannerException {
		SyntaxNode sn, tmp;

		sn = this.CondFact();
		while (this.sym == TokenKind.AND) {
			tmp = new SyntaxNode(this.la);
			this.scan();
			tmp.addChild(sn);
			tmp.addChild(this.CondFact());
			sn = tmp;
		}

		return sn;
	}

	/**
	 * CondFact = CondValue [("==" | "!=" | ">" | ">=" | "<" | "<=") CondValue].
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ScannerException
	 * @throws ParserException
	 */
	private SyntaxNode CondFact() throws ParserException, ScannerException {
		SyntaxNode sn, tmp;

		sn = this.CondValue();
		while ((this.sym == TokenKind.EQL) || (this.sym == TokenKind.NEQ) || (this.sym == TokenKind.GTR)
				|| (this.sym == TokenKind.GEQ) || (this.sym == TokenKind.LSS) || (this.sym == TokenKind.LEQ)) {
			tmp = new SyntaxNode(this.la);
			this.scan();
			tmp.addChild(sn);
			tmp.addChild(this.CondValue());
			sn = tmp;
		}

		return sn;
	}

	/**
	 * CondValue = ["!"] Expr.
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ScannerException
	 * @throws ParserException
	 */
	private SyntaxNode CondValue() throws ParserException, ScannerException {
		SyntaxNode sn;

		if (this.sym == TokenKind.CALL) {
			sn = new SyntaxNode(this.la);
			this.scan();
			sn.addChild(this.Expr());
		} else {
			sn = this.Expr();
		}

		return sn;
	}

	/**
	 * Expr = ["+" | "-"] Term {("+" | "-") Term}.
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ScannerException
	 * @throws ParserException
	 */
	private SyntaxNode Expr() throws ParserException, ScannerException {
		SyntaxNode sn, tmp, minus = null;

		if (this.sym == TokenKind.PLUS) {
			this.scan();
		} else if (this.sym == TokenKind.MINUS) {
			minus = new SyntaxNode(this.la);
			this.scan();
		}

		sn = this.Term();
		if (minus != null) {
			minus.addChild(sn);
			sn = minus;
		}

		while ((this.sym == TokenKind.PLUS) || (this.sym == TokenKind.MINUS)) {
			tmp = new SyntaxNode(this.la);
			this.scan();
			tmp.addChild(sn);
			tmp.addChild(this.Term());
			sn = tmp;
		}

		return sn;
	}

	/**
	 * Term = Factor [("*" | ".*" | "/" | "./" | "%") Factor].
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ScannerException
	 * @throws ParserException
	 */
	private SyntaxNode Term() throws ParserException, ScannerException {
		SyntaxNode sn, tmp;

		sn = this.Factor();
		while ((this.sym == TokenKind.TIMES) || (this.sym == TokenKind.DOTTIMES) || (this.sym == TokenKind.SLASH)
				|| (this.sym == TokenKind.DOTSLASH) || (this.sym == TokenKind.PERCENT)
				|| (this.sym == TokenKind.DOTPERCENT)) {
			tmp = new SyntaxNode(this.la);
			this.scan();
			tmp.addChild(sn);
			tmp.addChild(this.Factor());
			sn = tmp;
		}

		return sn;
	}

	/**
	 * Factor = Modifier [("^" | ".^") Modifier].
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ScannerException
	 * @throws ParserException
	 */
	private SyntaxNode Factor() throws ParserException, ScannerException {
		SyntaxNode sn, tmp;

		sn = this.Modifier();
		if ((this.sym == TokenKind.CARET) || (this.sym == TokenKind.DOTCARET)) {
			tmp = new SyntaxNode(this.la);
			this.scan();
			tmp.addChild(sn);
			tmp.addChild(this.Modifier());
			sn = tmp;
		}

		return sn;
	}

	/**
	 * Modifier = Value {ModNumeric | ModArray | ModStruct | "'"}.
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ParserException
	 * @throws ScannerException
	 */
	private SyntaxNode Modifier() throws ParserException, ScannerException {
		SyntaxNode sn, tmp;

		sn = this.Value();
		while (true) {
			if (this.sym == TokenKind.LBRACK) {
				tmp = this.ModNumeric();
			} else if (this.sym == TokenKind.LBRACE) {
				tmp = this.ModArray();
			} else if (this.sym == TokenKind.PERIOD) {
				tmp = this.ModStruct();
			} else if (this.sym == TokenKind.ACCENT) {
				tmp = new SyntaxNode(this.la);
				this.scan();
				tmp.addChild(sn);
			} else {
				break;
			}
			tmp.childs[0] = sn;
			sn = tmp;
		}

		return sn;
	}

	/**
	 * ModNumeric = [ Access [, Access] ].
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ScannerException
	 * @throws ParserException
	 */
	private SyntaxNode ModNumeric() throws ScannerException, ParserException {
		SyntaxNode sn;

		sn = new SyntaxNode(this.la);
		this.scan();
		sn.addChild(null);
		sn.addChild(this.Access());
		if (this.sym == TokenKind.COMMA) {
			this.scan();
			sn.addChild(this.Access());
		}
		this.check(TokenKind.RBRACK);

		return sn;
	}

	/**
	 * ModArray = { Expr [, Expr] }.
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ScannerException
	 * @throws ParserException
	 */
	private SyntaxNode ModArray() throws ScannerException, ParserException {
		SyntaxNode sn;

		sn = new SyntaxNode(this.la);
		this.scan();
		sn.addChild(null);
		sn.addChild(this.Expr());
		if (this.sym == TokenKind.COMMA) {
			this.scan();
			sn.addChild(this.Expr());
		}
		this.check(TokenKind.RBRACE);

		return sn;
	}

	/**
	 * ModStruct = "." ident ["(" [Statement {"," Statement}] ")"].
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ScannerException
	 * @throws ParserException
	 */
	private SyntaxNode ModStruct() throws ScannerException, ParserException {
		SyntaxNode sn;

		sn = new SyntaxNode(this.la);
		this.scan();
		sn.addChild(null);
		sn.addChild(new SyntaxNode(this.la));
		this.check(TokenKind.IDENT);

		if (this.sym == TokenKind.LPAR) {
			sn.tok.kind = TokenKind.METHOD;
			this.scan();
			if (this.sym != TokenKind.RPAR) {
				sn.addChild(this.Statement());
				while (this.sym == TokenKind.COMMA) {
					this.scan();
					sn.addChild(this.Statement());
				}
			}
			this.check(TokenKind.RPAR);
		}

		return sn;
	}

	/**
	 * Access = ":" | (Expr [":" Expr]).
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ScannerException
	 * @throws ParserException
	 */
	private SyntaxNode Access() throws ScannerException, ParserException {
		SyntaxNode sn, tmp;

		if (this.sym == TokenKind.COLON) {
			sn = new SyntaxNode(this.la);
			this.scan();
		} else {
			sn = this.Expr();
			if (this.sym == TokenKind.COLON) {
				tmp = new SyntaxNode(this.la);
				this.scan();
				tmp.addChild(sn);
				tmp.addChild(this.Expr());
				sn = tmp;
			}
		}

		return sn;
	}

	/**
	 * Value = "if" "(" Condition ")" Statement "else" Statement | "(" Statement
	 * ")" | "[" ValMatrix"]" | "{" ValMatrix}"}" | ident ["(" [Statement {","
	 * Statement}]")"] | number | string | "pi" | "e" | "i" | "true" | "false" |
	 * "null" | "end".
	 * 
	 * @return Syntax-tree
	 * 
	 * @throws ParserException
	 * @throws ScannerException
	 */
	private SyntaxNode Value() throws ParserException, ScannerException {
		SyntaxNode sn;

		switch (this.sym) {
		case IF:
			sn = new SyntaxNode(this.la);
			this.scan();
			this.check(TokenKind.LPAR);
			sn.addChild(this.Condition());
			this.check(TokenKind.RPAR);
			sn.addChild(this.Statement());
			this.check(TokenKind.ELSE);
			sn.addChild(this.Statement());
			break;
		case LBRACK:
			this.t = new Token();
			this.t.kind = TokenKind.NUMERIC;
			this.t.line = this.la.line;
			this.t.col = this.la.line;
			sn = new SyntaxNode(this.t);
			this.scan();
			sn.childs = this.ValMatrix();
			this.check(TokenKind.RBRACK);
			break;
		case LBRACE:
			this.t = new Token();
			this.t.kind = TokenKind.ARRAY;
			this.t.line = this.la.line;
			this.t.col = this.la.line;
			sn = new SyntaxNode(this.t);
			this.scan();
			sn.childs = this.ValMatrix();
			this.check(TokenKind.RBRACE);
			break;
		case LPAR:
			this.scan();
			sn = this.Statement();
			this.check(TokenKind.RPAR);
			break;
		case IDENT:
			sn = new SyntaxNode(this.la);
			this.scan();
			if (this.sym == TokenKind.LPAR) {
				// Function
				sn.tok.kind = TokenKind.FUNCTION;
				this.scan();
				if (this.sym != TokenKind.RPAR) {
					sn.addChild(this.Statement());
					while (this.sym == TokenKind.COMMA) {
						this.scan();
						sn.addChild(this.Statement());
					}
				}
				this.check(TokenKind.RPAR);

			} else {
				// Variable
				sn.tok.kind = TokenKind.VAR;
				if (!this.vars.contains(this.t.str)) {
					this.vars.add(this.t.str);
				}
			}
			break;
		case NUMBER:
		case STRING:
		case PI:
		case E:
		case I:
		case TRUE:
		case FALSE:
		case NULL:
		case END:
			sn = new SyntaxNode(this.la);
			this.scan();
			break;
		default:
			throw new ParserException("Missplaced token " + this.la.kind.getDescription(), this.la);
		}

		return sn;
	}

	/**
	 * ValMatrix = ValRow {";" ValRow}.
	 * 
	 * @return List of Syntax-nodes
	 * 
	 * @throws ParserException
	 * @throws ScannerException
	 */
	private SyntaxNode[] ValMatrix() throws ParserException, ScannerException {
		SyntaxNode sn;
		Token t;
		SyntaxNode[] list;

		t = new Token();
		t.kind = TokenKind.ROW;
		t.line = t.line;
		t.col = t.col;
		sn = new SyntaxNode(t);
		sn.childs = this.ValRow();
		list = new SyntaxNode[] { sn };

		while (this.sym == TokenKind.SEMICOLON) {
			sn = new SyntaxNode(this.la);
			this.scan();
			sn.childs = this.ValRow();
			list = Arrays.copyOf(list, list.length + 1);
			list[list.length - 1] = sn;
		}

		return list;
	}

	/**
	 * ValRow = Statement {[","] Statement}.
	 * 
	 * @return List of Syntax-nodes
	 * 
	 * @throws ParserException
	 * @throws ScannerException
	 */
	private SyntaxNode[] ValRow() throws ParserException, ScannerException {
		SyntaxNode[] list;

		list = new SyntaxNode[] { this.Statement() };
		while ((this.sym == TokenKind.COMMA) || ((this.sym != TokenKind.SEMICOLON) && (this.sym != TokenKind.RBRACK)
				&& (this.sym != TokenKind.RBRACE))) {
			if (this.sym == TokenKind.COMMA) {
				this.scan();
			}
			list = Arrays.copyOf(list, list.length + 1);
			list[list.length - 1] = this.Statement();
		}

		return list;
	}
}
