/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.utils;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

import at.lcm.bu21.general.utils.FileUtils;

/**
 * Open zip-archiv and export files
 * 
 * @author Werner Koppelsttter
 */
public class ReadZip {

	/** Zip-archive */
	ZipFile zip;

	/**
	 * Create class with zip-archive
	 * 
	 * @param file
	 *            Name of zip-file
	 * @throws IOException
	 * @throws ZipException
	 */
	public ReadZip(String file) throws ZipException, IOException {
		this(new File(file));
	}

	/**
	 * Create class with zip-archive
	 * 
	 * @param file
	 *            Zip-file
	 * @throws IOException
	 * @throws ZipException
	 */
	public ReadZip(File file) throws ZipException, IOException {
		this.zip = new ZipFile(file);
	}

	/**
	 * List entries of zip-archive
	 * 
	 * @return List of entries
	 */
	public String[] listEntries() {
		Enumeration<? extends ZipEntry> entries;
		LinkedList<String> list = new LinkedList<String>();

		entries = this.zip.entries();
		while (entries.hasMoreElements()) {
			list.add(entries.nextElement().getName());
		}

		return list.toArray(new String[0]);
	}

	/**
	 * Extract archive to directory
	 * 
	 * @param directory
	 *            Extract to this directory
	 * @throws IOException
	 */
	public void unzip(String directory) throws IOException {
		this.unzip(new File(directory));
	}

	/**
	 * Extract archive to directory
	 * 
	 * @param directory
	 *            Extract to this directory
	 * @throws IOException
	 */
	public void unzip(File directory) throws IOException {
		ZipEntry ze;

		for (String str : this.listEntries()) {
			ze = this.zip.getEntry(str);
			if (ze != null) {
				this.unzip(new File(directory, ze.getName()), ze);
			}
		}
	}

	/**
	 * Extract entry into directory
	 * 
	 * @param directory
	 *            Extract to this directory
	 * @param entry
	 *            Entry to extract
	 * @throws IOException
	 */
	public void unzip(String directory, String entry) throws IOException {
		this.unzip(new File(directory), entry);
	}

	/**
	 * Extract entry into directory
	 * 
	 * @param directory
	 *            Extract to this directory
	 * @param entry
	 *            Entry to extract
	 * @throws IOException
	 */
	public void unzip(File directory, String entry) throws IOException {
		ZipEntry ze;
		File file;

		ze = this.zip.getEntry(entry);
		if (ze != null) {
			file = new File(ze.getName());
			this.unzip(new File(directory, file.getName()), ze);
		}
	}

	/**
	 * Extract entry into file
	 * 
	 * @param file
	 *            Extract to this file
	 * @param entry
	 *            Entry to extract
	 * @throws IOException
	 */
	public void unzipFile(String file, String entry) throws IOException {
		this.unzipFile(new File(file), entry);
	}

	/**
	 * Extract entry into file
	 * 
	 * @param file
	 *            Extract to this file
	 * @param entry
	 *            Entry to extract
	 * @throws IOException
	 */
	public void unzipFile(File file, String entry) throws IOException {
		ZipEntry ze;

		ze = this.zip.getEntry(entry);
		if (ze != null) {
			this.unzip(file, ze);
		}
	}

	/**
	 * Extract entry to file
	 * 
	 * @param file
	 *            Extract to file
	 * @param entry
	 *            Entry to extract
	 * @throws IOException
	 */
	protected void unzip(File file, ZipEntry entry) throws IOException {
		BufferedInputStream is = null;
		BufferedOutputStream os = null;

		// Unzip
		if (entry.isDirectory()) {
			file.mkdirs();
		} else {
			file.getParentFile().mkdirs();
			os = new BufferedOutputStream(new FileOutputStream(file));
			is = new BufferedInputStream(this.zip.getInputStream(entry));
			FileUtils.copy(is, os);
			is.close();
			os.close();
		}
	}

	/**
	 * Get InputStream of entry
	 * 
	 * @param entry
	 *            Entry to unzip
	 * @return InputStream of entry
	 * @throws IOException
	 */
	public InputStream unzipStream(String entry) throws IOException {
		InputStream in = null;
		ZipEntry ze;

		ze = this.zip.getEntry(entry);
		if (ze != null) {
			in = this.zip.getInputStream(ze);
		}

		return in;
	}

	/**
	 * Close zip-archive
	 * 
	 * @throws IOException
	 */
	public void close() throws IOException {
		this.zip.close();
	}

}
