/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.gui.view.pubinterface;

import java.rmi.RemoteException;

import javax.swing.JTable;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.AbstractTableModel;

import at.lcm.bu21.magopt.structure.IParamContainer;
import at.lcm.bu21.magopt.structure.Parameter.ParameterPath;

/**
 * Table model for public-interface table.
 * 
 * @author Werner Koppelsttter
 */
public class PubInterfaceModel extends AbstractTableModel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/**
	 * List of parameters
	 */
	private IParamContainer container = null;

	/**
	 * Set items to display depends
	 * 
	 * @param container
	 *            ParamContainer
	 */
	public void setContainer(IParamContainer container) {
		this.container = container;

		// refresh table of model
		for (TableModelListener listener : this
				.getListeners(TableModelListener.class)) {
			if ((listener instanceof JTable)
					&& this.equals(((JTable) listener).getModel())) {
				listener.tableChanged(new TableModelEvent(this));
			}
		}
	}

	@Override
	public int getColumnCount() {
		return PubInterfaceColumn.values().length;
	}

	@Override
	public int getRowCount() {
		int len = 0;

		if (this.container != null) {
			try {
				len = this.container.listPublicInterfaceName().length + 1;
			} catch (RemoteException e) {
				e.printStackTrace();
			}
		}

		return len;
	}

	@Override
	public Object getValueAt(int row, int column) {
		ParameterPath path;
		Object obj = null;
		String[] list;
		String str;

		if (this.container != null) {
			try {
				list = this.container.listPublicInterfaceName();
				// Arrays.sort(list);
				if (row < list.length) {
					if (column == PubInterfaceColumn.NAME.getValue()) {
						obj = list[row];
					} else if (column == PubInterfaceColumn.PARAMETER
							.getValue()) {
						obj = this.container.getPublicInterfaceParam(list[row]);
					} else if (column == PubInterfaceColumn.CHECK.getValue()) {
						path = (ParameterPath) this.getValueAt(row, -1);
						obj = (path.getParameter() != null)
								|| (path.getPath().getPath() != null);
					} else {
						str = (String) this.getValueAt(row,
								PubInterfaceColumn.NAME.getValue());
						if ((str != null) && !str.isEmpty()) {
							obj = this.container.getPublicInterfacePath(str);
						}
					}
				}
			} catch (RemoteException e) {
				e.printStackTrace();
			}
		}

		return obj;
	}

	@Override
	public String getColumnName(int column) {
		for (PubInterfaceColumn e : PubInterfaceColumn.values()) {
			if (e.getValue() == column) {
				return e.getName();
			}
		}
		return null;
	}

	@Override
	public boolean isCellEditable(int row, int column) {
		boolean edit = false;
		Object obj;

		if (column == PubInterfaceColumn.NAME.getValue()) {
			edit = true;
		} else if (column == PubInterfaceColumn.PARAMETER.getValue()) {
			obj = this.getValueAt(row, PubInterfaceColumn.NAME.getValue());
			if ((obj != null) && !((String) obj).isEmpty()) {
				edit = true;
			}
		}

		return edit;
	}

	@Override
	public Class<?> getColumnClass(int column) {
		Class<?> type = String.class;

		if (column == PubInterfaceColumn.CHECK.getValue()) {
			type = Boolean.class;
		}

		return type;
	}

	@Override
	public void setValueAt(Object value, int row, int col) {
		String name;

		name = (String) this.getValueAt(row,
				PubInterfaceColumn.NAME.getValue());
		try {
			if (name == null) {
				if (col == PubInterfaceColumn.NAME.getValue()) {
					this.container.addPublicInterface((String) value);
				}
			} else {
				if (col == PubInterfaceColumn.NAME.getValue()) {
					if ((value == null) || ((String) value).isEmpty()) {
						this.container.removePublicInterface(name);
					} else {
						if (name.compareTo((String) value) != 0) {
							this.container.renamePublicInterface(name,
									(String) value);
						}
					}
				} else if (col == PubInterfaceColumn.PARAMETER.getValue()) {
					this.container.setPublicInterface(name, (String) value);
				}
			}
		} catch (RemoteException e) {
			e.printStackTrace();
		}
		this.fireTableDataChanged();
	}
}
