/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 251 $
 * $Date: 2020-06-22 12:45:24 +0200 (Mon, 22 Jun 2020) $
 */
package at.lcm.bu21.optim.algorithm.grid;

import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.optim.algorithm.JPanelSolver;
import at.lcm.bu21.optim.algorithm.Solver;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.VariableDesc;

/**
 * Grid implementation
 * 
 * @author Werner Koppelsttter
 */
public class Grid extends Solver {

	/** Index of next individual */
	private int index = 0;

	/**
	 * Create optimizer
	 * 
	 * @param problem
	 *            Based problem with description of variables and results
	 */
	public Grid(Problem problem) {
		super(problem);
	}

	@Override
	public String getOptimizerName() {
		return "Grid";
	}

	@Override
	public String getDescription() {
		return "Grid Algorithm";
	}

	@Override
	public JPanelSolver getPanel() {
		return new GridPanel(this);
	}

	/**
	 * Get size of archive
	 * 
	 * @return Size of archive
	 */
	public int getArchiveSize() {
		VariableDesc[] varDesc;
		int num = 1;

		varDesc = this.getProblem().getVariables();
		for (VariableDesc vd : varDesc) {
			num *= vd.getNumSteps();
		}

		return num;
	}

	@Override
	public Individual getNextIndividual() {
		Individual ind = null;
		int tmp;

		if (this.hasNext()) {
			// create individuals
			VariableDesc[] varDesc = this.getProblem().getVariables();
			tmp = this.index++;
			try {
				ind = new Individual(this.getProblem().getArchive());
				for (int n = 0; n < varDesc.length; n++) {
					if (varDesc[n].isVaried()) {
						ind.setVariable(n, varDesc[n].round(
								varDesc[n].getLB() + ((tmp % varDesc[n].getNumSteps()) * varDesc[n].getStepSize())));
						tmp /= varDesc[n].getNumSteps();
					} else {
						ind.setVariable(n, varDesc[n].getLB());
					}
				}
			} catch (Exception ex) {
				System.out.println("Error by creation individium");
			}
		}

		return ind;
	}

	@Override
	public void setResolvedIndividual(Individual ind) {
	}

	@Override
	public boolean hasNext() {
		return this.index < this.getArchiveSize();
	}

	@Override
	public Individual[] getArchive() {
		return new Individual[0];
	}

	@Override
	public void setArchive(Individual[] ind) {
	}

	@Override
	public Individual[] getPareto() {
		return new Individual[0];
	}

	@Override
	public TStruct getSettings() throws Exception {
		return new TStruct();
	}

	@Override
	public void setSettings(TStruct resumeData) throws Exception {
	}

	@Override
	public TStruct getState() throws Exception {
		return new TStruct();
	}

	@Override
	public void setState(TStruct resumeData) throws Exception {
	}

	@Override
	public String validateProblem() {
		String msg;
		double num = 1.0;

		msg = super.validateProblem();
		if (msg == null) {
			for (VariableDesc var : this.getProblem().getVariables()) {
				num *= var.getNumSteps();
			}
			if (Double.isInfinite(num)) {
				msg = "Gird is unlimited (Define steps greater null)";
			}
		}

		return msg;
	}
}
