import matplotlib.pyplot as plt
import numpy as np
from symspace.wrapper.symoptwrapper import SymArchive
from symspace.wrapper.symplot import SymSubPlot, SymPlot2Scatter, SymPlot2Line

try:
    '''latest symspace version only'''
    from symspace.plotting.matplot import FigureMPL
except ImportError:
    FigureMPL = None


def plotFromProject(prj_file, a_name, p_name):
    archive = SymArchive.load(prj_file, a_name)
    problem = archive.readProblem(p_name)
    plotFromProblem(problem)


def plotFromProblem(problem):
    local_pareto = False  # set full or local pareto
    plot_optimum = False  # plot only Min or Max of each Savepoint

    objectives_desc = problem.getObjectives()

    objectives_names = [desc.getName() for desc in objectives_desc]
    objectives_type = [str(desc.getObjective()).split('.')[-1] for desc in objectives_desc]

    d = {}
    for key in objectives_names:
        d[key] = [[], []]

    for sp_index in range(problem.getNumSavePoints()):
        # print('SavePoint[{:d}]'.format(sp_index))
        save_point = problem.getSavePoint(sp_index)
        if local_pareto:
            pareto = save_point.getLocalPareto()
        else:
            pareto = save_point.getFullPareto()
        for key in objectives_names:
            d[key][0].append([])
            d[key][1].append([])
        for i_index, individual in enumerate(pareto):
            # print('\tIndividual[{:d}]'.format(i_index))
            for name in objectives_names:
                objective_value = individual.getResult(name)
                # print('\t\t', name, objective_value)
                d[name][0][sp_index].append(sp_index)
                d[name][1][sp_index].append(objective_value)

    n_rows = len(objectives_names)

    if FigureMPL is not None:
        '''plot with matplotlib'''
        fig = plt.Figure()
        n_cols = 1  # adjust plot columns
        for index, key in enumerate(objectives_names):
            ylabel = '{} ({})'.format(key.split('.')[-1], objectives_type[index])
            ax = fig.add_subplot(n_rows, n_cols, 1 + index)
            x = d[key][0]
            y = d[key][1]

            if objectives_type[index] == 'MIN':
                y_flat = []
                x_flat = []
                for sp_index in range(len(x)):
                    y_flat.append(np.min(y[sp_index]))
                    x_flat.append(sp_index)
                ax.plot(x_flat, y_flat, 'r', marker='o')
            elif objectives_type[index] == 'MAX':
                y_flat = []
                x_flat = []
                for sp_index in range(len(x)):
                    y_flat.append(np.max(y[sp_index]))
                    x_flat.append(sp_index)
                ax.plot(x_flat, y_flat, 'r', marker='o')

            if not plot_optimum:
                y_flat = [item for sublist in y for item in sublist]
                x_flat = [item for sublist in x for item in sublist]
                ax.scatter(x_flat, y_flat)

            ax.set_xlabel('Savepoints')
            ax.set_ylabel(ylabel)
            ax.grid(True)

        fig_mpl = FigureMPL(fig)
        fig_mpl.showQT()

    else:
        '''plot with java plot'''
        sym_subplot = SymSubPlot(n_rows, 1)

        for index, key in enumerate(objectives_names):
            ylabel = '{} ({})'.format(key.split('.')[-1], objectives_type[index])
            x = d[key][0]
            y = d[key][1]

            if plot_optimum and objectives_type[index] == 'MIN':
                y_flat = []
                x_flat = []
                for sp_index in range(len(x)):
                    y_flat.append(np.min(y[sp_index]))
                    x_flat.append(sp_index)
                sym_subplot[index, 0] = SymPlot2Line()
                sym_subplot[index, 0].setXLimit(0, len(x_flat))
            elif plot_optimum and objectives_type[index] == 'MAX':
                y_flat = []
                x_flat = []
                for sp_index in range(len(x)):
                    y_flat.append(np.max(y[sp_index]))
                    x_flat.append(sp_index)

                sym_subplot[index, 0] = SymPlot2Line()
                sym_subplot[index, 0].setXLimit(0, len(x_flat))
            else:
                y_flat = [item for sublist in y for item in sublist]
                x_flat = [item for sublist in x for item in sublist]
                sym_subplot[index, 0] = SymPlot2Scatter()

            sym_subplot[index, 0].addSignal('', x_flat, y_flat)
            sym_subplot[index, 0].setXLabel('Savepoints')
            sym_subplot[index, 0].setYLabel(ylabel)

        sym_subplot.show()


if __name__ == '__main__':
    if 'Problem' in locals():
        '''Use from Optimizer Python Console in DebugMode (config.local -> OptimizerDebug=true)'''
        plotFromProblem(problem=locals()['Problem'])
    elif '__project__' in locals() or 'CurProject' in locals():
        '''Use from SyMSpace Console'''
        if '__project__' in locals():
            project = locals()['__project__']
        else:
            project = locals()['CurProject']
        if 'archive_name' not in locals():
            print('Unable to run script: Input parameters missing!')
            print('\tDefine "archive_name" in Python Console!')
        elif 'problem_name' not in locals():
            print('Unable to run script: Input parameters missing!')
            print('\tDefine "problem_name" in Python Console!')
        else:
            archive_name = locals()['archive_name']
            problem_name = locals()['problem_name']
            prj_file = project.javaObj.getFile()
            plotFromProject(prj_file, archive_name, problem_name)
    else:
        print('Unable to run script: Input parameters missing!\n')
        print('For usage in SyMSpace Center:')
        print('\tDefine "archive_name" in Python Console')
        print('\tDefine "problem_name" in Python Console')
        print('\tStart as Macro: SyMSpace Menu: Edit -> Run Macro (Ctrl+M)')
        print('\tOr assign "CurProject": SyMSpace Menu: '
              'Debug -> Python -> Assign Project Handle in Python Console (Ctrl+P)')

        print('For usage in SyMSpace Optimizer:')
        print('\tset in config.local -> OptimizerDebug=true')
        print('\tOptimizer Menu: Debug -> Assign problem to python')
