#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision$
$LastChangedDate::                           $

This file is part of SyMSpace.

Copyright (c) 2019, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
import logging
import tempfile
import numpy as np
import pandas as pd
from lcm.tdata_interface.pyjdata_interface import PyJDataInterface
from symspace.wrapper.symwrapper import SymProject, SymField, SymContainer, SymMaterial


class ModelDescription:
    """Class enables to:
    - create a description of a SyMSpace project including the most important informations
    - create a SyMSpace Project from a description file

    Excel <-> mop-File
    ------------------

    Convert a excel file to a SyMSpace project an vice versa.

    A Excel-file contains the sheets: Inparam, LinkParam and Components

    Inparam contains the columns:

    - Description: Read/write field description (Only for informational purpose)
    - Identifier:  Read/write field identifier
    - Value: Read/write field value
    - Unit: Read/write field value
    - Comment: Read/write field value (Needed because it may contain component webpath)
    - TexString: Read/write field value

    LinkParam contains the columns:

    - Flags: Field flags: Constant, Locked or ReadOnly
    - Identifier: Identifier of all fields in the (original) project
    - Link to: Create field links
    - Used by: Read field links (Only for informational purpose)

    Components contains the columns:

    - Description: Read/write component description (Only for informational purpose)
    - Container: Read container id / insert component in container
    - Identifier:  Read/write component identifier
    - Webpath:  Read component identifier from comment field of components / import component from webpath

     """

    @classmethod
    def convertDeprecatedXLS(cls, xls_file: str, index: int = None,
                             fail_on_error: bool = True, add_missing_containers: bool = True):

        prj_name = os.path.basename(xls_file).replace('.xlsx', '')
        xls_file_v2 = os.path.join(os.path.dirname(xls_file), prj_name + '_v2.xlsx')
        tmp_dir = tempfile.mkdtemp()
        mop_path = os.path.join(tmp_dir, prj_name + '.mop')
        project = SymProject.new(prj_name, prj_name, mop_path)
        tm = ModelDescription(project, index)
        tm.fail_on_error = fail_on_error
        tm.import_add_missing_containers = add_missing_containers
        tm.updateFromExcel(xls_file)
        tm.exportToExcel(xls_file_v2)

    def __init__(self, project: SymProject, index: int = None, log_level: str = 'INFO'):
        """
        Create a instance of this class

        Parameters
        ----------
        project : SymProject
            SyMSpace project to load (export Model-Description) or update (import Model-Description)

        index : int
            Project index, Zero-Based
        log_level: str
            Output only messages higher or equal to the loglevel
            e.g. 'DEBUG', 'INFO', 'WARNING', 'ERROR' or 'CRITICAL'
        """
        self.project = project
        self.logger = self.__getLogger(log_level)
        self.errors = 0
        self.error_msgs = []
        self.warning_msgs = []
        self.fail_on_error = False

        '''import switches'''
        self.import_insert_components = True
        self.import_add_missing_containers = False
        self.import_add_missing_fields = False
        '''export switches'''
        self.export_scalar_only = True
        self.export_eval_used_by = False

        if index is not None:
            self.project.setIndex(index)

    def getErrorString(self):
        """Return errors as a formatted string

        Returns
        -------

        str
        """
        return '\n'.join(self.error_msgs)

    def getWarningString(self):
        """Return warnings as a formatted string

        Returns
        -------

        str
        """
        return '\n'.join(self.warning_msgs)

    def exportToExcel(self, xls_file):
        self.__log('INFO', 'Start Excel Export')
        df_param = self.__getDynamicParams()
        df_link_param = self.__getLinkParams()
        df_comp = self.__getComponents()

        with pd.ExcelWriter(xls_file) as writer:
            df_param.to_excel(writer, sheet_name='Inparam', index=False)
            df_link_param.to_excel(writer, sheet_name='LinkParam', index=False)
            df_comp.to_excel(writer, sheet_name='Components', index=False)

        self.__log('INFO', 'Saved params to: {}'.format(xls_file))

    def updateFromExcel(self, xls_file):
        self.__log('INFO', 'Start update from Excel-file: {}'.format(xls_file))
        d = pd.read_excel(xls_file, sheet_name=None)

        if 'Components' in d and self.import_insert_components:
            self.__insertComponentsFromDF(d['Components'])
        else:
            '''Error if sheet is missing'''
            self.__log('ERROR', 'x-- "Components" sheet is missing!')

        if 'Inparam' in d:
            self.__updateParamsFromDF(d['Inparam'])
        else:
            '''Error if sheet is missing'''
            self.__log('ERROR', 'x-- "Inparam" sheet is missing!')

        if 'LinkParam' in d:
            self.__linkParamsFromDF(d['LinkParam'])
        else:
            # Todo: remove if xls files have been updated and replace with an error (same as in Components and Inparam)
            self.__log('WARNING', '|-- Deprecated XLS-File: "LinkParam" sheet is missing, '
                                  'trying to use "Inparam" for linking instead.')
            self.__linkParamsFromDF(d['Inparam'])

        if self.errors > 0:
            log_level = 'WARNING'
        else:
            log_level = 'INFO'
        self.__log(log_level, '\\--Total number of errors: {:d}'.format(self.errors))

        self.project.javaObj.restartJep()

    # def updateFromJSON(self, json_file):
    #     df = pd.read_json(json_file)
    #     self.updateParamsFromDF(df)
    #
    # def updateFromCSV(self, csv_file):
    #     df = pd.read_csv(csv_file)
    #     self.updateParamsFromDF(df)

    @staticmethod
    def __getLogger(log_level: str):
        logger = logging.getLogger('model_description')
        c_handler = logging.StreamHandler()
        c_format = logging.Formatter('%(levelname)8s - %(message)s')
        c_handler.setFormatter(c_format)
        if log_level.lower() == 'info':
            logger.setLevel(logging.INFO)
            c_handler.setLevel(logging.INFO)
        elif log_level.lower() == 'warning':
            logger.setLevel(logging.WARNING)
            c_handler.setLevel(logging.WARNING)
        elif log_level.lower() == 'error':
            logger.setLevel(logging.ERROR)
            c_handler.setLevel(logging.ERROR)
        elif log_level.lower() == 'critical':
            logger.setLevel(logging.CRITICAL)
            c_handler.setLevel(logging.CRITICAL)
        else:
            logger.setLevel(logging.DEBUG)
            c_handler.setLevel(logging.DEBUG)
        logger.addHandler(c_handler)
        return logger

    def __log(self, log, msg):
        if log.lower() == 'info':
            self.logger.info(msg)
        elif log.lower() == 'warning':
            self.logger.warning(msg)
            self.warning_msgs.append(msg)
        elif log.lower() == 'error':
            self.logger.error(msg)
            self.errors += 1
            self.error_msgs.append(msg)
            if self.fail_on_error:
                raise Exception(msg)
        elif log.lower() == 'critical':
            self.logger.critical(msg)
            self.errors += 1
            self.error_msgs.append(msg)
            if self.fail_on_error:
                raise Exception(msg)
        else:
            self.logger.debug(msg)

    @staticmethod
    def __getReducedIdentifier(tmp_param):
        return '.'.join(tmp_param.getIdentifier().split('.')[1:])

    def __getDynamicParams(self) -> pd.DataFrame:
        params = self.project.getParameters('SymField')
        data = dict(Description=[], Identifier=[], Value=[], Unit=[], Comment=[], TexString=[])

        for param in params:

            identifier = self.__getReducedIdentifier(param)
            # if identifier == '' or identifier is None:
            #     '''skip nan identifier'''
            #     self.__log('INFO', '|-- Skipped: {}'.format(param.getName()))
            #     continue

            if param.getLock() or param.getConstant() or param.getReadOnly():
                '''skip locked, constant and read-only params'''
                self.__log('INFO', '|-- Skipped: {}'.format(param.getName()))
                continue

            tmp_value = param.getData()
            if self.export_scalar_only:
                if not isinstance(tmp_value, (int, float)):
                    '''skip non-scalar'''
                    self.__log('INFO', '|-- Skipped: {}'.format(param.getName()))
                    continue
            elif isinstance(tmp_value, np.ndarray) and tmp_value.dtype == object:
                '''Skip TArray'''
                self.__log('INFO', '|-- Skipped: {}'.format(param.getName()))
                continue
            elif not isinstance(tmp_value, (bool, str, int, float, np.ndarray)):
                '''Skip TNull, TStruct, TPeriodic, TPeriodicMatrix'''
                self.__log('INFO', '|-- Skipped: {}'.format(param.getName()))
                continue

            data['Description'].append(param.getDescription())
            data['Identifier'].append(identifier)

            if self.export_scalar_only:
                value = param.getData()
            else:
                value = PyJDataInterface.convert_J2Py(param.getJavaData())  # required for numeric arrays

            if isinstance(value, bool):
                value = str(value)

            data['Value'].append(value)
            data['Unit'].append(param.getUnit())
            data['Comment'].append(param.getComment())
            data['TexString'].append(param.getTexString())

            self.__log('INFO', '+-- Adding Value: {} = {} {}'.format(data['Identifier'][-1], data['Value'][-1],
                                                                     data['Unit'][-1]))
        df = pd.DataFrame.from_dict(data)

        '''remove empty rows'''
        df = df.dropna(how='all')
        return df

    def __getLinkParams(self) -> pd.DataFrame:
        params = self.project.getParameters('SymParameter')

        data = {'Flags': [], 'Identifier': [], 'Link to': [], 'Used by': []}

        for param in params:

            identifier = self.__getReducedIdentifier(param)
            if identifier == '' or identifier is None:
                '''skip nan identifier'''
                self.__log('INFO', '|-- Skipped: {}'.format(param.getName()))
                continue

            flags = ''
            if param.getReadOnly():
                flags += 'R'
            if isinstance(param, SymField):
                if param.getLock():
                    flags += 'L'
                if param.getConstant():
                    flags += 'C'
            data['Flags'].append(flags)
            data['Identifier'].append(identifier)
            data['Link to'].append('')  # same as used by but needs to be set manually
            if self.export_eval_used_by:
                used_by_str = self.__getUsedBy(param, identifier)
            else:
                used_by_str = 'not-evaluated'
            data['Used by'].append(used_by_str)

        df = pd.DataFrame.from_dict(data)

        '''remove empty rows'''
        df = df.dropna(how='all')
        return df

    def __getUsedBy(self, param, identifier):
        used_by_str = ''
        try:
            if param.getUsed():
                '''check if parameter is used in a function, formula or linked field'''
                used_by_list = []

                for param_usedby in param.getUsedBy():
                    if not isinstance(param_usedby, SymMaterial):
                        '''
                        loop over all the parameters (functions, ...), 
                        where the parameter (field) is linked to 
                        '''
                        for key, value in param_usedby.getImportAssignments().items():
                            ''' loop over all the import assignments of the parameter (function, ...)'''
                            if value != '':
                                '''check if the import value is assigned to a field'''
                                param_used_id = self.__getReducedIdentifier(param_usedby.getObject(value))

                                if param_used_id == identifier:
                                    '''
                                    check if the id of the assigned field 
                                    is the same as the id of the parameter
                                    '''
                                    param_usedby_id = self.__getReducedIdentifier(param_usedby)
                                    '''
                                    create a id by combining the parameter id (functions, ...)
                                    with the import value
                                    '''
                                    import_id = '.'.join([param_usedby_id, key])

                                    if import_id not in used_by_list:
                                        '''check if the import id is already in the list'''
                                        used_by_list.append(import_id)

                if len(used_by_list) > 0:
                    used_by_str = ';'.join(used_by_list)
        except Exception as ex:
            self.__log('ERROR', 'x-- {} => {}'.format(identifier, str(ex)))
        return used_by_str

    def __getComponents(self) -> pd.DataFrame:
        params = self.project.getParameters('SymComponent')

        data = dict(Description=[], Identifier=[], Container=[], Webpath=[])

        for param in params:

            identifier = self.__getReducedIdentifier(param)
            if identifier == '':
                identifier = param.getIdentifier()
            container = '.'.join(param.getIdentifier().split('.')[1:-1])
            webpath = param.getComment()
            description = param.getDescription()

            if webpath == '':
                self.__log('WARNING', 'Missing webpath for Component at: "{}". '
                                      'Set the correct webpath manually!'.format(identifier))

            data['Description'].append(description)
            data['Identifier'].append(identifier)
            data['Container'].append(container)
            data['Webpath'].append(webpath)

        df = pd.DataFrame.from_dict(data)
        '''remove empty rows'''
        df = df.dropna(how='all')
        return df

    def __updateParam(self, identifier: str, value, comment: str, unit: str, description: str, tex: str):
        """
        Update a parameter

        Parameters
        ----------
        identifier: str
        value
        comment: str
        unit: str
        description: str
        tex: str
        """
        id_list = identifier.split('.')
        param_path = '.'.join(id_list[:-1])
        param_name = id_list[-1]

        if param_path == '':
            '''use toplevel identifier instead'''
            param_path = self.project.getIdentifier()
            self.__log('WARNING', '|-- Using Project as Container')

        if self.project[param_path] is None and self.import_add_missing_containers:
            '''add additional containers if the path does not exist'''
            param_path_list = param_path.split('.')
            param_path_list.insert(0, self.project.getIdentifier())

            for k in range(len(param_path_list)):
                tmp_param_path1 = '.'.join(param_path_list[:k + 1])
                if self.project[tmp_param_path1] is None:
                    self.__log('WARNING', '+-- Adding missing Container: {}'.format(tmp_param_path1))
                    tmp_param_path2 = '.'.join(param_path_list[:k])
                    container_name = param_path_list[k]
                    container = SymContainer(container_name)
                    self.project[tmp_param_path2].addParam(container)

        if self.project[identifier] is None and self.import_add_missing_fields:
            '''field does not exists, but should be created'''
            self.__log('WARNING', '+-- Adding Field: {}'.format(param_name))

            field = SymField(param_name)
            self.project[param_path].addParam(field)

        if self.project[identifier] is not None:
            '''field exists'''
            self.__log('INFO', '|-- Changing Value: {} = {} {}'.format(param_name, value, unit))
            field = self.project[identifier]  # type: # SymField
            field.setData(value)
            field.setDescription(description)
            field.setComment(comment)
            field.setUnit(unit)
            field.setTexString(tex)
        else:
            self.__log('ERROR', 'x-- Field Missing: {}'.format(identifier))

    def __updateParamsFromDF(self, df: pd.DataFrame):
        """
        Change or add fields with values

        Examples
        --------

        Excel Sheet Name: "Inparam"

        +-------------------------+-----------------------------+-------+-------+
        |       Description       |         Identifier          | Value | Unit  |
        +-------------------------+-----------------------------+-------+-------+
        | Nominal torque          | Specification.Tn            |   100 | N.m   |
        | Nominal speed           | Specification.nn            |  1800 | min-1 |
        | Minimum DC link voltage | Specification.Udc_min       |   500 | V     |
        | SkewingAngle            | Geometry.Rotor.SkewingAngle |    20 | °     |
        | SkewingSteps            | Geometry.Rotor.SkewingSteps |     3 |       |
        +-------------------------+-----------------------------+-------+-------+

        Parameters
        ----------
        df : pd.DataFrame

        """

        '''remove empty rows'''
        df = df.dropna(how='all')
        '''replace NaN in "Unit" column'''
        df['Unit'].fillna(value='')

        for index, row in df.iterrows():
            description = row['Description']
            identifier = row['Identifier']
            value = row['Value']
            comment = row['Comment'] if 'Comment' in row else ''
            tex = row['TexString'] if 'TexString' in row else ''
            unit = row['Unit']

            if isinstance(value, str):
                '''convert string to bool'''
                if value.lower() == 'true':
                    value = True
                elif value.lower() == 'false':
                    value = False
                elif value.startswith('[') and value.endswith(']'):
                    '''try to reconstruct an array'''
                    # Todo: this is just a workaround and should be replaced by a more stable conversion
                    if value.startswith('[[') and value.endswith(']]'):
                        value = np.fromstring(value[2:-2], sep=' ')
                    else:
                        value = np.fromstring(value[1:-1], sep=' ')

            if not isinstance(description, str):
                '''convert empty string'''
                description = ''
            if not isinstance(unit, str):
                '''convert empty string'''
                unit = ''
            if not isinstance(comment, str):
                '''convert empty string'''
                comment = ''
            if not isinstance(tex, str):
                '''convert empty string'''
                tex = ''

            if isinstance(identifier, str):
                try:
                    self.__updateParam(identifier, value, comment, unit, description, tex)
                except Exception as e:
                    self.__log('ERROR', 'x-- {} => {}'.format(identifier, str(e)))
            else:
                self.__log('ERROR', 'x-- Non string identifier in Inparam: row[{}],{}'.format(index, identifier))

    def __linkParamsFromDF(self, df: pd.DataFrame):
        """
        Link Fields to Functions

        Examples
        --------

        Excel Sheet Name: "LinkParam"

        +-------+----------------------------------------+---------+-----------------------------------------------+
        | Flags |               Identifier               | Link to |             Used by                           |
        +-------+----------------------------------------+---------+-----------------------------------------------+
        |       | Specification                          |         |                                               |
        |       | Specification.Tn                       |         | PMSM.Femm_Simulate.NoLoad.Tref                |
        |       | Specification.nn                       |         |                                               |
        |       | Specification.Udc_min                  |         | PMSM.PMSM_Loadpoint_Idq_RBF.Udc               |
        |       | Material_Library                       |         |                                               |
        |       | Material_Library.Material_Core         |         |                                               |
        | LC    | Material_Library.Material_Core.Type    |         |                                               |
        | LC    | Material_Library.Material_Core.Name    |         |                                               |
        | LC    | Material_Library.Material_Core.Dataset |         |                                               |
        +-------+----------------------------------------+---------+-----------------------------------------------+

        Parameters
        ----------
        df : pd.DataFrame

        """
        '''remove empty rows'''
        df = df.dropna(how='all')

        for index, row in df.iterrows():
            identifier = row['Identifier']

            import_assignment = row['Link to'] if 'Link to' in row else None

            if isinstance(identifier, str):
                try:
                    if isinstance(import_assignment, str):
                        import_list = import_assignment.split('.')
                        import_path = '.'.join(import_list[:-1])
                        import_name = import_list[-1]

                        self.project[import_path].javaObj.getDepend().setImportAssignmentsMap({import_name: identifier})
                        self.__log('INFO',
                                   '|-- Linking parameter {} -> {}'.format(identifier, import_assignment))
                except Exception as e:
                    self.__log('ERROR', 'x-- {} => {}'.format(identifier, str(e)))
            else:
                self.__log('ERROR', 'x-- Non string identifier in LinkParam: row[{}],{}'.format(index, identifier))

    def __insertComponentsFromDF(self, df):
        """
        Add components

        Examples
        --------

        Excel Sheet Name: "Components"

        +-----------------+------------------+-----------+-----------------------------------------------------------------+
        |   Description   | Identifier       | Container |                             Webpath                             |
        +-----------------+------------------+-----------+-----------------------------------------------------------------+
        | PMSM Simulation |                  |           | Simulation.Motor.Magnetic.PMSM_Interior_FEMM_V2                 |
        | Stator          | Geometry.Stator  | Geometry  | Part.MotorPart.StatorInterior.FreeCAD.Stator_Interior1c_FreeCAD |
        | Rotor           | Geometry.Rotor   | Geometry  | Part.MotorPart.RotorPMInterior.FreeCAD.Rotor_Interior26_FreeCAD |
        +-----------------+------------------+-----------+-----------------------------------------------------------------+

        Parameters
        ----------
        df : pd.DataFrame
        """

        '''remove empty rows'''
        df = df.dropna(how='all')

        for index, row in df.iterrows():
            description = row['Description']
            container_id = row['Container']
            identifier = row['Identifier']
            webpath = row['Webpath']

            if not isinstance(identifier, str):
                identifier = ''

            try:
                if webpath == '':
                    raise Exception('Missing Component-Webpath for "{}"'.format(description))

                if isinstance(container_id, str):
                    container_id = self.project.getName() + '.' + container_id
                else:
                    '''if container id is empty add below project'''
                    container_id = self.project.getName()

                if isinstance(webpath, str):

                    if self.project[container_id] is None and self.import_add_missing_containers:
                        '''add additional containers if the path does not exist'''
                        param_path_list = container_id.split('.')

                        for k in range(len(param_path_list)):
                            tmp_param_path1 = '.'.join(param_path_list[:k + 1])
                            if self.project[tmp_param_path1] is None:
                                self.__log('WARNING', '+-- Adding missing Container: {}'.format(tmp_param_path1))
                                tmp_param_path2 = '.'.join(param_path_list[:k])
                                container_name = param_path_list[k]
                                container = SymContainer(container_name)
                                self.project[tmp_param_path2].addParam(container)

                    if self.project[container_id] is not None:
                        '''container exists already'''
                        self.__log('INFO', '+-- Adding Component: {} => {}'.format(description, webpath))
                        comp_param = self.project[container_id].addComponent(webpath)

                        comp_name = identifier.split('.')[-1]
                        if comp_name != '':
                            '''set the name of the component parameter'''
                            comp_param.setName(comp_name)

                    else:
                        '''container does not exists'''
                        raise Exception('Container does not exist {}'.format(container_id))

                else:
                    self.__log('ERROR', 'x-- Non string webpath')

            except Exception as e:
                self.__log('ERROR', 'x-- {}'.format(str(e)))


if __name__ == '__main__':
    '''Needs to be executed in the SyMSpace Python Console.'''

    _prj_name = 'Demo'
    _xls_file = 'x:/Model.xlsx'
    _mop_path = 'x:/Model'
    _project = SymProject.new(_prj_name, _prj_name, _mop_path)
    _tm = ModelDescription(_project)
    _tm.updateFromExcel(_xls_file)
    _project.save()
